/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright   Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-201), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright   Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   This code is released only for internal review by Nature Communications, manuscript: NCOMMS-21-29568.

   *EndCopyright:
*/

#ifndef __AGENT_H__
#define __AGENT_H__

#include "enums.h"
#include "statistical_areas.h"
#include "debug.h"

#include <iostream>
#include <vector>
#include <map>
#include <memory>

/*
 * Person data structure
 */

class Agent
{
private:



   /* set infectious period and peak viral titer based off influenza kinetics */
   unsigned LATENT_PERIOD_ = 4;              // in half-days. non-infectious period between exposure and incubation
   //unsigned INFECTIOUS_PERIOD_        = 34;
   //unsigned RECOVERY_TIME_            = 34;
   unsigned PEAK_VIRAL_TITER_ = 10;          // 14; time from exposure to peak infectiousness (includes latent period)
   unsigned INFECTION_PERIOD_   = 34;          // total time from exposure to recovery
   // Percentage of infected agents that become symptomatic
   float ASYMPTOMATIC_INFECTIVITY_ = 0.3;   // as a proportion of symptomatic
   float PERCENT_SYMPTOMATIC       = 0.669;
   float CHILD_PERCENT_SYMPTOMATIC = 0.134; //0.223;



   constexpr static float DAYS_PER_STEP_ = 0.5;

   constexpr static float exponent_  = 10.0;
   constexpr static float exponent2_ = 20.0;

   /* Constant attributes associated to this agent */
   const unsigned int ID_;
   const AgeGroup     age_group_;

   bool new_illness_;
   bool newly_detected_;

   std::vector<std::shared_ptr<WorkingGroup> > working_groups_;
   std::shared_ptr<Household> household_;

   /* Disease states of the agent */
   DiseaseState state_;
   VacState     vacstate_; /*Kit's edit -intervention*/
   AVState      AVstate_;  /*Kit's edit -AVE*/
   bool         infected_prior_;
   bool         illness_detected_;
   bool         detected_recovered_;
   //unsigned steps_in_disease_state_;
   unsigned steps_since_av_applied_;
   unsigned steps_since_av_exposure_;
   unsigned steps_since_infection_;


   //CZ COV19 SD 17032020 - SD compliance states
   SD_State SDstate_;

   //CZ COV19 24032020 - Case Isolation states
   CI_State CIstate_;

   //CZ COV19 31032020
   HQ_State HQstate_;

   // state for school closure condition
   SC_State SCstate_;

   Compliance_State Compliance_state_;

   Isolation_State Isolation_state_; // CZ COV19 31032020 - for distinguishing between different types of isolation/distancing

   float isolation_factor_home_;
   float isolation_factor_work_;
   float isolation_factor_community_;

   int HQclock_;                     // a timer for HQ duration trigger

   double VacGroup;                   /*Kit's edit -intervention*/

   //static std::vector<double> viral_load_;
   // std::vector<double> viral_load_;
   // CZ: removing static designation


   /*
    * Bitfield representing the length of time this agent spends in
    *  - latent (max 2 days),
    *  - incubating (max 1 days)
    *  - infectious (max 6 days), and
    *  - in withdrawal (max 3 days)
    */

   /*
    * struct
    * {
    * unsigned long latent : 3;
    * unsigned long incubating : 2;
    * unsigned long infectious : 4;
    * unsigned long withdrawal : 2;
    * } SIRTransitionSteps_;
    */

   int incubating_period_;
   int AVEd_illness_supression_;

   /* Add to individual nighttime mixing groups */
   inline void addToHouseholdProbability(double factor)
   {
      this->getHH()->addInfectedAgent(this->getAgeGroup(), factor);
   }

   inline void addToClusterProbability(double factor)
   {
      this->getHC()->addInfectedAgent(this->isAdult(), factor);
   }

   inline void addToCommunityProbability(double factor)
   {
     //std::cout << "adding infected agent to CD" << std::endl;
      this->getCD()->addInfectedAgent(this->getAgeGroup(), factor);
      //std::cout << "added infected agent to CD" << std::endl;
   }

   inline void addToNeighbourhoodProbability(double factor)
   {
      this->getSLA()->addInfectedAgent(this->getAgeGroup(), factor);
   }

public:

   /* Default constructor */
   Agent(unsigned long ID, AgeGroup ag)
      : ID_(ID),
        age_group_(ag),
        household_(NULL),
        state_(DiseaseStates::StateSusceptible),
        vacstate_(VacStates::notvaccinated),
        AVstate_(AVStates::Notexposed),
        SDstate_(SD_States::SD_noncompliant), // CZ COV19 17032020
        CIstate_(CI_States::CI_noncompliant), // CZ COV19 24032020
        HQstate_(HQ_States::HQ_noncompliant),
        SCstate_(SC_States::SC_nil),
        Compliance_state_(Compliance_States::nil),
        Isolation_state_(Isolation_States::nil),
        isolation_factor_home_(1),
        isolation_factor_work_(1),
        isolation_factor_community_(1),
        HQclock_(0),
        infected_prior_(false),
        illness_detected_(false),
        detected_recovered_(false),
        AVEd_illness_supression_(0),
        steps_since_av_exposure_(0),                                                    /* Kit's edit -AVE*/
        steps_since_av_applied_(0),                                                     /* new AV dose every 10 days */
        steps_since_infection_(0),
        new_illness_(false),
        newly_detected_(false)
   {
   }

   /*
    * Query methods
    */
   inline bool isInfected(void)
   {
      if ((state_ == DiseaseStates::StateSymptomatic) || (state_ == DiseaseStates::StateAsymptomatic))
      {
         return(true);
      }
      return(false);
   }

   inline bool isIll(void)
   {
      if (state_ == DiseaseStates::StateSymptomatic)
      {
         return(true);
      }
      return(false);
   }

   inline bool isInfectedPrior(void)
   {
      return(infected_prior_);
   }

   inline bool isIllnessDetected(void)
   {
      return(illness_detected_);
   }

   inline bool wasIllnessDetected(void)
   {
     return(detected_recovered_);
   }

   inline bool isSusceptible(void)
   {
      return(state_ == DiseaseStates::StateSusceptible);
   }

   inline bool isRemoved(void)
   {
      return(state_ == DiseaseStates::StateRemoved);
   }

   inline bool isWithdrawn(void)
   {
      return(state_ == DiseaseStates::StateWithdrawn);
   }

   inline bool isChild(void)
   {
      return(age_group_ == AgeGroups::Age0to4 || age_group_ == AgeGroups::Age5to18);
   }

   inline bool isStudent(void)
   {
      return(age_group_ == AgeGroups::Age5to18 && working_groups_.size() == 3);
   }

   inline bool isAdult(void)
   {
      return(!this->isChild());
   }

   inline bool isTeacher(void)
   {
      return(this->isAdult() && working_groups_.size() == 3);
   }

   inline bool willBeSymptomatic(void)
   {
      if (incubating_period_ > 0)
      {
         //std::cout << "Agent " << this->getID() << " will be symptomatic after " << incubating_period_ << " steps." << std::endl;
         return(true);
      }
      return(false);
   }

   /*
    * Getter methods
    */
   // CZ 10 06 2020 Getter methods for disease parameter variables

   unsigned getPEAK_VIRAL_TITER_(void)
   {
      return(PEAK_VIRAL_TITER_);
   }

   unsigned getLATENT_PERIOD_(void)
   {
      return(LATENT_PERIOD_);
   }

   double getASYMPTOMATIC_INFECTIVITY_(void)
   {
      return(ASYMPTOMATIC_INFECTIVITY_);
   }

/*
 * unsigned getINFECTIOUS_PERIOD_(void)
 * {
 *    return(INFECTIOUS_PERIOD_);
 * }
 */

/*
 * unsigned getRECOVERY_TIME_(void)
 * {
 *    return(RECOVERY_TIME_);
 * }
 */
   unsigned getINFECTION_PERIOD_(void)
   {
      return(INFECTION_PERIOD_);
   }

   double getPERCENT_SYMPTOMATIC(void)
   {
      return(PERCENT_SYMPTOMATIC);
   }

   double getCHILD_PERCENT_SYMPTOMATIC(void)
   {
      return(CHILD_PERCENT_SYMPTOMATIC);
   }

   inline unsigned long getID(void)
   {
      return(ID_);
   }

   inline const std::shared_ptr<Household>& getHH(void)
   {
      return(household_);
   }

   inline const std::shared_ptr<HouseholdCluster>& getHC(void)
   {
      return(household_->getHC());
   }

   inline const std::shared_ptr<CensusDistrict>& getCD(void)
   {
      return(household_->getHC()->getCD());
   }

   inline const std::shared_ptr<StatisticalLocalArea>& getSLA(void)
   {
      return(household_->getHC()->getCD()->getSLA());
   }

   inline const std::vector<std::shared_ptr<WorkingGroup> >& getWGs(void)
   {
      return(working_groups_);
   }

   inline const std::shared_ptr<DestinationZone>& getDZ(void)
   {
      if (!working_groups_.empty())
      {
         return(working_groups_.front()->getDZ());
      }
      else
      {
         std::cout << "\ttesting... get DZ" << std::endl;
         return(NULL);
      }
   }

   inline bool isNewIllness(void)
   {
      return(new_illness_);
   }

   inline bool isNewlyDetected(void)
   {
      return(newly_detected_);
   }

   inline void resetNewIllness(void)
   {
      new_illness_ = false;
   }

   inline void resetNewDetection(void)
   {
      newly_detected_ = false;
   }

   inline AgeGroup getAgeGroup(void)
   {
      return(age_group_);
   }

   inline DiseaseState getState(void)
   {
      return(state_);
   }

   inline VacState getVacState(void)
   {
      return(vacstate_);
   }                                                         /* Kit's edit -intervention*/

   inline AVState getAVState(void)
   {
      return(AVstate_);
   }                                                      /* Kit's edit -AVE*/

///////// CZ COV19 - new states

   inline SD_State getSDState(void)
   {
      return(SDstate_);
   }                                // CZ COV19 SD 17032020

   inline CI_State getCIState(void) // CZ COV19 SD 24032020
   {
      return(CIstate_);
   }

   inline HQ_State getHQState(void)
   {
      return(HQstate_);
   }

   inline SC_State getSCState(void)
   {
      return(SCstate_);
   }

   inline Isolation_State getIsolationState(void)
   {
      return(Isolation_state_);
   }

   inline Compliance_State getComplianceState(void)
   {
      return(Compliance_state_);
   }

   inline float getIsolationFactor_home(void)
   {
      return(isolation_factor_home_);
   }

   inline float getIsolationFactor_work(void)
   {
      return(isolation_factor_work_);
   }

   inline float getIsolationFactor_community(void)
   {
      return(isolation_factor_community_);
   }

   void eval_Compliance_state(void);

   int findParent_ID(std::map<unsigned long, std::shared_ptr<Agent> >&);

/////////

   double getDaytimeProbabilityOfInfection(void);

   double getDaytimeProbabilityOfInfection(unsigned);

   double getNighttimeProbabilityOfInfection(void);

/*
 *  ///// CZ COV19 SD 17032020
 *
 *  double getDaytimeProbabilityOfInfection_SD( double );
 *
 *  double getDaytimeProbabilityOfInfection_SD( unsigned, double );
 *
 *  double getNighttimeProbabilityOfInfection_SD( void );
 *
 *  //////
 */

   double getHouseholdProbabilityOfInfection(void);

   double getHouseholdClusterProbabilityOfInfection(void);

   double getNeighbourhoodProbabilityOfInfection(void);

   double getCommunityProbabilityOfInfection(void);

   /*
    * Setter methods
    */
   // CZ 10 06 2020 - set disease parameter variables

   inline void setPEAK_VIRAL_TITER_(unsigned new_peak_time)
   {
      PEAK_VIRAL_TITER_ = new_peak_time;
   }

   inline void setLATENT_PERIOD_(unsigned new_latent_period)
   {
      LATENT_PERIOD_ = new_latent_period;
   }

/*
 * inline void setINFECTIOUS_PERIOD_(unsigned new_infectious_period)
 * {
 *    INFECTIOUS_PERIOD_ = new_infectious_period;
 * }
 */

/*
 * inline void setRECOVERY_TIME_(unsigned new_recovery_time)
 * {
 *    RECOVERY_TIME_ = new_recovery_time;
 * }
 */
   inline void setINFECTION_PERIOD_(unsigned new_infection_period)
   {
      INFECTION_PERIOD_ = new_infection_period;
   }

   inline void setPERCENT_SYMPTOMATIC(double new_factor)
   {
      PERCENT_SYMPTOMATIC = new_factor;
   }

   inline void setCHILD_PERCENT_SYMPTOMATIC(double new_factor)
   {
      CHILD_PERCENT_SYMPTOMATIC = new_factor;
   }

   inline void setASYMPTOMATIC_INFECTIVITY_(double new_factor)
   {
      ASYMPTOMATIC_INFECTIVITY_ = new_factor;
   }

   ////


   inline void addWG(std::shared_ptr<WorkingGroup> wg)
   {
      working_groups_.push_back(wg);
   }

   inline void setHH(std::shared_ptr<Household> hh)
   {
      household_ = hh;
   }

   inline void setState(DiseaseState new_state)
   {
      this->state_ = new_state;
   }

   inline void setVacState(VacState new_vacstate)
   {
      this->vacstate_ = new_vacstate;
   }                                                                                     /* Kit's edit -intervention*/

   inline void setAVState(AVState new_AVstate)
   {
      this->AVstate_ = new_AVstate;
   }                                                                                /* Kit's edit -AVE*/                                                                                  /* Kit's edit -AVE*/

/////// CZ COV19 new states

   inline void setSDState(SD_State new_SDstate)
   {
      this->SDstate_ = new_SDstate;
   }                                            // CZ COV19 SD 17032020

   inline void setCIState(CI_State new_CIstate) // CZ COV19 SD 24032020
   {
      this->CIstate_ = new_CIstate;
   }

   inline void setIsolationState(Isolation_State new_state)
   {
      this->Isolation_state_ = new_state;
   }

   inline void setSCState(SC_State new_state)
   {
      this->SCstate_ = new_state; // could add some error checking here.
   }

   inline void setHQState(HQ_State new_HQstate)
   {
      this->HQstate_ = new_HQstate;
   }

   inline void setComplianceState(Compliance_State new_state)
   {
      this->Compliance_state_ = new_state;
   }

///////

   /* Misc (non-trivial) methods */

   double computeFactor(double& VEi_1, double& VEi_2, double& AVEi);

   // CZ COV19 17032020 reverting to original function to avoid having to pass variables in
   //double computeFactor( void );

   void init(void);

   void step(void);

   void step_HQ(int duration_HQ); // CZ COV19 31032020 - updates home quarantine clock


   void setIsolationFactor_home(double FoI_SD_home, double FoI_CI_home, double FoI_HQ_home, double FoI_SC_home);

   void setIsolationFactor_work(double FoI_SD_work, double FoI_CI_work, double FoI_HQ_work, double FoI_SC_work);

   void setIsolationFactor_community(double FoI_SD_community, double FoI_CI_community, double FoI_HQ_community, double FoI_SC_community);



   bool Illness_detection(double& Detection, int Response_delay, double& TAP, double& GTAP);

   int stepAV(int& Response_delay);    /* Kit's edit -AVE*/

   void AVdose(int& Response_delay);   /* Kit's edit -AVE*/

   /* Add to all mixing groups during daytime */
   void addToDaytimeProbabilityOfInfection(double& VEi_1, double& VEi_2, double& AVEi); /* Kit's edit -intervention*/
   // CZ COV19 24032020

   /* Add to all mixing groups during nighttime */
   void addToNighttimeProbabilityOfInfection(double& VEi_1, double& VEi_2, double& AVEi); /* Kit's edit -intervention*/
   // CZ COV19 24032020, 06122020


   // CZ COV19 SD 17032020 // CZ COV19 24032020
   void addToNighttimeProbabilityOfInfection_SD(double& VEi_1, double& VEi_2, double& AVEi);

   void addToDaytimeProbabilityOfInfection_SD(double& VEi_1, double& VEi_2, double& AVEi);



   /* Add to only household mixing group during nighttime */
   void addToHouseholdProbabilityOfInfection(void);

   /* Add to all but household mixing group during nighttime */
   void addToCommunityProbabilityOfInfection(void);

   bool symptomatic_supression(void);

   void infect(double& AVEd, double& VEd_1, double& VEd_2); // CZ: added vaccine efficacy for disease (06122020)

   bool TAP_Sympotmatic_present(void);   // kit edit - intervention

   bool CD_Sympotmatic_present(void);    // kit edit - intervention

   // void infectWithSymptoms( void ){ this->init( true ); this->infect(); } /* Kit's edit -AVE*/

   //static void computeViralLoad(void);
   double computeViralLoad(int t);
   void printTiterVsTime(void);

   // CZ - removed static designation because variables can be modified

   void GTAP_administer(double& GTAP); /* Kit's edit -AVE*/
   void TAP_administer(double& TAP);   /* Kit's edit -AVE*/
   void reset_intervention(void);
};

typedef std::pair<unsigned long, std::shared_ptr<Agent> >   agent_pair_type;
typedef std::map<unsigned long, std::shared_ptr<Agent> >    agent_map_type;
//typedef std::pair<unsigned long, std::shared_ptr<Agent> >   agent_pair_type;
//typedef std::map<unsigned long, std::shared_ptr<Agent> >    agent_map_type;

#endif
