/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright   Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-201), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright   Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   This code is released only for internal review by Nature Communications, manuscript: NCOMMS-21-29568.

   *EndCopyright:
*/

#include "create_population.h"

const std::string CreatePopulation::school_grade_filename   = "/tmp/SchoolGrades.dat";
const std::string CreatePopulation::school_student_filename = "/tmp/SchoolStudents.dat";

/* Associate CDs and POW with an SLA */
void CreatePopulation::mapCD_SLA_DZ(cd_map_type& census_districts, dz_map_type& destination_zones, sla_map_type& statistical_local_areas, const std::string& database_input)
{
   /* SLA_2011.dat will only contain one column with 5 digit code */
   std::ifstream sla_data(database_input + "SLA.dat");
   unsigned long sla_id;
   unsigned long dummy;

   while (sla_data >> sla_id >> dummy)
   {
      statistical_local_areas.emplace(sla_id, std::make_shared <StatisticalLocalArea>(sla_id));
   }
   std::cout << "Created: " << statistical_local_areas.size() << " SLAs." << std::endl;

   sla_data.close();

   /* CD_2011.dat should only contain cd_id, latitude and longitude */
   std::ifstream      cd_data(database_input + "CDinSLA.dat");
   unsigned long      cd_id;
   unsigned long long dz_id;
   double             latitude;
   double             longitude;
   while (cd_data >> cd_id >> sla_id)
   {
      auto csla = statistical_local_areas.find(sla_id);
      if (csla != statistical_local_areas.end())
      {
         census_districts.emplace(cd_id, std::make_shared <CensusDistrict>(cd_id, csla->second, latitude, longitude));
      }
      else
      {
         std::cerr << "Unknown SLA: " << sla_id << std::endl;
         exit(1);
      }
   }
   cd_data.close();
   std::cout << "Created: " << census_districts.size() << " CDs." << std::endl;

   std::ifstream dz_data(database_input + "DZinSLA.dat");

   unsigned long pow_idx = 0;
   while (dz_data >> sla_id >> dz_id)
   {
      auto csla = statistical_local_areas.find(sla_id);
      if (csla != statistical_local_areas.end())
      {
         destination_zones.emplace(dz_id, std::make_shared <DestinationZone>(dz_id, csla->second));
      }
      else
      {
         std::cerr << "Unknown SLA: " << sla_id << std::endl;
         exit(1);
      }
   }
   std::cout << "Created: " << destination_zones.size() << " DZs." << std::endl;
}

void CreatePopulation::mapWG_POW(wg_map_type& working_groups, wg_map_type& schools, dz_map_type& destination_zones, const std::string& database_input)
{
   std::ifstream wgFile(database_input + "WG.dat");

   unsigned long long in_DZ;
   unsigned long      WG_start_id;
   unsigned           n_WGs;

   while (wgFile >> in_DZ >> WG_start_id >> n_WGs)
   {
      auto cdz = destination_zones.find(in_DZ);
      if (cdz != destination_zones.end())
      {
         /* fill in all these WGs */
         for (unsigned long i = WG_start_id; i < WG_start_id + n_WGs; ++i)
         {
            working_groups.emplace(i, std::make_shared <WorkingGroup>(i, WorkingGroupTypes::Workplace, cdz->second));
         }
      }
      else
      {
         std::cerr << "Unknown DZ:" << in_DZ << " (line:" << __LINE__ << ")" << std::endl;
         exit(1);
      }
   }

   wgFile.close();
   std::cout << "Created: " << working_groups.size() << " WGs." << std::endl;

   std::ifstream schoolFile(database_input + "SCHOOLinDZ.dat");

   /* TODO: set up grades and classes for working groups too */
   unsigned int  school_sz;
   unsigned long school_idx;
   for (unsigned int i = 0;
        schoolFile >> school_idx >> in_DZ >> school_sz;
        ++i)
   {
      auto cdz = destination_zones.find(in_DZ);
      if (cdz != destination_zones.end())
      {
         /* fill in all these schools */
         schools.emplace(school_idx, std::make_shared <WorkingGroup>(school_idx, WorkingGroupTypes::SchoolWhole, cdz->second));
      }
      else
      {
         std::cerr << "Unknown DZ:" << in_DZ << " (line:" << __LINE__ << ")" << std::endl;
         exit(1);
      }
   }
   schoolFile.close();
   std::cout << "Created: " << schools.size() << " Schools." << std::endl;
}

void CreatePopulation::addGradesAndClasses(agent_map_type& agents, wg_map_type& schools, wg_map_type& grades, wg_map_type& classes)
{
   std::cout << "Adding grades and classes to schools..." << std::endl;
   unsigned long grade_id = 0, class_id = 0;

   unsigned long num_unassigned_teachers = 0;

   std::ofstream school_grade_stream(school_grade_filename);
   std::ofstream school_student_stream(school_student_filename);
   for (auto& school_entry : schools)
   {
      auto&      cschool = school_entry.second;
      const auto my_dz   = cschool->getDZ();

      /* Make a list of all students/teachers attending school */
      std::vector <unsigned long> students, teachers;
      for (auto agent_idx : cschool->getAgentIdxs())
      {
         auto& cagent = agents.at(agent_idx);
         if (cagent->isChild())
         {
            students.push_back(agent_idx);
         }
         else
         {
            teachers.push_back(agent_idx);
         }
      }

      /* Assign students to grades */
      unsigned long n_students_per_grade = ceil(double( students.size()) / double( N_GRADES ));
      unsigned long n_classes_in_grade   = ceil(double( n_students_per_grade ) / N_STUDENTS_PER_CLASS);

      /*
       * std::cout << n_students_per_grade << " students per grade in school:" << cschool->getID()
       * << " (sz: " << students.size() << ") and " << n_classes_in_grade << " classes in each grade." << std::endl;
       */

      school_grade_stream << cschool->getID() << " ";
      school_student_stream << cschool->getID() << " ";
      for (unsigned i = 0; i < N_GRADES; ++i, ++grade_id)
      {
         std::vector <unsigned long> classes_in_grade;

         /* Make new grade */
         grades.emplace(grade_id, std::make_shared <WorkingGroup>(grade_id, WorkingGroupTypes::SchoolGrade, my_dz));
         for (unsigned cl = 0; cl < n_classes_in_grade; ++cl, ++class_id)
         {
            classes.emplace(class_id, std::make_shared <WorkingGroup>(class_id, WorkingGroupTypes::SchoolClass, my_dz));
            classes_in_grade.push_back(class_id);
         }
         auto my_grade = grades.rbegin()->second; //CZ: this is already deterministic
         school_grade_stream << classes_in_grade.size() << " ";

         /* Add students until grade full */
         unsigned long j;
         int           assigned_class = -1;
         for (j = 0; j < n_students_per_grade; ++j)
         {
            if (students.size() == 0)
            {
               break;
            }

            /* Select random student from list */  // CZ: this is where I will change it

            auto me = agents.at(students.front()); //CZ: using first element in list
            //  std::cout << "student: " << me->getID() << std::endl;
            //unsigned long assigned_student = rand() % students.size();
            //auto me = agents.at( students.at( assigned_student ) );

            /* Select class to assign student to */

            // CZ: class index loops through the classes in the grade
            if (assigned_class < classes_in_grade.size())
            {
               assigned_class++;
            }
            if (assigned_class >= classes_in_grade.size())
            {
               assigned_class = 0;
            }

            //unsigned long assigned_class = rand() % classes_in_grade.size();
            //auto my_class = classes.at( classes_in_grade.at( assigned_class ) );
            //  std::cout << "assigned class: " << assigned_class + 1 << std::endl;
            //  std::cout << "of " << classes_in_grade.size() << std::endl;
            auto my_class = classes.at(classes_in_grade.at(assigned_class));

            /* Assign me to grade and class */
            my_grade->addAgentIdx(me->getID());
            my_class->addAgentIdx(me->getID());

            /* Assign grade and class to me */
            me->addWG(my_grade);
            me->addWG(my_class);

            /* Remove student from list */
            students.erase(students.begin()); //CZ: just erase first element, the student we placed
            //auto start_it = students.begin();
            //students.erase( start_it + assigned_student );
         }

         school_student_stream << j << " ";

         /* Assign one teacher per class */ //CZ: I thought there were 3 teachers per class?
         for (auto class_id : classes_in_grade)
         {
            auto my_class = classes.at(class_id);

            for (unsigned j = 0; j < N_TEACHERS_PER_CLASS; ++j)
            {
               if (teachers.size() == 0)
               {
                  break;
               }

               /* Select random teacher from list */
               //unsigned long assigned_teacher_idx = rand() % teachers.size();
               //auto me = agents.at( teachers.at( assigned_teacher_idx ) );

               // CZ: add deterministically
               auto me = agents.at(teachers.front());
               //  std::cout << "teacher: " << me->getID() << std::endl;
               /* Assign me to grade and class */
               my_grade->addAgentIdx(me->getID());
               my_class->addAgentIdx(me->getID());

               /* Assign grade and class to me */
               me->addWG(my_grade);
               me->addWG(my_class);

               /* Remove teacher from list */
               //auto start_it = teachers.begin();
               //teachers.erase( start_it + assigned_teacher_idx );

               // CZ:
               teachers.erase(teachers.begin());
            }

            num_unassigned_teachers = teachers.size();
         }
      }

      school_student_stream << std::endl;
      school_grade_stream << std::endl;
   }
   school_student_stream.close();
   school_grade_stream.close();

   //  std::cout << "CZ: number of unassigned teachers = " << num_unassigned_teachers << std::endl;
   //std::cout << "Press enter to continue ...";
   //    std::cin.get();
}

void CreatePopulation::readPopulation(agent_map_type& agents, hh_map_type& households, hc_map_type& household_clusters,
                                      cd_map_type& census_districts, dz_map_type& destination_zones, sla_map_type& statistical_local_areas,
                                      wg_map_type& working_groups, wg_map_type& schools, wg_map_type& grades, wg_map_type& classes, const std::string& database_input, const std::string& population_datafile)
{
   mapCD_SLA_DZ(census_districts, destination_zones, statistical_local_areas, database_input);
   mapWG_POW(working_groups, schools, destination_zones, database_input);

   long unsigned    in_CD;        // census district
   long long signed in_WG;        // working group
   long unsigned    in_HC, in_HH; // household and cluster
   short            in_AG;        // age group

   std::ifstream pop_file(population_datafile);

   std::cout << "Creating agents..." << std::endl;

   for (unsigned p_idx = 0;
        pop_file >> in_AG >> in_HH >> in_HC >> in_WG >> in_CD;
        ++p_idx)
   {
      agents.emplace(p_idx, std::make_shared <Agent>(p_idx, (AgeGroup)(in_AG)));

      /* Check if we need to initialise a new cluster */
      auto c_hc = household_clusters.find(in_HC);
      if (c_hc == household_clusters.end())
      {
         /* Fill in household cluster info */
         auto ccd = census_districts.find(in_CD);
         if (ccd != census_districts.end())
         {
            household_clusters.emplace(in_HC, std::make_shared <HouseholdCluster>(in_HC, ccd->second));
            c_hc = household_clusters.find(in_HC);
         }
         else
         {
            std::cerr << "Unknown census collection district: " << in_CD << std::endl;
            exit(1);
         }
      }

      /* Check if we need to initialise a new household and assign it to cluster */
      auto c_hh = households.find(in_HH);
      if (c_hh == households.end())
      {
         households.emplace(in_HH, std::make_shared <Household>(in_HH, household_clusters.at(in_HC)));
         c_hh = households.find(in_HH);
      }

      /* Assign working group */
      if (in_WG > 0)
      {
         auto cwg = working_groups.find(in_WG);
         if (cwg != working_groups.end())
         {
            agents.at(p_idx)->addWG(cwg->second);
         }
         else
         {
            std::cerr << "Unknown place of work: " << in_WG << std::endl;
            exit(1);
         }
      }
      else if (in_WG < 0)
      {
         auto cschool = schools.find(-in_WG);
         if (cschool != schools.end())
         {
            agents.at(p_idx)->addWG(cschool->second);
         }
         else
         {
            std::cerr << "Unknown school: " << -in_WG << " (out of " << schools.size() << " schools)." << std::endl;
            exit(1);
         }
      }

      /* Assigned household cluster */
      agents.at(p_idx)->setHH(c_hh->second);

      /* Add agent index to the household and LOWEST working group (we recursively add to everything in the hierarchy) */
      agents.at(p_idx)->getHH()->addAgentIdx(p_idx);
      if (agents.at(p_idx)->getWGs().size() > 0)
      {
         agents.at(p_idx)->getWGs().front()->addAgentIdx(p_idx);
      }

      if (p_idx % 1000000 == 0)
      {
         std::cout << "\rCreated agent: " << agents.at(p_idx)->getID() << std::flush;
      }
   }
   std::cout << std::endl;

   pop_file.close();

   addGradesAndClasses(agents, schools, grades, classes);

   for (auto csla : statistical_local_areas)
   {
      csla.second->computeDensity();
   }

   for (auto ccd : census_districts)
   {
      ccd.second->computeDensity();
   }

   for (auto cwg : working_groups)
   {
      cwg.second->computeDensity();
   }

   for (auto csch : schools)
   {
      csch.second->computeDensity();
   }

   for (auto cgr : grades)
   {
      cgr.second->computeDensity();
   }

   for (auto ccl : classes)
   {
      ccl.second->computeDensity();
   }

   std::cout << "Created: " << working_groups.size() << " Working Groups." << std::endl;
   std::cout << "Created: " << schools.size() << " Schools." << std::endl;
   std::cout << "Created: " << grades.size() << " Grades." << std::endl;
   std::cout << "Created: " << classes.size() << " Classes." << std::endl;
   std::cout << "Created: " << households.size() << " Households." << std::endl;
   std::cout << "Created: " << household_clusters.size() << " Household Clusters." << std::endl;
   std::cout << "Created: " << agents.size() << " Agents." << std::endl;
}
