/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright   Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-201), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright   Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   This code is released only for internal review by Nature Communications, manuscript: NCOMMS-21-29568.

   *EndCopyright:
*/

#include "epidemic.h"

#include <fstream>
#include <sstream>
#include <string>
#include <random>
#include <chrono>

std::vector <std::pair <unsigned long, std::vector <unsigned long> > > Epidemic::n_commuters_to_airport_slas_;
std::vector <std::pair <unsigned long, std::vector <unsigned long> > > Epidemic::n_commuters_to_airport_agents_;
std::vector <unsigned long> Epidemic::agent_list_;

const double Epidemic::expected_number_infected_ = 0.00004;//0.00004;

std::string Epidemic::seeded_sla_fname { "SeedSLAs.dat" };

void Epidemic::init_transmission_rates(double p_trans, double scaler)
{
   /* Initialise all transmission rates for all statistical areas */
   WorkingGroup::addTransmissionRate(scaler * p_trans * contact_rate_wg);
   for (auto c_tr : transmission_rate_schools)
   {
      double p = 1 - pow(1 - c_tr, 1.0 / 4.1);
      WorkingGroup::addTransmissionRate(scaler * p);
   }

   for (auto c_rate : contact_rate_sla)
   {
      StatisticalLocalArea::addTransmissionRate(scaler * p_trans * c_rate);
   }

   for (auto c_rate : contact_rate_cd)
   {
      CensusDistrict::addTransmissionRate(scaler * p_trans * c_rate);
   }

   for (unsigned i = 0; i < contact_rate_hc.size(); ++i)
   {
      auto contact_rate_child_or_adult = contact_rate_hc.at(i);
      for (unsigned j = 0; j < contact_rate_child_or_adult.size(); ++j)
      {
         /* Add transmission rates from child (0) or adult (1) */
         HouseholdCluster::addTransmissionRate(i, scaler * p_trans * contact_rate_child_or_adult.at(j));
      }
   }

   for (unsigned i = 0; i < transmission_rate_hh.size(); ++i)
   {
      auto c_hh_size = transmission_rate_hh.at(i);
      for (unsigned j = 0; j < c_hh_size.size(); ++j)
      {
         double p = 1 - pow(1 - c_hh_size.at(j), 1.0 / 4.1);

         /* Add transmission rate for household size i+1 */
         Household::addTransmissionRate(i, scaler * p);
      }
   }

#if 0
   /* OLIVER: Withdrawal rates? Not implemented */
   for (unsigned i = 0; i < 3; ++i)
   {
      for (unsigned j = 0; j < 3; j++)
      {
         WDpro[i][j] *= WDrate[i]; //WDpro and WDrate stand for?
      }
   }
#endif
}

void Epidemic::init_random_population_seed(const agent_map_type& agents)
{
   for (auto& cagent : agents)
   {
      agent_list_.push_back(cagent.first);
   }
}

void Epidemic::init_airport_seeded_slas(const std::string& database, const sla_map_type& sla_map)
{
   /* Open the file for reading */
   std::string seed_file(database + seeded_sla_fname);

   std::ifstream seed_stream(seed_file);

   std::string cline;
   unsigned    ap = 0;

   while (std::getline(seed_stream, cline))
   {
      std::stringstream i_ss;
      i_ss << cline;

      unsigned incoming_passengers;
      i_ss >> incoming_passengers;
      n_commuters_to_airport_slas_.emplace_back(incoming_passengers, std::vector <unsigned long>());
      n_commuters_to_airport_agents_.emplace_back(incoming_passengers, std::vector <unsigned long>());

      unsigned long c_sla;

      while (i_ss >> c_sla)
      {
         n_commuters_to_airport_slas_.back().second.push_back(c_sla);

         for (auto& agent_idx : sla_map.at(c_sla)->getAgentIdxs())
         {
            n_commuters_to_airport_agents_.back().second.push_back(agent_idx);
         }
      }
   }
   seed_stream.close();
}

unsigned Epidemic::get_one_airport_agent_idx(unsigned airport, agent_map_type& agents)
{
   auto com_to_ap = n_commuters_to_airport_agents_.begin();

   std::advance(com_to_ap, airport);

   /* Choose one agent in SLAs surrounding this airport */
   auto&         potential_agent_idxs = com_to_ap->second;
   unsigned int  rval             = rand() % potential_agent_idxs.size();
   unsigned long chosen_agent_idx = potential_agent_idxs.at(rval);

   std::cout << "Chosen agent " << chosen_agent_idx << " (out of "
             << potential_agent_idxs.size() << " options) in airport " << airport << std::endl;
   return(chosen_agent_idx);
}

void Epidemic::seed_epi_airports(agent_map_type& agents, std::vector <unsigned long>& infected_ids, double& AVEd, double& VEd_1, double& VEd_2)
{
   std::default_random_engine generator(rand());        /* Kit's edit 06.07.18 - random seed */
   unsigned n_infected = 0;

   for (auto& com_to_agents : n_commuters_to_airport_agents_)
   {
      /* Draw number to infect from binomial */
      unsigned int max_to_infect = com_to_agents.first;
      std::binomial_distribution <int> bd(max_to_infect, expected_number_infected_);
      unsigned int n_to_infect = bd(generator);
      //std::cout << "Infecting " << n_to_infect << " agents in SLAs with " << max_to_infect << " potentials." << std::endl;

      /* Infect random agents surrounding this airport SLA */
      auto& potential_agent_idxs = com_to_agents.second;
      for (unsigned i = 0; i < n_to_infect; ++i)
      {
         unsigned int  rval             = rand() % potential_agent_idxs.size();
         unsigned long chosen_agent_idx = potential_agent_idxs.at(rval);
         if (!agents.at(chosen_agent_idx)->isInfected())
         {
            agents.at(chosen_agent_idx)->infect(AVEd, VEd_1, VEd_2);
            infected_ids.push_back(chosen_agent_idx);
            n_infected++;
         }
      }
   }
}

void Epidemic::seed_epi_sla(unsigned n_to_infect,
                            StatisticalLocalArea& sla_to_infect,
                            agent_map_type& agents, double& AVEd, double& VEd_1, double& VEd_2)
{
   /* choose random agents in that SLA to be infected */
   std::vector <unsigned long> valid_agent_idx(sla_to_infect.getAgentIdxs());
   for (unsigned i = 0; i < n_to_infect; ++i)
   {
      bool infected = false;
      while (!infected)
      {
         if (valid_agent_idx.size() < n_to_infect - i)
         {
            std::cerr << "Not enough uninfected agents SLA:" << sla_to_infect.getID()
                      << " (originally " << sla_to_infect.getAgentIdxs().size()
                      << " agents, and infected " << i << " so far)" << std::endl;
            exit(1);
         }

         unsigned long rval             = rand() % valid_agent_idx.size();
         unsigned long chosen_agent_idx = valid_agent_idx.at(rval);

         /* If we've already infected this agent, continue searching, else infect and move on */
         if (!agents.at(chosen_agent_idx)->isInfected())
         {
            agents.at(chosen_agent_idx)->infect(AVEd, VEd_1, VEd_2);
            infected = true;
         }
         valid_agent_idx.erase(valid_agent_idx.begin() + rval);
      }
   }
}

void Epidemic::seed_epi_agents(double p_to_infect,
                               agent_map_type& agents, double& AVEd, double& VEd_1, double& VEd_2)
{
   /* choose random agents in that SLA to be infected */
   std::vector <unsigned long> agent_list(agent_list_);

   if (p_to_infect > 1.0)
   {
      std::cerr << "Proportion of agents to infect (" << p_to_infect << ")"
                << ") greater than 100%%. Exiting." << std::endl;
      exit(1);
   }

   unsigned long n_to_infect = (unsigned long)(double(agent_list.size()) * p_to_infect);
   std::cout << "Infecting " << n_to_infect << " agents randomly." << std::endl;

   for (unsigned i = 0; i < n_to_infect; ++i)
   {
      unsigned long chosen_agent_idx;
      do
      {
         unsigned long rval = rand() % agent_list.size();
         chosen_agent_idx = agent_list.at(rval);
      } while (agents.at(chosen_agent_idx)->isInfected());

      agents.at(chosen_agent_idx)->infect(AVEd, VEd_1, VEd_2);
   }
   std::cout << "Done." << std::endl;
}
