/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright   Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-201), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright   Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   This code is released only for internal review by Nature Communications, manuscript: NCOMMS-21-29568.

   *EndCopyright:
*/

#include <limits.h>
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <vector>
#include <stdarg.h>
#include <math.h>
#include <string>
#include <algorithm>
#include <map>
#include <sstream>

#include "debug.h"
#include "utils.hpp"
#include "generate_population.h"
#include "enums.h"

void generatePopulation(std::vector <std::shared_ptr <SimpleAgent> >&);
void printPopulation(const std::vector <std::shared_ptr <SimpleAgent> >&, const std::string&);

std::shared_ptr <SimpleAgent> generateAgent(unsigned long, cd_type, unsigned long, unsigned long, const std::vector <std::pair <AgeGroup, double> > &);

int main(int argc, char *argv[])
{
   srand(time(NULL));

   std::vector <std::shared_ptr <SimpleAgent> > agents;
   const std::string fname = database_input + "population.dat";
   std::cout << "Initiating program." << std::endl;
   generatePopulation(agents);

   printPopulation(agents, fname);
}

void generatePopulation(std::vector <std::shared_ptr <SimpleAgent> >& agents)
{
   std::cout << "Database: " << database_input << std::endl;
   std::string fname(database_input + "TravelToWork.dat");
   std::cout << "Preprocessing " << fname << " input file..." << std::endl;

   std::ifstream TTWFile(fname);
   if (!TTWFile.good())
   {
      std::cerr << "File " << fname << " does not exist." << std::endl;
   }

   /* Generate WG.dat file by counting how many people are in each DZ and dividing by 10 */

   /* map from DZ to WG ID to WG structure (holding number in group and number assigned to group) */
   std::map <dz_type, unsigned long>  dz_numAgents_map;
   std::map <sla_type, unsigned long> sla_numStudents_map;
   std::map <dz_type, std::vector <unsigned long> > dz_to_adult_idx;

   cd_dz_nagents_type cd_dz_nagents_map;

   unsigned long in_DZ, in_CD;
   unsigned int  in_numAgents;
   while (TTWFile >> in_CD >> in_DZ >> in_numAgents)
   {
      /* Initialise this entry if needed destination zone map */
      if (dz_numAgents_map.find(in_DZ) == dz_numAgents_map.end())
      {
         dz_numAgents_map[in_DZ] = 0;
      }

      dz_numAgents_map[in_DZ] += in_numAgents;
      cd_dz_nagents_map[in_CD].push_back(std::pair <dz_type, unsigned long>(in_DZ, in_numAgents));
   }
   TTWFile.close();

   /* Map all CDs to an SLA */
   std::map <cd_type, sla_type> cd_sla_map;
   std::ifstream cd_sla(database_input + "CDinSLA.dat");
   unsigned long in_SLA;
   double         in_lat, in_long;
   while (cd_sla >> in_CD >> in_SLA >> in_lat >> in_long)
   {
      cd_sla_map[in_CD]           = in_SLA;
      sla_numStudents_map[in_SLA] = 0;
   }
   cd_sla.close();

   /* How many working groups total? */
   unsigned long cumNumWGs = 0, numWGs;
   std::string   wg_fname  = database_input + "WG.dat";
   std::ofstream WGFile(wg_fname);
   std::map <dz_type, struct WGLimitStruct> dz_wglim_map;
   struct WGLimitStruct wg_lim;
   for (auto dz_to_numWg : dz_numAgents_map)
   {
      numWGs = ceil((double)dz_to_numWg.second / (double)NUM_PEOPLE_IN_WG);
      WGFile << dz_to_numWg.first << " " << cumNumWGs << " " << numWGs << std::endl;
      wg_lim.lower = cumNumWGs;
      cumNumWGs   += numWGs;
      wg_lim.upper = cumNumWGs;
      dz_wglim_map[dz_to_numWg.first] = wg_lim;
   }
   WGFile.close();

   std::cout << "Created " << cumNumWGs << " working groups in " << wg_fname << "." << std::endl;

   /* Create all working groups */
   int HH_count;

   std::cout << "Pre-loading family, age, house, school datafiles..." << std::endl;

   unsigned long min_CD = ULONG_MAX;
   unsigned long max_CD = 0;
   std::string   cline;

   std::ifstream hh_comp_fs(database_input + "HOUSEHOLD_COMPOSITION.dat");
   std::map <cd_type, std::vector <double> > hh_comp_map;
   {
      /* pre-load maps of CDs to data */
      double hh_comp_dist;
      while (std::getline(hh_comp_fs, cline))
      {
         std::istringstream iss(cline);
         iss >> in_CD;
         hh_comp_map.emplace(in_CD, std::vector <double>());
         while (iss >> hh_comp_dist)
         {
            hh_comp_map.at(in_CD).push_back(hh_comp_dist);
         }
      }
   }
   hh_comp_fs.close();

   std::cout << "Household composition map size: " << hh_comp_map.size() << std::endl;

   auto hh_comp_it = hh_comp_map.begin();

   std::ifstream age_group_fs(database_input + "AGE.dat");
   std::map <cd_type, std::pair <unsigned long, std::vector <double> > > age_map;
   {
      /* Load all age distributions */
      double         age_dist;
      unsigned long cd_size;
      while (std::getline(age_group_fs, cline))
      {
         std::istringstream iss(cline);
         iss >> in_CD >> cd_size;
         std::vector <double> dist;
         while (iss >> age_dist)
         {
            dist.push_back(age_dist);
         }
         age_map.emplace(in_CD, std::pair <unsigned long, std::vector <double> >(cd_size, dist));
      }
   }
   age_group_fs.close();

   std::cout << "Agegroup map size: " << age_map.size() << std::endl;

   std::vector <unsigned> students;

   typedef std::vector <std::pair <AgeGroup, double> >   agegroup_cdf_type;

   /* Generate the distributions */
   unsigned long n_CDs           = 0;
   unsigned long agent_id        = 0, hh_id = 0, hc_id = 0;
   unsigned long cum_agents      = 0, n_adults = 0, n_children = 0;
   unsigned long n_agents_age[5] = { 0 };
   for (auto age_it : age_map)
   {
      unsigned long c_CD = age_it.first;

      /* Check this CD is known to all data (census) files */
      if (cd_sla_map.find(c_CD) == cd_sla_map.end())
      {
         std::cerr << std::endl << "CD " << c_CD << " not found in CDinSLA.dat." << std::endl;
         exit(1);
      }
      sla_type c_SLA = cd_sla_map.at(c_CD);

      if ((hh_comp_it = hh_comp_map.find(c_CD)) == hh_comp_map.end())
      {
         std::cerr << std::endl << "CD " << c_CD << " not found in HOUSEHOLD_COMPOSITION.dat" << std::endl;
         continue;
      }

      unsigned long c_cd_size     = age_it.second.first;
      auto&         agegroup_dist = age_it.second.second;

      auto& hh_comp_dist = hh_comp_it->second;

      std::cout << "\rCreating CD:" << c_CD;
      //std::cout << "Creating CD:" << c_CD << std::endl;

      auto n_adult_males   = agegroup_dist[SexAgeGroups::M19to29] + agegroup_dist[SexAgeGroups::M30to64] + agegroup_dist[SexAgeGroups::M65Plus];
      auto n_adult_females = agegroup_dist[SexAgeGroups::F19to29] + agegroup_dist[SexAgeGroups::F30to64] + agegroup_dist[SexAgeGroups::F65Plus];

      auto n_child_males   = agegroup_dist[SexAgeGroups::M0to4] + agegroup_dist[SexAgeGroups::M5to18];
      auto n_child_females = agegroup_dist[SexAgeGroups::F0to4] + agegroup_dist[SexAgeGroups::F5to18];

      /*
       * Generate the CDFs for this census collection district.
       * generateCDF takes in the domain (first argument) and distribution (second argument) and returns a vector
       * to be used with sampleCDF (which returns a value from the domain).
       */
      auto hh_comp_cdf = generateCDF(hh_comps, hh_comp_dist);

      auto adult_sex_cdf = generateCDF(sexes, std::vector <double> { n_adult_males, n_adult_females });

      auto child_sex_cdf = generateCDF(sexes, std::vector <double> { n_child_males, n_child_females });

      agegroup_cdf_type adult_male_agegroup_cdf = generateCDF(adult_agegroups,
                                                              std::vector <double> { agegroup_dist[SexAgeGroups::M19to29],
                                                                                    agegroup_dist[SexAgeGroups::M30to64],
                                                                                    agegroup_dist[SexAgeGroups::M65Plus] });
      agegroup_cdf_type adult_female_agegroup_cdf = generateCDF(adult_agegroups,
                                                                std::vector <double> { agegroup_dist[SexAgeGroups::F19to29],
                                                                                      agegroup_dist[SexAgeGroups::F30to64],
                                                                                      agegroup_dist[SexAgeGroups::F65Plus] });

      agegroup_cdf_type child_male_agegroup_cdf = generateCDF(child_agegroups,
                                                              std::vector <double> { agegroup_dist[SexAgeGroups::M0to4],
                                                                                    agegroup_dist[SexAgeGroups::M5to18] });
      agegroup_cdf_type child_female_agegroup_cdf = generateCDF(child_agegroups,
                                                                std::vector <double> { agegroup_dist[SexAgeGroups::F0to4],
                                                                                      agegroup_dist[SexAgeGroups::F5to18] });

      agegroup_cdf_type adult_at_parents_agegroup_cdf = generateCDF(std::vector <AgeGroup> { AgeGroups::Age19to29 },
                                                                    std::vector <double> { 1.0 });

      unsigned long cum_agents_in_cd = 0;
      unsigned long n_hh_in_cluster  = 0;
      while (cum_agents_in_cd < c_cd_size)
      {
         /* Draw the current household composition */

         /*
          * TODO: make sure composition size is less than/equal to # agents left in CD
          */
         HouseholdComposition c_hh_comp = HouseholdComposition(sampleCDF(hh_comp_cdf));
         //hh_comp_drawn[c_hh_comp]++;

         /* Vector of CDFs that we need to sample to get age groups of the agents */
         std::vector <agegroup_cdf_type> agegroup_cdfs;
         unsigned long n_adults_to_sample   = 0;
         unsigned long n_children_to_sample = 0;

         /* Single/lone household compositions */
         if (c_hh_comp <= HouseholdCompositions::Single)
         {
            n_adults_to_sample = 1;
         }
         /* Couple with/without children household compositions */
         else if (c_hh_comp <= HouseholdCompositions::CouplesWithFourChildren)
         {
            agegroup_cdfs.push_back(adult_male_agegroup_cdf);
            agegroup_cdfs.push_back(adult_female_agegroup_cdf);
            n_adults += 2;

            /* Couple with children household compositions */
            if (c_hh_comp >= HouseholdCompositions::CouplesWithOneChild)
            {
               n_children_to_sample = c_hh_comp - HouseholdCompositions::CouplesWithOneChild + 1;
            }
         }
         /* Single parent family household compositions */
         else if (c_hh_comp <= HouseholdCompositions::SingleParentFamilyFiveChildren)
         {
            n_adults_to_sample   = 1;
            n_children_to_sample = c_hh_comp - HouseholdCompositions::SingleParentFamilyOneChild + 1;
         }
         /* Group household compositions */
         else
         {
            n_adults_to_sample = c_hh_comp - HouseholdCompositions::GroupTwo + 2;
         }

         unsigned n_adults_at_parents = 0;
         for (unsigned child_id = 0; child_id < n_children_to_sample; ++child_id)
         {
            /* Are these "children" actually adults living at home? */
            if (rand() / double( RAND_MAX ) <= PERC_ADULTS_AT_HOME)
            {
               agegroup_cdfs.push_back(adult_at_parents_agegroup_cdf);
               n_adults_at_parents++;
            }
            /* Else add the children (<18) age group CDFs for sampling */
            else if (sampleCDF(child_sex_cdf) == Sex::Male)
            {
               agegroup_cdfs.push_back(child_male_agegroup_cdf);
            }
            else
            {
               agegroup_cdfs.push_back(child_female_agegroup_cdf);
            }
         }
         n_children += n_children_to_sample - n_adults_at_parents;

         for (unsigned adult_id = 0; adult_id < n_adults_to_sample; ++adult_id)
         {
            if (sampleCDF(adult_sex_cdf) == Sex::Male)
            {
               agegroup_cdfs.push_back(adult_male_agegroup_cdf);
            }
            else
            {
               agegroup_cdfs.push_back(adult_female_agegroup_cdf);
            }
         }
         n_adults += n_adults_to_sample + n_adults_at_parents;

         for (auto c_cdf : agegroup_cdfs)
         {
            auto cagent = generateAgent(agent_id++, c_CD, hh_id, hc_id, c_cdf);
            agents.push_back(cagent);
            n_agents_age[cagent->my_age]++;
         }

         hh_id++;
         /* Ensure we don't have clusters going between CDs */
         if (++n_hh_in_cluster == 4)
         {
            hc_id++;
            n_hh_in_cluster = 0;
         }

         cum_agents_in_cd += agegroup_cdfs.size();
      }
      hc_id++;

      cd_dz_nagents_type::iterator cd_dz_nagents_it;

      std::map <dz_type, std::map <wg_type, unsigned long> > dz_wg_num_map;

      /* If there are actually adults commuting from this CD to any DZ... */
      if ((cd_dz_nagents_it = cd_dz_nagents_map.find(c_CD)) != cd_dz_nagents_map.end())
      {
         /* ...pre-select all adults in area. */
         std::vector <unsigned long> adult_idx_in_dz;
         for (unsigned i = 0; i < cum_agents_in_cd; ++i)
         {
            unsigned long agent_id = cum_agents + i;
            if (agents[agent_id]->isAdult())
            {
               /* 35% of agents aged 65+ work */
               if (agents[agent_id]->my_age < AgeGroups::Age65Plus ||
                   rand() / double(RAND_MAX) < 0.35)
               {
                  adult_idx_in_dz.push_back(agent_id);
               }
            }
            else if (agents[agent_id]->isStudentAge())
            {
               if (rand() / double(RAND_MAX) < SCHOOL_ATTENDANCE_LIKELIHOOD)
               {
                  students.push_back(agent_id);
                  if (sla_numStudents_map.find(c_SLA) == sla_numStudents_map.end())
                  {
                     sla_numStudents_map[c_SLA] = 0;
                  }
                  sla_numStudents_map.at(c_SLA)++;
               }
            }
         }

         /* Iterate through each DZ this CD can go to. */
         for (auto dz_to_nagents_it : cd_dz_nagents_it->second)
         {
            dz_type       c_DZ = dz_to_nagents_it.first;
            unsigned long n_commuting_agents = dz_to_nagents_it.second;

            if (dz_to_adult_idx.find(c_DZ) == dz_to_adult_idx.end())
            {
               dz_to_adult_idx[c_DZ] = std::vector <unsigned long>();
            }

            /* Gives which working groups we can select from */
            struct WGLimitStruct wg_lims  = dz_wglim_map.at(c_DZ);
            unsigned             wg_range = wg_lims.upper - wg_lims.lower;

            if (dz_wg_num_map.find(c_DZ) == dz_wg_num_map.end())
            {
               dz_wg_num_map[c_DZ] = std::map <wg_type, unsigned long>();
            }

            /*
             * Now, randomly assign each agent from list to a working group
             ***(don't care how many are in each group because it should be E{18} due to number of groups)
             */
            for (unsigned i = 0; i < n_commuting_agents; ++i)
            {
               if (adult_idx_in_dz.size() == 0)
               {
                  std::cerr << std::endl << "Number of commuting agents (" << n_commuting_agents - i
                            << " more commuters than working adults in DZ:" << c_DZ << ")." << std::endl;
                  break;
               }
               unsigned long selectedAdult = rand() % adult_idx_in_dz.size();
               wg_type       selectedWG    = rand() % wg_range + wg_lims.lower;

               if (selectedWG > wg_lims.upper)
               {
                  std::cerr << "err" << std::endl;
               }

               dz_to_adult_idx[c_DZ].push_back(adult_idx_in_dz.at(selectedAdult));

               agents[adult_idx_in_dz.at(selectedAdult)]->my_wg = selectedWG;
               adult_idx_in_dz.erase(adult_idx_in_dz.begin() + selectedAdult);

               if (dz_wg_num_map.at(c_DZ).find(c_DZ) == dz_wg_num_map.at(c_DZ).end())
               {
                  dz_wg_num_map.at(c_DZ)[selectedWG] = 0;
               }
               dz_wg_num_map.at(c_DZ)[selectedWG]++;
            }
         }
      }

      cum_agents += cum_agents_in_cd;
   }

   std::cout << std::endl << "Created " << agents.size() << " agents"
             << " comprising " << n_adults << " adults and " << n_children << " children." << std::endl;

   std::cout << "Agent in age groups: ";
   for (unsigned i = 0; i < 5; ++i)
   {
      std::cout << n_agents_age[i] << " ";
   }
   std::cout << std::endl;

   /*
    * Make schools and assign teachers and students
    */

   /* First, list all DZs in each SLA */
   unsigned long sla, dz;
   std::map <dz_type, sla_type> dz_sla_map;
   std::map <sla_type, std::vector <struct School *> > sla_school_map;
   std::map <sla_type, std::vector <struct School *> > sla_schoolBin_map;
   std::ifstream sla_dz(database_input + "DZinSLA.dat");
   while (sla_dz >> sla >> dz)
   {
      sla_school_map[sla]    = std::vector <struct School *>();
      sla_schoolBin_map[sla] = std::vector <struct School *>();
      dz_sla_map[dz]         = sla;
      //std::cout << "DZ:" << dz << " in SLA:" << sla << std::endl;
   }
   sla_dz.close();

   std::cout << "Number of SLAs with DZs: " << sla_school_map.size() << std::endl;

   /* For each state... */
   unsigned short state;
   unsigned       numSchools[SCHOOL_BOUNDS.size()] = { 0 };
   std::ifstream  school(database_input + "SCHOOL.dat");

   std::map <dz_type, sla_type>::iterator       dz_sla_it_end, dz_sla_it = dz_sla_map.begin();
   std::map <sla_type, unsigned long>::iterator sla_numStudents_it_end, sla_numStudents_it = sla_numStudents_map.begin();

   unsigned long cum_schools = 1;
   std::map <dz_type, unsigned> dz_numTeachers_map;

   /* Write which DZ schools are in */
   std::ofstream school_in_dz(database_input + "SCHOOLinDZ.dat");

   std::map <sla_type, unsigned long> sla_numStudentsFilled_map;

   unsigned long global_school_idx = 1;

   /* For generating the CDFs */
   std::map <sla_type, std::vector <unsigned long> > sla_school_id_map;
   std::map <sla_type, std::vector <double> >         sla_school_dist_map;

   std::map <unsigned long, struct School *> school_id_map;

   while (std::getline(school, cline))
   {
      unsigned long sum_max_students = 0;

      std::istringstream iss(cline);
      iss >> state;
      state--;
      for (unsigned i = 1; i < SCHOOL_BOUNDS.size(); ++i)
      {
         iss >> numSchools[i];
         sum_max_students += SCHOOL_BOUNDS.at(i) * numSchools[i];
      }

      if (state > STATE_STRINGS.size())
      {
         std::cerr << "SCHOOL.dat file has more rows (" << state << ") than known states (" << STATE_STRINGS.size() << ")" << std::endl;
      }
      std::cout << "Filling schools in state " << STATE_STRINGS.at(state) << std::endl;

      /* Find all DZs in this state (TODO use DZinSLA data)... */
      if ((dz_sla_it_end = dz_sla_map.lower_bound((state + 2) * DZ_MULTIPLIER)) == dz_sla_map.end())
      {
         if (state + 1 < STATE_STRINGS.size())
         {
            std::cerr << "Could not find DZ lower bound: " << (state + 2) * DZ_MULTIPLIER << std::endl;
         }
      }
      /* ...and make a list of available DZs (and number of agents working there) to put schools in */
      std::vector <dz_type> dzs_in_state;
      for (; dz_sla_it != dz_sla_it_end; ++dz_sla_it)
      {
         dzs_in_state.push_back(dz_sla_it->first);
         dz_numTeachers_map[dz_sla_it->first] = 0;
      }
      std::cout << "Added " << dzs_in_state.size() << " DZs for state " << STATE_STRINGS.at(state) << std::endl;

      /* Find all SLAs in this state... */
      if ((sla_numStudents_it_end = sla_numStudents_map.lower_bound((state + 2) * SLA_MULTIPLIER)) == sla_numStudents_map.end())
      {
         if (state + 1 < STATE_STRINGS.size())
         {
            std::cerr << "Could not find SLA lower bound: " << (state + 2) * SLA_MULTIPLIER << std::endl;
         }
      }

      std::cout << "Found SLA:" << sla_numStudents_it_end->first << " as end of state " << STATE_STRINGS.at(state) << std::endl;

      /* ...and make a list of available SLAs (and number of students living there) to put schools in */
      for (; sla_numStudents_it != sla_numStudents_it_end; ++sla_numStudents_it)
      {
         sla_numStudentsFilled_map[sla_numStudents_it->first] = 0;
      }
      std::cout << "Added " << sla_numStudentsFilled_map.size() << " SLAs total." << std::endl;

      bool state_full = false;

      /* ...iterate through each school size upper bound in that state... */
      for (unsigned bound_idx = 1; bound_idx < SCHOOL_BOUNDS.size() && ~state_full; ++bound_idx)
      {
         unsigned bound_range = SCHOOL_BOUNDS.at(bound_idx) - SCHOOL_BOUNDS.at(bound_idx - 1);

         std::cout << "There are " << numSchools[bound_idx] << " schools in bound "
                   << SCHOOL_BOUNDS.at(bound_idx - 1) << " - " << SCHOOL_BOUNDS.at(bound_idx)
                   << "(range: " << bound_range << "; num of bounds: " << bound_range / SCHOOL_STEP_SIZE << ")." << std::endl;

         unsigned remaining_schools = numSchools[bound_idx];

         while (remaining_schools > 0)
         {
            unsigned num_bins = std::min(remaining_schools, bound_range);

            unsigned num_schools_per_bin = remaining_schools / num_bins;

            for (unsigned bin_idx = 1; bin_idx < num_bins + 1; ++bin_idx)
            {
               auto cschool_bin_sz         = (bin_idx * bound_range) / (num_bins + 1) + 1;
               auto max_students_in_school = SCHOOL_BOUNDS.at(bound_idx - 1) + cschool_bin_sz;
               auto max_students_scaled    = SCHOOL_BOUNDS.at(bound_idx - 1) * MAX_STUDENT_SCALER + cschool_bin_sz;

               unsigned num_teachers_required = ceil(max_students_in_school / NUM_STUDENTS_TO_TEACHERS);

               unsigned num_schools_in_bound = remaining_schools / num_bins;

               /*
                * std::cout << "Assigning " << num_schools_in_bound << " schools of size " << max_students_in_school
                *                      << "(" << num_teachers_required << " teachers)." << std::endl;
                */

               /* ... and all schools in that upper bound. */
               unsigned long local_school_idx;
               for (local_school_idx = 0; local_school_idx < num_schools_in_bound; ++local_school_idx)
               {
                  unsigned num_working_adults = 0, num_students_left_in_sla = 0;
                  dz_type  selectedDZ;
                  sla_type selectedSLA;

                  while (num_working_adults < num_teachers_required ||
                         num_students_left_in_sla <= max_students_scaled)
                  {
                     if (dzs_in_state.size() == 0)
                     {
                        std::cerr << "No more DZs left in state!" << std::endl;
                        std::cout << "Still have " << num_schools_in_bound - local_school_idx
                                  << " schools left in bound " << max_students_in_school << " to go." << std::endl;
                        state_full = true;
                        break;
                     }

                     /* Randomly select a DZ in that state */
                     unsigned rval = rand() % dzs_in_state.size();
                     selectedDZ = dzs_in_state.at(rval);

                     /* Check DZ has enough available teachers for this school size */
                     num_working_adults = 0;
                     if (dz_numAgents_map.find(selectedDZ) != dz_numAgents_map.end())
                     {
                        num_working_adults = dz_numAgents_map.at(selectedDZ) - dz_numTeachers_map.at(selectedDZ);
                     }

                     selectedSLA = dz_sla_map.at(selectedDZ);

                     /* Check SLA has the right number of students for this school size */
                     if (sla_numStudents_map.find(selectedSLA) != sla_numStudents_map.end())
                     {
                        if (sla_numStudentsFilled_map.find(selectedSLA) == sla_numStudentsFilled_map.end())
                        {
                           sla_numStudentsFilled_map[selectedSLA] = 0;
                        }
                        num_students_left_in_sla = sla_numStudents_map.at(selectedSLA) - sla_numStudentsFilled_map.at(selectedSLA);
                     }

                     if (num_students_left_in_sla <= max_students_scaled ||
                         num_working_adults < num_teachers_required)
                     {
                        dzs_in_state.erase(dzs_in_state.begin() + rval);
                     }
                  }

                  if (state_full)
                  {
                     break;
                  }

                  if (dz_to_adult_idx.find(selectedDZ) == dz_to_adult_idx.end())
                  {
                     std::cerr << "Somehow DZ:" << selectedDZ << " in numAgents map but not in agentIdx map!" << std::endl;
                  }

                  /* Assign school to DZ */
                  sla_type       sla     = dz_sla_map.at(selectedDZ);
                  struct School *cschool = new struct School;
                  cschool->ID           = local_school_idx + global_school_idx;
                  cschool->max_students = max_students_in_school;
                  cschool->num_students = 0;
                  cschool->SLA          = sla;
                  cschool->DZ           = selectedDZ;

                  school_id_map[cschool->ID] = cschool;
                  sla_school_map[sla].push_back(cschool);

                  school_in_dz << cschool->ID << " " << cschool->DZ << " " << cschool->max_students << std::endl;

                  /* Add teachers to DZ */
                  dz_numTeachers_map[selectedDZ] += num_teachers_required;

                  sla_numStudentsFilled_map[selectedSLA] += max_students_scaled;
                  if (sla_numStudentsFilled_map[selectedSLA] > sla_numStudents_map.at(selectedSLA))
                  {
                     sla_numStudentsFilled_map[selectedSLA] = sla_numStudents_map.at(selectedSLA);
                  }
               }
               global_school_idx += local_school_idx;
            }
            remaining_schools -= num_schools_per_bin * num_bins;
         }
      }
   }
   school.close();
   school_in_dz.close();

   /*
    * Load pairwise distances between SLAs
    */
   std::map <sla_type, std::map <double, sla_type> > sla_distance_sla_map;
   std::ifstream sla_distance_matrix(database_input + "DistanceSLA.txt");
   std::cout << "Loading SLA distance matrix" << std::endl;

   /* First get which SLAs we have distances for... */
   std::getline(sla_distance_matrix, cline);
   std::istringstream     iss(cline);
   sla_type               csla;
   std::vector <sla_type> sla_list;
   while (iss >> csla)
   {
      sla_list.push_back(csla);
   }
   std::cout << "Stored " << sla_list.size() << " SLAs" << std::endl;

   /* ...then for every SLA, create a map of (every other) SLA to the distance between them */
   unsigned sla_row_idx = 0;
   while (std::getline(sla_distance_matrix, cline))
   {
      std::istringstream iss(cline);
      double    dist;
      sla_type sla_row = sla_list.at(sla_row_idx++);
      sla_distance_sla_map[sla_row].emplace();

      for (sla_type sla_col_idx = 0; iss >> dist; ++sla_col_idx)
      {
         if (dist < MAX_STUDENT_COMMUTE_DISTANCE)
         {
            sla_type sla_col = sla_list.at(sla_col_idx);
            sla_distance_sla_map.at(sla_row)[dist] = sla_col;
         }
      }
   }

   unsigned num_without_schools = 0;

   /* Generate CDFs for every entry in the SLA distance matrix */
   std::map <sla_type, std::map <double, std::vector <std::pair <unsigned long, double> > > > sla_dist_school_id_cdf_map;
   std::map <sla_type, double> school_zone_scaler;
   std::cout << "Generating CDFs for all SLA->SLA distances..." << std::endl;
   unsigned long i = 0;
   for (const auto& distance_sla_map : sla_distance_sla_map)
   {
      auto from_SLA = distance_sla_map.first;
      sla_dist_school_id_cdf_map[from_SLA].emplace();

      unsigned long total_school_sizes = 0;
      std::vector <unsigned long> school_ids;
      std::vector <unsigned long> school_sizes;

      for (const auto& centry : distance_sla_map.second)
      {
         auto cdist  = centry.first;
         auto to_SLA = centry.second;

         /* This is an ordered map, so every distance entry before is less than this distance */
         if (sla_school_map.find(to_SLA) != sla_school_map.end())
         {
            for (const auto& cschool : sla_school_map.at(to_SLA))
            {
               school_ids.push_back(cschool->ID);
               school_sizes.push_back(cschool->max_students);
            }
         }

         if (school_sizes.size() > 0)
         {
            std::vector <double> school_dist;
            /* Re-normalise distributions */
            for (auto csize : school_sizes)
            {
               school_dist.push_back((double)csize);
            }
            sla_dist_school_id_cdf_map.at(from_SLA)[cdist] = generateCDF(school_ids, school_dist);

            /*
             * std::cout << "CDF from SLA:" << from_SLA << " to SLA:" << to_SLA << " (distance: " << cdist
             *                      << ") has " << school_ids.size() << " schools in it." << std::endl;
             */
         }
      }
      std::cout << "\rGenerated CDF for SLA:" << from_SLA << " (number " << ++i << "/" << sla_distance_sla_map.size() << "; num distances: "
                << sla_dist_school_id_cdf_map.at(from_SLA).size() << "; num schools: " << sla_dist_school_id_cdf_map.at(from_SLA).rbegin()->second.size() << ")" << std::fflush;

      school_zone_scaler[from_SLA] = 1.0;

      if (sla_dist_school_id_cdf_map.at(from_SLA).rbegin()->second.size() == 0)
      {
         num_without_schools++;
      }
   }
   std::cout << std::endl;

   std::cout << "Number of SLAs without schools within " << MAX_STUDENT_COMMUTE_DISTANCE * 100 << "km: " << num_without_schools << std::endl;

   unsigned long n_students_assigned = 0, n_students_assigned_to_their_sla = 0, n_students_not_assigned = 0;

   /* Assign students to schools based on zones */
   unsigned long tot_num_students = students.size();

   double tot_school_dist = 0.0;
   while (students.size() > 0)
   {
      unsigned selected_student = rand() % students.size();
      unsigned agent_idx        = students.at(selected_student);
      students.erase(students.begin() + selected_student);

      if (agents[agent_idx]->my_wg != 0)
      {
         std::cerr << "Student #:" << agent_idx << " already assigned WG." << std::endl;
         continue;
      }

      sla_type c_SLA = cd_sla_map.at(agents[agent_idx]->my_cd);

      if (sla_distance_sla_map.find(c_SLA) == sla_distance_sla_map.end())
      {
         std::cerr << "Student #:" << agent_idx << " in unknown SLA:" << c_SLA << std::endl;
         continue;
      }
      ;

      auto& dist_school_id_cdf_map = sla_dist_school_id_cdf_map.at(c_SLA);

      bool increase_zone = false;

      if (dist_school_id_cdf_map.size() == 0 || dist_school_id_cdf_map.rbegin()->second.size() == 0)
      {
         //std::cerr << "SLA: " << c_SLA << " has no schools within " << MAX_STUDENT_COMMUTE_DISTANCE * 100 << "km" << std::endl;
      }
      else
      {
         for (unsigned i = 0; agents.at(agent_idx)->my_wg >= 0 && i < MAX_ALLOCATION_ATTEMPTS; ++i)
         {
            /* Use CDF to find  */
            double chosen_dist = invFoldedNormCDF(school_zone_scaler.at(c_SLA) * SCHOOL_ZONE_SIGMA, rand() / double(RAND_MAX));

            /* Find the closest SLA that is within this distance */
            auto farthest_dist_it = dist_school_id_cdf_map.begin();

            if (dist_school_id_cdf_map.size() > 1)
            {
               if ((farthest_dist_it = dist_school_id_cdf_map.lower_bound(chosen_dist)) == dist_school_id_cdf_map.end())
               {
                  /*
                   * std::cerr << "SLA: " << c_SLA << " has no schools within " << chosen_dist * 100 << "km. Max distance is "
                   *                      << dist_school_id_cdf_map.rbegin()->first * 100 << "km"
                   *                      << " (num distances: " << dist_school_id_cdf_map.size() << ")." << std::endl;
                   */
                  continue;
               }
            }

            auto school_id_cdf = farthest_dist_it->second;

            unsigned long chosen_school_id = sampleCDF(school_id_cdf);

            /*
             * std::cout << "Chose School:" << chosen_school_id << " for agent " << agent_idx
             *                      << " (allowable catchment: " << chosen_dist << "; num SLAs: "
             *                      << std::distance( dist_school_id_cdf_map.begin(), farthest_dist_it )
             *                      << "; num schools: "<< school_id_cdf.size() << ")." << std::endl;
             */
            struct School *my_school = school_id_map.at(chosen_school_id);

            if (my_school->num_students >= my_school->max_students)
            {
               /*
                * std::cerr << "School:" << my_school->ID << " at capacity (" << school_id_cdf.size()
                *                      << " schools within chosen dist: " << chosen_dist
                *                      << "; schools within 150km of " << c_SLA << ":" << dist_school_id_cdf_map.rbegin()->second.size() << ")." << std::endl;
                */
               increase_zone = true;
               continue;
            }

            agents[agent_idx]->my_wg = -my_school->ID;
            my_school->num_students++;

            if (my_school->SLA == c_SLA)
            {
               n_students_assigned_to_their_sla++;
            }
            n_students_assigned++;

            tot_school_dist += chosen_dist;
         }
      }

      if (agents[agent_idx]->my_wg >= 0)
      {
         n_students_not_assigned++;
         if (increase_zone)
         {
            school_zone_scaler.at(c_SLA) += 0.1;
         }
      }

      if ((n_students_assigned + n_students_not_assigned) % 1000 == 0)
      {
         std::cout << "\r# students assigned: " << n_students_assigned << "/" << tot_num_students
                   << " (" << students.size() << " students left). # assigned to their SLA: " << n_students_assigned_to_their_sla
                   << " (" << n_students_assigned - n_students_assigned_to_their_sla << " commuters; "
                   << n_students_not_assigned << " not assigned; avg dist: " << tot_school_dist / (double)n_students_assigned << ")." << std::flush;
      }
   }

   std::cout << "# students assigned: " << n_students_assigned << ". # assigned to their SLA: " << n_students_assigned_to_their_sla
             << "(" << n_students_assigned - n_students_assigned_to_their_sla << " commuters; " << n_students_not_assigned << "not assigned)." << std::endl;

   unsigned long n_teachers_assigned = 0;
   /* Iterate through all SLAs */
   for (auto& sla_school_it : sla_school_map)
   {
      /* Iterate through all schools in this SLA */
      for (auto& cschool : sla_school_it.second)
      {
         dz_type currentDZ = cschool->DZ;

         unsigned num_teachers_required = ceil((double)cschool->num_students / (double)NUM_STUDENTS_TO_TEACHERS);

         std::cout << "\rNumber of students in school:" << cschool->ID << " is " << cschool->num_students;

         /* Randomly pick the teachers from this DZ */
         if (dz_to_adult_idx.find(currentDZ) == dz_to_adult_idx.end())
         {
            std::cerr << "Could not find DZ:" << currentDZ << " in list." << std::endl;
            continue;
         }
         auto& potential_adults = dz_to_adult_idx.at(currentDZ);

         unsigned num_teachers_assigned = 0;
         while (num_teachers_assigned < num_teachers_required)
         {
            unsigned long rval      = rand() % potential_adults.size();
            unsigned long adult_idx = potential_adults.at(rval);
            agents[adult_idx]->my_wg = -cschool->ID;
            num_teachers_assigned++;

            std::cout << "\rAssigned adult " << adult_idx << " (out of " << potential_adults.size()
                      << ") to school:" << cschool->ID
                      << " (" << num_teachers_assigned << "/" << num_teachers_required << ")";

            potential_adults.erase(potential_adults.begin() + rval);
            n_teachers_assigned++;

            if (potential_adults.size() == 0)
            {
               std::cerr << "No potential adults left in DZ:" << currentDZ << std::endl;
               break;
            }
         }
      }
      std::cout << std::endl;
   }
   std::cout << "Number of teachers assigned: " << n_teachers_assigned << std::endl;
}

std::shared_ptr <SimpleAgent> generateAgent(unsigned long id, cd_type cd, unsigned long hh, unsigned long hc,
                                            const std::vector <std::pair <AgeGroup, double> >& age_cdf)
{
   std::shared_ptr <SimpleAgent> nagent = std::make_shared <SimpleAgent>();
   nagent->my_id = id;
   nagent->my_hh = hh;
   nagent->my_hc = hc;
   nagent->my_cd = cd;
   nagent->my_wg = 0;

   /* Sample my age from the age CDF */
   nagent->my_age = sampleCDF(age_cdf);
   return(nagent);
}

void printPopulation(const std::vector <std::shared_ptr <SimpleAgent> >& agents, const std::string& fname)
{
   std::ofstream pFile(fname);

   for (auto& cagent : agents)
   {
      pFile << cagent->my_age << " " << cagent->my_hh << " " << cagent->my_hc << " " << cagent->my_wg << " " << cagent->my_cd << std::endl;
   }
   pFile.close();
}
