/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright   Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-201), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright   Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   This code is released only for internal review by Nature Communications, manuscript: NCOMMS-21-29568.

   *EndCopyright:
*/

#ifdef _MSC_VER
#  define _CRT_SECURE_NO_WARNINGS
#endif

#include "agent.h"
#include "create_population.h"
#include "epidemic.h"
#include "debug.h"
#include "enums.h"


#include <cstddef>
#include <fstream>
#include <iostream>
#include <map>
#include <string>
#include <vector>
#include <memory>
#include <time.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <getopt.h>
#include <unistd.h>
#include <random>
#include <stack>

// HH, HC, CD, SLA, WG, CLASS, GRADE, SCH
std::vector<unsigned long> g_cumulative_infected_in_group {
   0, 0, 0, 0, 0, 0, 0, 0
};


SeedType seed_type;

// unless otherwise stated:
// illness refers to symptomatic cases
// incidence refers to detected cases (symptomatic or not)
// infected refers to all infections, symptomatic or not
// attack rate refers to total detected cases, symptomatic or not

unsigned long g_detected_incidence  = 0, // counts all newly detected cases (symptomatic or not)
              g_detected_prevalence = 0, // counts all currently infected, detected agents (symptomatic or not)
              g_detected_cumulative = 0, // totals all detected cases (symptomatic or not)
              g_infected_incidence  = 0, // counts all new infections
              g_infected_prevalence = 0, // hcounts all currently infected agents
              g_infected_cumulative = 0, // totals all infections
              g_illness_incidence   = 0, // counts all new illnesses, counted at time of infection (not symptom onset)
              g_illness_prevalence  = 0, // counts all currently ill agents
              g_illness_cumulative  = 0, // totals all illnesses
              g_number_removed      = 0, // counts all removed agents, symptomatic or not
              g_detected_incidence_VACge = 0,
              g_detected_incidence_VACpr = 0,
              g_infected_incidence_VACge = 0,
              g_infected_incidence_VACpr = 0;

int num_CD_incidence  = 0,               // counts number of CD regions newly infected (0 -> at least one new case)
    num_CD_prevalence = 0;               // coutns number of CD regions with at least one acive case (detected infection)

// These will count DETECTED inicidence and prevalence, not illnesses
std::map<unsigned long, unsigned long> g_sla_prevalence_map,
                                       g_sla_number_removed_map,
                                       g_sla_incidence_map,
                                       g_sla_attack_rate_map;

// These will count DETECTED incidence and prevalence, not illnesses
std::map<AgeGroup, std::map<unsigned long, unsigned long> > g_ag_sla_prevalence_map,
                                                            g_ag_sla_number_removed_map,
                                                            g_ag_sla_incidence_map;

// CZ: 12122020 - adding an age-group global inidence and prevalence output
// for illness, detection, and infection, to avoid any confusion and facilitate analysis / sanity checking
std::map<AgeGroup, unsigned long> g_ag_infection_incidence_map,
                                  g_ag_infection_prevalence_map,
                                  g_ag_illness_incidence_map,
                                  g_ag_illness_prevalence_map,
                                  g_ag_detection_incidence_map,
                                  g_ag_detection_prevalence_map,
                                  g_ag_detection_incidence_VACpr_map,
                                  g_ag_detection_incidence_VACge_map,
                                  g_ag_infection_incidence_VACpr_map,
                                  g_ag_infection_incidence_VACge_map,
                                  ag_pops_map;


void seedInfection(sla_map_type&);
void updateAgentState(std::shared_ptr<Agent>, double, unsigned);
void simulateDaytime(agent_map_type&);
void simulateNighttime(agent_map_type&);
void simulateDaytimeWeekend(agent_map_type&); //CZ COV19 21112020


//////////
//CZ COV19 SD 17032020 - modified daytime and nighttime functions during SD period
void simulateDaytime_SD(agent_map_type&); // CZ COV19 31032020 - expanding generality of these functions
void simulateNighttime_SD(agent_map_type&);

void simulateDaytimeWeekend_SD(agent_map_type&);
//CZ COV19 21112020, same as simulatenighttime_sd but removed bookkeeping to avoid double-counting

void step_and_detect(agent_map_type&);  // CZ had to separate this from simulateDaytime etc., so isolation states can be updated accurately
void step_without_detection(agent_map_type&);

void assign_SD_compliance(agent_map_type&);
void assign_CI_compliance(agent_map_type&);
void assign_HQ_compliance(agent_map_type&);                   // CZ COV19 HQ 31032020, HQ compliance will be conditional on CI compliance.

void assign_SC_compliance_students_teachers(agent_map_type&); // this is static
void assign_SC_compliance_parents(agent_map_type&, double);   // this is re-computed dynamically

void detect_ill_agents(agent_map_type&, double P_detect);     // making this independent of the 'agent->step()' function


// update HQ steps the quarantine clock for those agents that are isolated in HQ
void update_HQ(agent_map_type&);            // will update quarantine states based on household cases and compliance

//determines isolation state of each agent based on measures implemented (triggered) and compliance state
void update_Isolation_states(agent_map_type&, bool, bool, bool, bool);

void reZero(hh_map_type&, hc_map_type&, cd_map_type&, dz_map_type&, wg_map_type&, school_map_type&, wg_map_type&, wg_map_type&, sla_map_type&);

// uses the independently-computed compliance states to determine combined state (i.e., SD_CI_HQ...)
void set_Compliance_states_all(agent_map_type&);

void set_Isolation_Factors_home(agent_map_type&, double, double, double, double);
void set_Isolation_Factors_work(agent_map_type&, double, double, double, double);
void set_Isolation_Factors_community(agent_map_type&, double, double, double, double);

///////

//CZ: 10 06 2020 - disease parameter inputs for sensitivity analysis and uncertainty analysis
void parse_Parameters(const std::string& input_fname);
void parse_SD_intervention_inputs(const std::string& input_fname);                                    //CZ COV19 SD 17032020
void parse_pharm_intervention_params(const std::string& database_input, const sla_map_type& sla_map); /* Kit's edit - intervention */
/* Kit's edit - AVE */
void dynamic_intervention(agent_map_type&);
// CZ: pre-pandemic and progressive vaccination
void progressive_vaccination_priority(agent_map_type&, std::stack<unsigned long>&, std::stack<unsigned long>&, std::stack<unsigned long>&);
void progressive_vaccination_general(agent_map_type&, std::stack<unsigned long>&, std::stack<unsigned long>&, std::stack<unsigned long>&);
void pre_epidemic_vaccination(agent_map_type&, std::stack<unsigned long>&, std::stack<unsigned long>&, std::stack<unsigned long>&);
void count_vaccinated_agents(agent_map_type&);


//void dynamic_quarantine(agent_map_type&);                                                     /* Kit's edit - quarantine */
// CZ COV19 31032020 - I am removing Kit's dynamic quarantine methods, as they may cause issues with the other SD measures
// consult previous versions (including the CDIP release version) if neighborhood-targeted dynamic quarantine is desired. Or we may re-implement using the new SD structures

int count_CD_incidence(cd_map_type& census_districts);
int count_CD_prevalence(cd_map_type& census_districts);

unsigned long getNumAgentsIll(agent_map_type&);
unsigned long getNumAgentsInfected(agent_map_type&);
unsigned long getNumAgentsRemoved(agent_map_type&);
void getSLAIncidencePrevalence(sla_map_type&, std::map<unsigned long, double>&, std::map<unsigned long, double>&);

void parse_arguments(int argc, char *argv[], double&, double&, unsigned&,
                     unsigned&, unsigned&, double&, double&, std::string &,
                     std::string &, bool&, std::string &, std::string &, std::string &);
/* Kit's edit - random seed - intervention */
int makeDir(const std::string&);


/*beginning Kit's edit -intervention*/
// These are public varibles used for the intervention things
int                   vac_SLA_bool = 0;    // vaccinate selected sla boolean
std::vector<unsigned> vac_SLAs;            // SLAs to vaccinat
std::vector<unsigned> AV_SLAs;             // SLAs to apply anti virals

// Intervention variables
double perc_vac = 0.0;                        // proportion of population to pre-vaccinate
double GTAP     = 0.0;                        // proportion of population for geographical targeted AV
double TAP      = 0.0;                        // proportion of population for targeted AV

//CZ: progressive vaccination - some vaccines are delivered pre-epidemic, and some are delivered during epidemic
int n_vac_per_week_1     = 0;             // 80k is the planned weekly roll-out rate
int n_vac_pre_epidemic_1 = 5000000;       // this is the number of immunisations delivered prior to epidemic outset

int n_vac_per_week_2     = 0;             // 80k is the planned weekly roll-out rate
int n_vac_pre_epidemic_2 = 10000000;      // this is the number of immunisations delivered prior to epidemic outset

// CZ: age-stratified vaccination
double        p_vac_65_and_over = 100;      // agegroup 4
double        p_vac_18_to_64    = 10;       // age groups 2 and 3
double        p_vac_children    = 1;        // age groups 0 and 1

unsigned long num_vac          = 0;
unsigned long num_vac_65_plus  = 0;
unsigned long num_vac_18_to_64 = 0;
unsigned long num_vac_children = 0;

unsigned long num_vac_PV          = 0;
unsigned long num_vac_65_plus_PV  = 0;
unsigned long num_vac_18_to_64_PV = 0;
unsigned long num_vac_children_PV = 0;

unsigned long num_vac_GV          = 0;
unsigned long num_vac_65_plus_GV   = 0;
unsigned long num_vac_18_to_64_GV  = 0;
unsigned long num_vac_children_GV  = 0;

// CZ: these are our vaccine parameters leftover from Kit's implementation
// for the influenza interventions study (interfering with influenza...)
// does not include efficacy for disease...
// will examine how this was implemented in the case of antivirals (AVEd)

double VEs_1 = 0.68;                          // vaccine efficacy of susceptibility
double VEi_1 = 0.68;                          // vaccine efficacy of infectiousness
double VEd_1 = 0.68;                          // adding vaccine efficacy for disease
                                              // (modifies probability of being symptomatic)

double VEs_2 = 0.386;                         // vaccine efficacy of susceptibility
double VEi_2 = 0.386;                         // vaccine efficacy of infectiousness
double VEd_2 = 0.386;                         // efficacy for disease


double AVEs = 0.3;                         // Anti viral efficacy of susceptibility
double AVEi = 0.62;                        // Anti viral efficacy of infectiousness
double AVEd = 0.6;                         // Anti viral efficacy of disease

bool   pharm_intervention = true;          // Boolean for Running intervention
int    Response_delay     = 14;            // delay between Sympotmatic_infection and intervention
int    AV_doses           = 0;             // Count of AV doses used
int    no_infected        = 0;             // Temporary check kit fair 25/07/18

//CZ: modifying detection to allow asymptomatic and pre-symptomatic detection
double Detection       = 1.0;                 // probability of detection of illness
double Detection_asymp = 0.2;                 // probability of detection of pre- or asymptomatic infection
/*end Kit's edit -intervention*/
//unsigned reuse_seed = 0;         /* Kit's edit - random seed */


// CZ COV19 17032020
// default global parameters (run values will be parsed in from input file)
// Socia distancing parameters
// intervention threshold
bool   SD_intervention       = true;
int    n_detected_SD_trigger = 1000;  // cumulative attack rate trggering intervention
double compliance_rate_SD    = 0.5;   // fraction of people who comply
int    duration_SD           = 14;    // duration of social distancing measures (in days)
double FoI_SD_community      = 0.1;   // factor by which an infected person's community contribution to FoI is multiplied with SD
double FoI_SD_work           = 0.25;  // factor by which an infected person's work/school contribution to FoI is multiplied with SD
double FoI_SD_home           = 1;     // factor by which an infected person's home contribution to FoI is multiplied with SD

bool   School_closure        = false; // turn on for school closure
int    n_detected_SC_trigger = 1000;
int    duration_SC           = 14;
double FoI_SC_community      = 0.1;
double FoI_SC_work           = 0.25;
double FoI_SC_home           = 1;
double p_parent_stays_home   = 0.25;

bool   CI_intervention    = false; // Case Isolation, from beginning of epidemic
double compliance_rate_CI = 0.7;
double FoI_CI_community   = 0.25;
double FoI_CI_work        = 0.25;
double FoI_CI_home        = 1;
int    g_CI_agents        = 0; // the count of agents affected by CI, for output.

//CZ COV19 31032020
bool   HQ_intervention    = false; // Home quarantine, in response to illness in household
double compliance_rate_HQ = 0.7;   // compliance probability given CI compliance
int    duration_HQ        = 14;    // length of quarantine period in days
double FoI_HQ_community   = 0.25;
double FoI_HQ_work        = 0.25;
double FoI_HQ_home        = 2.0;
int    g_HQ_agents        = 0; // the count of agents affected by HQ, for output.


//CZ: COV19 - disease parameters for sensitivity analysis etc. 10 06 2020
// set default disease parameters - these can be altered by Parameters_input.dat
// read by the parse_Parameters() function.
double latent_period         = 0;
double incubation_period_mu  = 1.644;  //from Lauer et al.
double incubation_period_sig = 0.363;  // from Lauer et al.
double recovery_period_hi    = 10;     //time between end of exponential period an recovery - upper bound - CDC replication competent virus reports
double recovery_period_low   = 5;
//double infectious_period          = 15; // this will be the recovery period plus the incubation period
double adult_symptomatic_fraction = 0.67;
double child_symptomatic_fraction = 0.134;
double asymptomatic_infectivity   = 0.3;

//CZ 11 06 2020 incubation period will be set from a lognormal distribution
// CZ 11 06 2020 symptomatic period (time from onset to recovery) will be set as a uniform
// distribution (Qifang Bi et al., Figure S5 ) from 14 - 28 days
std::default_random_engine generator;



// CZ: set to re-use rng seed
bool reuse_seed = true;


int main(int argc, char *argv[])
{
   double      p_trans                   = 0.08;
   double      trans_scaler              = 2.75;
   unsigned    n_runs                    = 1;
   unsigned    n_days                    = 180;
   unsigned    n_slas_to_infect          = 1;
   double      p_agents_to_infect_in_sla = -1.0;
   double      p_pop_to_infect           = -1.0;

   std::string current_dir = (char *)get_current_dir_name();

   std::string database_output = current_dir + "/AmTraC_outputs/2016_population_1/";
   std::string database_input  = current_dir + "/AmTraC_inputs/2016_population_1/";

   std::string SD_intervention_input_fname = "SD_intervention_input.dat";

   std::string Pharm_intervention_input_fname = "pharm_intervention_input.dat";

   std::string Parameter_input_fname = "Parameter_input.dat";

   parse_arguments(argc,
                   argv,
                   p_trans,
                   trans_scaler,
                   n_runs, n_days,
                   n_slas_to_infect,
                   p_agents_to_infect_in_sla,
                   p_pop_to_infect,
                   database_output,
                   database_input,
                   reuse_seed,
                   SD_intervention_input_fname,
                   Pharm_intervention_input_fname,
                   Parameter_input_fname);



   std::string pop_fname = database_input + "population.dat";

   std::cout << "Starting simulations with following settings:"
             << "\n\toutput directory: " << database_output
             << "\n\tinput directory: " << database_input
             << "\n\tReuse_seed: " << reuse_seed     /* Kit's edit  random seed */
             << "\n\tRuns: " << n_runs
             << "\n\tDays per run: " << n_days
             << "\n\tP(transmission): " << p_trans
             << "\n\tTransmission Scaler: " << trans_scaler
             << "\n\tSLAs infected: " << n_slas_to_infect
             << "\n\tAgents infected per SLA: " << p_agents_to_infect_in_sla
             << "\n\tProportion of population to infect: " << p_pop_to_infect
             << std::endl;

   seed_type = SeedType::Airport;
   if (p_pop_to_infect > 0)
    {
       seed_type = SeedType::PopulationRandom;
    }
   else if (p_agents_to_infect_in_sla > 0)
    {
       seed_type = SeedType::NumberOfSLAs;
    }

   /* Set up data structures */
   agent_map_type agents;
   hh_map_type    households;
   hc_map_type    household_clusters;
   cd_map_type    census_districts;
   wg_map_type    working_groups;
   wg_map_type    schools, grades, classes;
   dz_map_type    destination_zones;
   sla_map_type   statistical_local_areas;

   /* Scale contact rates by `p_trans' for inputting to agents */
   Epidemic::init_transmission_rates(p_trans, trans_scaler);

   /* Make a folder for the current set of results */
   struct timeval tv;
   gettimeofday(&tv, NULL);
   struct tm      *tm_info = localtime(&tv.tv_sec);
   char           buffer[26];
   strftime(buffer, 26, "%Y-%m-%d_%H-%M-%S", tm_info);

   std::stringstream base_dir_ss;
   base_dir_ss << database_output << "results/"
               << "P" << p_trans
               << "_T" << trans_scaler
               << "_D" << buffer;

   if (seed_type == SeedTypes::Airport)
    {
       base_dir_ss << "_AP";
    }
   else if (seed_type == SeedTypes::NumberOfSLAs)
    {
       base_dir_ss << "_S" << n_slas_to_infect
                   << "_A" << p_agents_to_infect_in_sla;
    }
   else if (seed_type == SeedTypes::PopulationRandom)
    {
       base_dir_ss << "_PP" << p_pop_to_infect;
    }

   std::size_t last_fs  = pop_fname.find_last_of('L');
   std::size_t last_dot = pop_fname.find_last_of('.');
   if (last_fs != -1)
    {
       std::string pl_str = pop_fname.substr(last_fs + 1, last_dot - last_fs - 1);
       base_dir_ss << "_L" << pl_str;
    }

   std::string base_dir(base_dir_ss.str());

   std::size_t dot_loc;
   while ((dot_loc = base_dir.find(".")) != std::string::npos)
    {
       base_dir.replace(dot_loc, 1, "-");
    }

   if (makeDir(base_dir) != 0)
    {
       std::cerr << "Making directory: " << base_dir << " failed. Exiting." << std::endl;
       return(1);
    }
   std::cout << "Created folder " << base_dir << "." << std::endl;

   // beginning Kit's edit - random seed
   // This is used to determine the random seed number to ensure all repeitions are the same. All related edits are labeled "- random seed"
   std::vector<unsigned> random_seed;
   unsigned              seed;

   std::ofstream         saved_seed(base_dir + "/saved_seed.dat");
   std::ifstream         loaded_seed(database_input + "loaded_seed.dat");
   if (reuse_seed)
    {
       if (!loaded_seed.good())
        {
           std::cout << "File loaded_seed.dat does not exist." << std::endl;
        }
       while (loaded_seed >> seed)
        {
           random_seed.push_back(seed);
        }
       loaded_seed.close();

       seed = random_seed[0];
    }
   else
    {
       seed = time(NULL);
    }
   srand(seed);

   std::stringstream run_savefile_seed;
   run_savefile_seed << base_dir << "/saved_seed";
   std::fstream      outfile_stream_seed(run_savefile_seed.str() + ".dat");
   // end Kit's edit - random seed


   /* Create the population */
   std::cout << "Reading Population..." << std::endl;

   CreatePopulation::readPopulation(agents, households, household_clusters, census_districts, destination_zones, statistical_local_areas,
                                    working_groups, schools, grades, classes, database_input, pop_fname);

   /* read population sets the id of the parameters in input */
   std::cout << "Done" << std::endl;

   /* Pre-load the global maps */
   std::stringstream SLA_pop_savefile;
   SLA_pop_savefile << base_dir << "/SLA_pop.dat";

   std::ofstream pop_dat_stream(SLA_pop_savefile.str());

   for (auto sla_entry : statistical_local_areas)
    {
       g_sla_prevalence_map.emplace(sla_entry.first, 0);
       g_sla_number_removed_map.emplace(sla_entry.first, 0);
       g_sla_incidence_map.emplace(sla_entry.first, 0);
       g_sla_attack_rate_map.emplace(sla_entry.first, 0);
       pop_dat_stream << sla_entry.second->getID() << " " << double( sla_entry.second->getAgentIdxs().size()) << std::endl;
    }
   pop_dat_stream.close();

   for (AgeGroup ag = AgeGroups::Age0to4; ag <= AgeGroups::Age65Plus; ag = AgeGroup(ag + 1))
    {
       g_ag_sla_prevalence_map.emplace(ag, g_sla_prevalence_map);
       g_ag_sla_number_removed_map.emplace(ag, g_sla_number_removed_map);
       g_ag_sla_incidence_map.emplace(ag, g_sla_incidence_map);

       //CZ: copy the above for the global age-group incidence maps:
       g_ag_infection_incidence_map.emplace(ag, 0); // map pairs are age group -> number infected etc...
       g_ag_infection_prevalence_map.emplace(ag, 0);
       g_ag_illness_incidence_map.emplace(ag, 0);
       g_ag_illness_prevalence_map.emplace(ag, 0);
       g_ag_detection_incidence_map.emplace(ag, 0);
       g_ag_detection_prevalence_map.emplace(ag, 0);
       g_ag_detection_incidence_VACpr_map.emplace(ag, 0);
       g_ag_detection_incidence_VACge_map.emplace(ag, 0);
       g_ag_infection_incidence_VACpr_map.emplace(ag, 0);
       g_ag_infection_incidence_VACge_map.emplace(ag, 0);
       ag_pops_map.emplace(ag, 0);
    }

   std::stringstream AgeGroup_pop_savefile;
   AgeGroup_pop_savefile << base_dir << "/AgeGroup_pop.dat";
   std::ofstream     AG_pop_dat_stream(AgeGroup_pop_savefile.str());

   for (auto agent_entry : agents)
    {
       auto ag = (agent_entry.second)->getAgeGroup();
       ag_pops_map.at(ag)++;
    }

   for (AgeGroup ag = AgeGroups::Age0to4; ag <= AgeGroups::Age65Plus; ag = AgeGroup(ag + 1))
    {
       AG_pop_dat_stream << ag_pops_map.at(ag) << " ";
    }

   AG_pop_dat_stream << std::endl;
   AG_pop_dat_stream.close();



   /* Seeded SLAs are those with international airports */
   if (seed_type == SeedTypes::Airport)
    {
       Epidemic::init_airport_seeded_slas(database_input, statistical_local_areas);
    }
   else if (seed_type == SeedTypes::PopulationRandom)
    {
       Epidemic::init_random_population_seed(agents);
    }


   // CZ 10 06 2020 parse input parameter file
   // CZ: 06 12 2020 order matters, parameter input now has booleans
   // for pharmaceutical and SD interventions.
   std::string   Parameter_input_file = database_input + "Parameter_input/" + Parameter_input_fname;

   std::ifstream Parameter_file(Parameter_input_file);

   if (!Parameter_file.good())
    {
       std::cout << "Parameter input file named: " << Parameter_input_file << " does not exist, using default parameters" << std::endl;
    }
   else
    {
       parse_Parameters(Parameter_input_file);
    }


   // beginning Kit's edit -intervention
   // This section reads the intervention inputs
   // CZ: modified for consistency and naming conventions

   //if (pharm_intervention)
   //{
   std::string   Pharm_intervention_input_file = database_input + "Pharm_intervention_input/" + Pharm_intervention_input_fname;
   std::ifstream Pharm_intervention_file(Pharm_intervention_input_file);
   if (!Pharm_intervention_file.good())
    {
       std::cout << "File intervention_input.dat does not exist." << std::endl;
    }
   else
    {
       parse_pharm_intervention_params(Pharm_intervention_input_file, statistical_local_areas);
    }
   //}
   // end Kit's edit -interventio/

   //if (SD_intervention)
   //{
   std::string   SD_intervention_input_file = database_input + "SD_intervention_input/" + SD_intervention_input_fname;
   std::ifstream SD_intervention_file(SD_intervention_input_file);
   if (!SD_intervention_file.good())
    {
       std::cout << "File SD_intervention_input.dat named: " << SD_intervention_input_file << " does not exist, using default parameters" << std::endl;
    }
   else
    {
       parse_SD_intervention_inputs(SD_intervention_input_file);
    }
   //}


   // save parameters
   std::stringstream config_savefile;
   config_savefile << base_dir << "/config.dat";

   std::ofstream configfile_stream(config_savefile.str());

   configfile_stream
      << "\n" << "-----general parameters-----"
      << "\n" << "transmission probability: " << p_trans
      << "\n" << "transmission scalar (scales R0): " << trans_scaler
      << "\n" << "population: " << agents.size()
      << "\n" << "number of days: " << n_days
      << "\n" << "number of runs: " << n_runs
      << "\n" << "illness detection probability (per ill individual, per day): " << Detection
      << "\n" << "pre and asymptomatic detection probability (i.e., from testing): " << Detection_asymp
      << "\n"
      << "\n" << "-----disease parameters-----"
      << "\n" << "incubation period mu: " << incubation_period_mu
      << "\n" << "incubation period sigma: " << incubation_period_sig
      << "\n" << "days from end of incubation to recovery (upper bound): " << recovery_period_hi
      << "\n" << "days from end of incubation to recovery (lower bound): " << recovery_period_low
      << "\n" << "adult symptomatic fraction: " << adult_symptomatic_fraction
      << "\n" << "child symptomatic fraction: " << child_symptomatic_fraction
      << "\n" << "asymptomatic infectivity factor: " << asymptomatic_infectivity
      << "\n"
      << "\n" << "-----antiviral intervention parameters-----"
      << "\n" << "pharmaceutical intervention (or pre-vac) on? [0=no, 1=yes]: " << pharm_intervention
      << "\n" << "GTAP coverage: " << GTAP
      << "\n" << "TAP coverage: " << TAP
      << "\n" << "antiviral efficacy for susceptibility: " << AVEs
      << "\n" << "antiviral efficacy for infectiousness: " << AVEi
      << "\n" << "antiviral efficacy for disease (symptom expression): " << AVEd
      << "\n" << "logistical delay time: " << Response_delay
      << "\n"
      << "\n" << "-----vaccine parameters-----"
      << "\n" << "pharmaceutical intervention (or pre-vac) on? [0=no, 1=yes]: " << pharm_intervention
      << "\n" << "number of people pre-vaccinated (priority vaccine): " << n_vac_pre_epidemic_1
      << "\n" << "number of people vaccinated per week (priority vaccine): " << n_vac_per_week_1
      << "\n" << "number of people pre-vaccinated (general vaccine): " << n_vac_pre_epidemic_2
      << "\n" << "number of people vaccinated per week (general vaccine): " << n_vac_per_week_2
      << "\n" << "probability of vaccination given Age 65+: " << p_vac_65_and_over
      << "\n" << "probability of vaccination given Age 18-64: " << p_vac_18_to_64
      << "\n" << "probability of vaccination given Age under 18: " << p_vac_children
      << "\n" << "vaccine efficacy for susceptibility (priority vaccine): " << VEs_1
      << "\n" << "vaccine efficacy for infectiousness (priority vaccine): " << VEi_1
      << "\n" << "vaccine efficacy for disease (symptom expression, priority vaccine): " << VEd_1
      << "\n" << "vaccine efficacy for susceptibility (general vaccine): " << VEs_2
      << "\n" << "vaccine efficacy for infectiousness (general vaccine): " << VEi_2
      << "\n" << "vaccine efficacy for disease (symptom expression, general vaccine): " << VEd_2
      << "\n"
      << "\n" << "----Social Distancing for COV19----"
      << "\n" << "Social distancing on? [0=no, 1=yes]: " << SD_intervention
      << "\n" << "cumulative attack rate threshold: " << n_detected_SD_trigger
      << "\n" << "SD compliance rate: " << compliance_rate_SD
      << "\n" << "duration of social distancing: " << duration_SD
      << "\n" << "Factor of social distancing for FoI in community: " << FoI_SD_community
      << "\n" << "Factor of social distancing for FoI at work and school: " << FoI_SD_work
      << "\n" << "Factor of social distancing for FoI at home: " << FoI_SD_home
      << "\n"
      << "\n" << "---School closure for COV19---"
      << "\n" << "School closure on? [0=no, 1=yes]: " << School_closure
      << "\n" << "cumulative attack rate threshold: " << n_detected_SC_trigger
      << "\n" << "duration of school closure: " << duration_SC
      << "\n" << "Factor of school closure for FoI in community: " << FoI_SC_community
      << "\n" << "Factor of school closure for FoI at work and school: " << FoI_SC_work
      << "\n" << "Factor of school closure for FoI at home: " << FoI_SC_home
      << "\n"
      << "\n" << "----Case Isolation for COV19----"
      << "\n" << "Case Isolation on? [0=no, 1=yes]: " << CI_intervention
      << "\n" << "CI compliance rate: " << compliance_rate_CI
      << "\n" << "Factor of case isolation for FoI in community: " << FoI_CI_community
      << "\n" << "Factor of case isolation for FoI at work and school: " << FoI_CI_work
      << "\n" << "Factor of case isolation for FoI at home: " << FoI_CI_home
      << "\n"
      << "\n" << "----Home Quarantine for COV19----"
      << "\n" << "Home Quarantine on? [0=no, 1=yes]: " << HQ_intervention
      << "\n" << "HQ compliance rate: " << compliance_rate_HQ
      << "\n" << "Quarantine duration (days): " << duration_HQ
      << "\n" << "Factor of home quarantine for FoI in community: " << FoI_HQ_community
      << "\n" << "Factor of home quarantine for FoI at work and school: " << FoI_HQ_work
      << "\n" << "Factor of home quarantine for FoI at home: " << FoI_HQ_home
      << std::endl;

   configfile_stream.close();


   // CZ - did away with this, now viral load is computed when needed
   //Agent::computeViralLoad();

   /* TODO: parallelise runs */
   for (unsigned run = 0; run < n_runs; ++run)
    {
       clock_t tStart = clock();

       std::cout << "Initiating run " << run << "." << std::endl;

       std::cout << "\n" << "=== disease parameters ==="
                 << "\n" << "incubation period mu: " << incubation_period_mu
                 << "\n" << "incubation period sigma: " << incubation_period_sig
                 << "\n" << "days from end of incubation to recovery (upper bound): " << recovery_period_hi
                 << "\n" << "days from end of incubation to recovery (lower bound): " << recovery_period_low
                 << "\n" << "latent period, time between infection and onset of infectiousness (days): " << latent_period
                 << "\n" << "adult symptomatic fraction: " << adult_symptomatic_fraction
                 << "\n" << "child symptomatic fraction: " << child_symptomatic_fraction
                 << "\n" << "asymptomatic infectivity factor: " << asymptomatic_infectivity
                 << "\n" << "illness detection probability (per ill individual, per day): " << Detection
                 << "\n" << "pre and asymptomatic detection probability (i.e., from testing): " << Detection_asymp
                 << "\n"
                 << std::endl;



       // CZ COV19 SD 17032020
       bool SD_trigger = false;
       int  SD_timer   = 0;

       //CZ COV19 SC 31032020
       bool SC_trigger = false;
       int  SC_timer   = 0;


       // beginning Kit's edit  - random seed
       // This is used to determine the random seed number to ensure all repeitions are the same. All related edits are labeled "- random seed"
       if (reuse_seed)
        {
           seed = random_seed[run];
        }
       else
        {
           if (run > 0)
            {
               seed = time(NULL);
            }
        }
       std::cout << "Used random number seed: " << seed << std::endl;
       srand(seed);

       outfile_stream_seed << unsigned(seed) << std::endl;
       std::cout << "printing seed : " << seed << std::endl;
       // end Kit's edit  - random seed

       // seed generator for distributed parameters
       generator.seed(seed + 1);

       // use mu and sigma parameters to make lognormal distribution for incubation period
       double incubation_period;
       double recovery_period;


       std::lognormal_distribution<double> incubation_period_dist(incubation_period_mu, incubation_period_sig);
       std::uniform_int_distribution<int>  recovery_period_dist(recovery_period_low, recovery_period_hi);

       // set disease parameters for each agent
       double incubation_period_avg = 0;
       double recovery_period_avg   = 0;
       int    n_agents              = 0;
       std::cout << "setting disease parameters for agents..." << std::endl;
       for (auto agent_entry : agents)
        {
           n_agents              += 1;
           incubation_period      = incubation_period_dist(generator);
           incubation_period_avg += incubation_period;

           recovery_period      = recovery_period_dist(generator);
           recovery_period_avg += recovery_period;

           agent_entry.second->setLATENT_PERIOD_((unsigned)latent_period * 2);

           agent_entry.second->setPEAK_VIRAL_TITER_((unsigned)(incubation_period * 2 + latent_period * 2));

           //agent_entry.second->setINFECTIOUS_PERIOD_((unsigned)(incubation_period * 2 + recovery_period * 2 + latent_period * 2));

           //agent_entry.second->setRECOVERY_TIME_((unsigned)(incubation_period * 2 + recovery_period * 2 + latent_period * 2));

           agent_entry.second->setINFECTION_PERIOD_((unsigned)(incubation_period * 2 + recovery_period * 2 + latent_period * 2));

           agent_entry.second->setPERCENT_SYMPTOMATIC(adult_symptomatic_fraction);

           agent_entry.second->setCHILD_PERCENT_SYMPTOMATIC(child_symptomatic_fraction);

           agent_entry.second->setASYMPTOMATIC_INFECTIVITY_(asymptomatic_infectivity);

           if (agent_entry.first == 0)
            {
               agent_entry.second->printTiterVsTime();
            }
           //CZ COV19 25032020 - added a SD and CI state reset to init() in agent.cpp
        }

       incubation_period_avg /= n_agents;
       recovery_period_avg   /= n_agents;

       std::cout << "average incubation period for run " << run << " : " << incubation_period_avg << std::endl;
       std::cout << "average recovery period for run " << run << " : " << recovery_period_avg << std::endl;

       // Rezero the SLA prevalence
       for (auto& ccd : census_districts)
        {
           ccd.second->rezero_n_infecteds_tot();
        }

       //unsigned long n_airport_illnesses = 0;
       // this counts the number of symptomatic imporations
       // however, they are not symptomatic at the time of importation
       // for efficiency, they are counted upon imporation, not upon symptom expression

       unsigned long n_airport_infections = 0;

       // CZ: zero cumulative global variables
       for (auto& cg : g_cumulative_infected_in_group)
        {
           cg = 0;
        }

       // CZ: zeroing g_sla_attack_rate_map for the run
       for (auto sla_entry : statistical_local_areas)
        {
           g_sla_attack_rate_map.at(sla_entry.first) = 0;
        }

       g_detected_cumulative = 0;
       g_illness_cumulative  = 0;
       g_infected_cumulative = 0;

       char buffer[26];
       gettimeofday(&tv, NULL);
       tm_info = localtime(&tv.tv_sec);
       strftime(buffer, 26, "%Y-%m-%d_%H-%M-%S", tm_info);
       std::stringstream run_savefile,
                         run_savefile_incidence,
                         run_savefile_prevalence,
                         run_savefile_population,
                         run_savefile_breakdown,
                         run_savefile_CI_HQ, /* Kit's edit -intervention */
                         run_savefile_infection_I_ag,
                         run_savefile_illness_I_ag,
                         run_savefile_detection_I_ag,
                         run_savefile_infection_P_ag,
                         run_savefile_illness_P_ag,
                         run_savefile_detection_P_ag,
                         run_savefile_P_vaccinations,
                         run_savefile_G_vaccinations,
                         run_savefile_infection_I_VACge_ag,
                         run_savefile_infection_I_VACpr_ag,
                         run_savefile_detection_I_VACge_ag,
                         run_savefile_detection_I_VACpr_ag,
                         run_savefile_breakthrough;

       run_savefile << base_dir << "/R" << run << "_D" << buffer;
       run_savefile_breakdown << run_savefile.str() << "/Iinf_Pinf_Iill_Pill_Idet_Pdet_CDI_CDP.dat";

       run_savefile_incidence << run_savefile.str() << "/I_SLA";
       run_savefile_prevalence << run_savefile.str() << "/P_SLA";
       run_savefile_CI_HQ << run_savefile.str() << "/CI_HQ.dat";
       run_savefile_infection_I_ag << run_savefile.str() << "/I_infection_Ag.dat";
       run_savefile_illness_I_ag << run_savefile.str() << "/I_illness_Ag.dat";
       run_savefile_detection_I_ag << run_savefile.str() << "/I_detection_Ag.dat";
       run_savefile_infection_P_ag << run_savefile.str() << "/P_infection_Ag.dat";
       run_savefile_illness_P_ag << run_savefile.str() << "/P_illness_Ag.dat";
       run_savefile_detection_P_ag << run_savefile.str() << "/P_detection_Ag.dat";
       run_savefile_P_vaccinations << run_savefile.str() << "/Vaccinations_priority_Ag.dat";
       run_savefile_G_vaccinations << run_savefile.str() << "/Vaccinations_general_Ag.dat";

       run_savefile_detection_I_VACge_ag << run_savefile.str() << "/I_detection_VACge_Ag.dat";
       run_savefile_detection_I_VACpr_ag << run_savefile.str() << "/I_detection_VACpr_Ag.dat";
       run_savefile_infection_I_VACge_ag << run_savefile.str() << "/I_infection_VACge_Ag.dat";
       run_savefile_infection_I_VACpr_ag << run_savefile.str() << "/I_infection_VACpr_Ag.dat";

       run_savefile_breakthrough << run_savefile.str() << "/IinfGE_IinfPR_IdetGE_IdetPR.dat";

       if (makeDir(run_savefile.str()) != 0)
        {
           std::cerr << "Making directory: " << run_savefile.str() << " failed. Exiting." << std::endl;
           return(1);
        }
       std::cout << "Created folder " << run_savefile.str() << " for results." << std::endl;

       run_savefile << "/IP.dat";
       std::ofstream outfile_stream(run_savefile.str());
       std::ofstream outfile_stream_breakdown(run_savefile_breakdown.str());
       std::ofstream outfile_stream_incidence(run_savefile_incidence.str() + ".dat");
       std::ofstream outfile_stream_prevalence(run_savefile_prevalence.str() + ".dat");
       //CZ COV19 output for HQ and CI numbers
       std::ofstream outfile_stream_CI_HQ(run_savefile_CI_HQ.str());
       //CZ COV19 v7.6, outputs for infection, illness, and detection by age
       std::ofstream outfile_stream_infection_I_ag(run_savefile_infection_I_ag.str());
       std::ofstream outfile_stream_illness_I_ag(run_savefile_illness_I_ag.str());
       std::ofstream outfile_stream_detection_I_ag(run_savefile_detection_I_ag.str());
       std::ofstream outfile_stream_infection_P_ag(run_savefile_infection_P_ag.str());
       std::ofstream outfile_stream_illness_P_ag(run_savefile_illness_P_ag.str());
       std::ofstream outfile_stream_detection_P_ag(run_savefile_detection_P_ag.str());

       std::ofstream outfile_stream_detection_I_VACge_ag (run_savefile_detection_I_VACge_ag.str());
       std::ofstream outfile_stream_detection_I_VACpr_ag (run_savefile_detection_I_VACpr_ag.str());

       std::ofstream outfile_stream_infection_I_VACge_ag (run_savefile_infection_I_VACge_ag.str());
       std::ofstream outfile_stream_infection_I_VACpr_ag (run_savefile_infection_I_VACpr_ag.str());

       std::ofstream outfile_stream_breakthrough(run_savefile_breakthrough.str());


       //CZ: outputs for vaccination
       std::ofstream outfile_stream_P_vaccination(run_savefile_P_vaccinations.str());
       std::ofstream outfile_stream_G_vaccination(run_savefile_G_vaccinations.str());


       std::map<AgeGroup, std::shared_ptr<std::ofstream> > ag_outfile_stream_incidence;
       std::map<AgeGroup, std::shared_ptr<std::ofstream> > ag_outfile_stream_prevalence;
       for (AgeGroup ag = AgeGroups::Age0to4; ag <= AgeGroups::Age65Plus; ag = AgeGroup(ag + 1))
        {
           std::string inc_fname(run_savefile_incidence.str() + "-" + std::to_string(ag) + ".dat");
           std::string prev_fname(run_savefile_prevalence.str() + "-" + std::to_string(ag) + ".dat");
           ag_outfile_stream_incidence.emplace(ag, std::make_shared<std::ofstream>(inc_fname));
           ag_outfile_stream_prevalence.emplace(ag, std::make_shared<std::ofstream>(prev_fname));
        }

       /* Randomly select SLA to infect */
       std::vector<std::shared_ptr<StatisticalLocalArea> > seedSLA;
       std::vector<unsigned>                               num_agent_to_be_infected;

       if (seed_type == SeedTypes::NumberOfSLAs)
        {
           auto tmp_sla_list(statistical_local_areas);
           for (unsigned i = 0; i < n_slas_to_infect; ++i)
            {
               unsigned long num_agents_in_sla;
               auto          sla_it = tmp_sla_list.begin();

               unsigned long rval;
               rval   = rand() % tmp_sla_list.size();
               sla_it = tmp_sla_list.begin();
               std::advance(sla_it, rval);
               num_agents_in_sla = sla_it->second->getAgentIdxs().size();

               seedSLA.push_back(sla_it->second);
               num_agent_to_be_infected.push_back(p_agents_to_infect_in_sla * double(num_agents_in_sla));
               configfile_stream << seedSLA.back()->getID() << "," << num_agent_to_be_infected.back() << std::endl;
               tmp_sla_list.erase(sla_it);
            }
        }

       /* reset starting configuration */
       std::cout << "Initialising all agents..." << std::endl;
       for (auto agent_entry : agents)
        {
           agent_entry.second->setVacState(VacState(VacStates::notvaccinated));
           agent_entry.second->setAVState(AVState(AVStates::Notexposed));
           agent_entry.second->reset_intervention();
           agent_entry.second->init(); //Kit's edit -intervention
           //CZ COV19 25032020 - added a SD and CI state reset to init() in agent.cpp
        }

       // CZ COV19 01042020 - compute compliance for each measure, then compute overall compliance states
       // in set_Compliance_states_all()
       if (SD_intervention)
        {
           assign_SD_compliance(agents);
           std::cout << "Assigning social distancing compliance to agents" << std::endl;
        }

       if (CI_intervention)
        {
           assign_CI_compliance(agents);
           std::cout << "assigning case isolation compliance to agents at rate: " << compliance_rate_CI << std::endl;
           if (HQ_intervention)
            {
               assign_HQ_compliance(agents);
               std::cout << "done assigning CI compliance, now assigning HQ compliance at rate: " << compliance_rate_HQ << std::endl;
            }
        }

       if (School_closure)
        {
           assign_SC_compliance_students_teachers(agents);
           std::cout << "Assigning school closure compliance to students and teachers (parents comply dynamically)" << std::endl;
        }

       if (HQ_intervention && (!CI_intervention))
        {
           std::cout << "WARNING: Home quarantine has not been implemented, it is conditional on Case Isolation" << std::endl;
        }

       set_Compliance_states_all(agents);



       // beginning Kit's edit -intervention
       /* vaccinating the population */
       // displaying the intervention inputs

       // make lists of unvaccinated agents in each age bracket
       std::vector<unsigned long> unvaccinated_child_IDs;
       std::vector<unsigned long> unvaccinated_adult_IDs;
       std::vector<unsigned long> unvaccinated_senior_IDs;
       std::stack<unsigned long>  unvaccinated_child_IDs_s;
       std::stack<unsigned long>  unvaccinated_adult_IDs_s;
       std::stack<unsigned long>  unvaccinated_senior_IDs_s;

       if (pharm_intervention)
        {
           std::cout << "\n==Pharmaceutical Intervention parameters=="
                     << "\nIntervention update version: 09.01.2021"
                     << "\nVaccination:"
                     << "\n\tNumber of people pre-vaccinated (priority): " << n_vac_pre_epidemic_1
                     << "\n\tNumber of people vaccinated per week (priority): " << n_vac_per_week_1
                     << "\n\tNumber of people pre-vaccinated (general): " << n_vac_pre_epidemic_2
                     << "\n\tNumber of people vaccinated per week (general): " << n_vac_per_week_2
                     << "\n\tproportion of vaccines to 65+ agegroup: " << p_vac_65_and_over
                     << "\n\tproportion of vaccines to 18 to 64 agegroups: " << p_vac_18_to_64
                     << "\n\tproportion of vaccines to under 18 agegroups: " << p_vac_children
                     << "\n\tVaccine efficacy for susceptibility, priority (VEs_1): " << VEs_1
                     << "\n\tVaccine efficacy for infectiousness, priority (VEi_1): " << VEi_1
                     << "\n\tVaccine efficacy for disease, priority (VEd_1): " << VEd_1
                     << "\n\tVaccine efficacy for susceptibility, general (VEs_2): " << VEs_2
                     << "\n\tVaccine efficacy for infectiousness, general (VEi_2): " << VEi_2
                     << "\n\tVaccine efficacy for disease, general (VEd_2): " << VEd_2
                     << "\nTargeted Antivirals:"
                     << "\n\tProportion of Targeted anti-viral prophylaxis, (TAP): " << TAP
                     << "\n\tProportion of Geographically TAP, (GTAP): " << GTAP
                     << "\n\tIntervention response delay in days, (Intervention_delay): " << Response_delay
                     << "\n\tAnti-viral efficacy for susceptibility, (AVEs): " << AVEs
                     << "\n\tAnti-viral efficacy for infectiousness, (AVEi): " << AVEi
                     << "\n\tAnti-viral efficacy for illness, (AVEd): " << AVEd
                     << std::endl;
           if (vac_SLA_bool)
            {
               std::cout << "\tVaccinating only in spcified SLAs" << std::endl;
            }
           else
            {
               std::cout << "... Vaccinating in all SLAs ..." << std::endl;
            }

           //int num_vac = 0;


           // CZ: this is where people are vaccinated in Kit's implementation
           // CZCOVID-19 v7.5
           // to implement vaccine efficacy for disease, we need to modify the agent class
           // and we need to modify the infect() function to include vaccine efficacy for disease
           // query vaccine state, just like AVEd is used currently.

           // CZ: modified age-stratified pre-vaccination
           // zeroing the vaccine buckets
           num_vac          = 0;
           num_vac_65_plus  = 0;
           num_vac_18_to_64 = 0;
           num_vac_children = 0;

           num_vac_PV          = 0;
           num_vac_65_plus_PV  = 0;
           num_vac_18_to_64_PV = 0;
           num_vac_children_PV = 0;

           num_vac_GV          = 0;
           num_vac_65_plus_GV  = 0;
           num_vac_18_to_64_GV = 0;
           num_vac_children_GV = 0;


           for (auto agent_entry : agents)
            {
               auto c_agent = agent_entry.second;

               if (c_agent->getAgeGroup() == AgeGroups::Age65Plus)
                {
                   unvaccinated_senior_IDs.push_back(c_agent->getID());
                }
               else if (c_agent->getAgeGroup() == AgeGroups::Age0to4 ||
                        c_agent->getAgeGroup() == AgeGroups::Age5to18)
                {
                   unvaccinated_child_IDs.push_back(c_agent->getID());
                }
               else
                {
                   unvaccinated_adult_IDs.push_back(c_agent->getID());
                }
            }

           unsigned shuf_seed = 0;

           while (shuf_seed == 0)
            {
               shuf_seed = rand();
            }

           std::shuffle(unvaccinated_senior_IDs.begin(), unvaccinated_senior_IDs.end(), std::default_random_engine(shuf_seed));
           std::shuffle(unvaccinated_adult_IDs.begin(), unvaccinated_adult_IDs.end(), std::default_random_engine(shuf_seed + (unsigned)rand()));
           std::shuffle(unvaccinated_child_IDs.begin(), unvaccinated_child_IDs.end(), std::default_random_engine(shuf_seed + (unsigned)rand()));

           for (int i = 0; i < unvaccinated_senior_IDs.size(); i++)
            {
               unvaccinated_senior_IDs_s.push(unvaccinated_senior_IDs.at(i));
            }
           for (int i = 0; i < unvaccinated_adult_IDs.size(); i++)
            {
               unvaccinated_adult_IDs_s.push(unvaccinated_adult_IDs.at(i));
            }
           for (int i = 0; i < unvaccinated_child_IDs.size(); i++)
            {
               unvaccinated_child_IDs_s.push(unvaccinated_child_IDs.at(i));
            }

           std::cout << "there are " << unvaccinated_senior_IDs_s.size() << " unvaccinated agents 65+" << std::endl;
           std::cout << "there are " << unvaccinated_adult_IDs_s.size() << " unvaccinated agents 18-64" << std::endl;
           std::cout << "there are " << unvaccinated_child_IDs_s.size() << " unvaccinated agents under 18" << std::endl;

// CZ: this should be random - otherwise there might be some kind of bias introduced.
           pre_epidemic_vaccination(agents,
                                    unvaccinated_senior_IDs_s,
                                    unvaccinated_adult_IDs_s,
                                    unvaccinated_child_IDs_s);
        }

       if (SD_intervention || CI_intervention || School_closure)
        {
           std::cout << "\n==Social Distancing Intervention Parameters=="
                     << "\nIntervention update version: 06.12.2020"
                     << "\n"
                     << "\n" << "----Social Distancing for COV19----"
                     << "\n" << "Social distancing on? [0=no, 1=yes]: " << SD_intervention
                     << "\n" << "cumulative attack rate threshold: " << n_detected_SD_trigger
                     << "\n" << "SD compliance rate: " << compliance_rate_SD
                     << "\n" << "duration of social distancing: " << duration_SD
                     << "\n" << "Factor of social distancing for FoI in community: " << FoI_SD_community
                     << "\n" << "Factor of social distancing for FoI at work and school: " << FoI_SD_work
                     << "\n" << "Factor of social distancing for FoI at home: " << FoI_SD_home
                     << "\n"
                     << "\n" << "---School Closure for COV19---"
                     << "\n" << "School closure on? [0=no, 1=yes]: " << School_closure
                     << "\n" << "cumulative attack rate threshold: " << n_detected_SC_trigger
                     << "\n" << "duration of school closure: " << duration_SC
                     << "\n" << "Factor of school closure for FoI in community: " << FoI_SC_community
                     << "\n" << "Factor of school closure for FoI at work and school: " << FoI_SC_work
                     << "\n" << "Factor of school closure for FoI at home: " << FoI_SC_home
                     << "\n" << "Chance that a parent stays home (per student): " << p_parent_stays_home
                     << "\n"
                     << "\n" << "----Case Isolation for COV19----"
                     << "\n" << "Case Isolation on? [0=no, 1=yes]: " << CI_intervention
                     << "\n" << "CI compliance rate: " << compliance_rate_CI
                     << "\n" << "Factor of case isolation for FoI in community: " << FoI_CI_community
                     << "\n" << "Factor of case isolation for FoI at work and school: " << FoI_CI_work
                     << "\n" << "Factor of case isolation for FoI at home: " << FoI_CI_home
                     << "\n"
                     << "\n" << "----Home Quarantine for COV19----"
                     << "\n" << "Home Quarantine on? [0=no, 1=yes]: " << HQ_intervention
                     << "\n" << "HQ compliance rate: " << compliance_rate_HQ
                     << "\n" << "Quarantine duration (days): " << duration_HQ
                     << "\n" << "Factor of home quarantine for FoI in community: " << FoI_HQ_community
                     << "\n" << "Factor of home quarantine for FoI at work and school: " << FoI_HQ_work
                     << "\n" << "Factor of home quarantine for FoI at home: " << FoI_HQ_home
                     << std::endl;
        }


       if (seed_type == SeedTypes::NumberOfSLAs)  // incidence will not be correctly calculated for SLA seeding
        {
           /* Seed selected SLAs with infectious people */
           std::cout << "Seeding " << seedSLA.size() << " SLAs with infections..." << std::endl;
           for (unsigned i = 0; i < seedSLA.size(); ++i)
            {
               std::cout << "Infecting " << num_agent_to_be_infected.at(i)
                         << " people in SLA:" << seedSLA.at(i)->getID() << "." << std::endl;
               Epidemic::seed_epi_sla(num_agent_to_be_infected.at(i), *(seedSLA.at(i)), agents, AVEd, VEd_1, VEd_2);
            }
        }
       else if (seed_type == SeedTypes::PopulationRandom)
        {
           Epidemic::seed_epi_agents(p_pop_to_infect, agents, AVEd, VEd_1, VEd_2);        // incidence will not be correctly calculated for random seeding
        }

       g_infected_prevalence = getNumAgentsInfected(agents);
       g_illness_prevalence  = getNumAgentsIll(agents);

       std::cout << "\nStarting simulation (max days="
                 << n_days << " and "
                 << g_infected_prevalence << "/" << agents.size()
                 << " agents already infected,\n"
                 << g_illness_prevalence << "/" << agents.size()
                 << " agents already symptomatic" << std::endl;




       // initialize the weekday counter
       unsigned day_of_week = 0;
       bool     weekend     = false;

       for (unsigned day = 1;
            day < n_days && (day < 10 || (g_infected_prevalence > 0 && g_number_removed < agents.size()));
            ++day)
        {
           std::cout << "\t=== Start of day " << day << " ===" << std::endl;

           std::cout << "**********RNG CHECK: " << rand() << std::endl;

           // CZ: 7.2 figure out if it's a weekend
           day_of_week += 1;
           weekend      = false;
           if (day_of_week >= 8)
            {
               day_of_week = 0;
            }
           else
            {
               if (day_of_week == 6 || day_of_week == 7)
                {
                   weekend = true;
                }
            }

           if (weekend)
            {
               std::cout << "it's the weekend" << std::endl;
            }
           else
            {
               std::cout << "it's a weekday" << std::endl;
            }



           /* Zero the buckets (seems like this could be done quicker) */
           reZero(households, household_clusters, census_districts, destination_zones,
                  working_groups, schools, grades, classes, statistical_local_areas);


           if ((seed_type == SeedTypes::Airport) && ((g_detected_cumulative < 20) || !SD_intervention))
            {
               /* Seed at SLAs around international airports */
               std::vector<unsigned long> infected_agent_ids;
               std::cout << "... Seeding via international airports ..." << std::endl;
               Epidemic::seed_epi_airports(agents, infected_agent_ids, AVEd, VEd_1, VEd_2);

               for (auto& agent_id : infected_agent_ids)
                {
                   auto c_agent = agents.at(agent_id);

                   n_airport_infections++;
                   g_infected_incidence++;

                   g_ag_infection_incidence_map.at(c_agent->getAgeGroup())++;

                   if (c_agent->getVacState()==VacStates::vaccinated_priority)
                   {
                     g_ag_infection_incidence_VACpr_map.at(c_agent->getAgeGroup())++;
                    }
                    else if (c_agent->getVacState() == VacStates::vaccinated_general)
                    {
                      g_ag_infection_incidence_VACge_map.at(c_agent->getAgeGroup())++;
                    }


                   // if (agents.at(agent_id)->willBeSymptomatic())
                   //  {
                   //     n_airport_illnesses++;
                   //  }
                }
               std::cout << "\tInfected " << infected_agent_ids.size() << " agents in SLAs near airports/urban centres." << std::endl;
            }

           if (!weekend)
            {
               std::cout << "it's not a weekend so we're simulating daytime then nighttime" << std::endl;

               /* Update probabilities and infect agents during the first part of the day (at work/school) */

               if (SD_intervention || CI_intervention || School_closure)
                {
                   // CZ: if prevalence threshold was reached at the end of the previous day,
                   // trigger Social Distancing regimen
                   if (SD_timer == 1)
                    {
                       std::cout << "SD trigger reached." << std::endl;
                    }

                   if (SC_timer == 1)
                    {
                       std::cout << "SC trigger reached. Children, Teachers, and some parents will stay home." << std::endl;
                    }

                   if (SD_trigger)
                    {
                       SD_timer += 1;
                       std::cout << "SD applied, timer: " << SD_timer << std::endl;
                    }

                   if (SC_trigger)
                    {
                       assign_SC_compliance_parents(agents, p_parent_stays_home);
                       SC_timer += 1;
                       std::cout << "SC applied, timer: " << SC_timer << std::endl;
                    }


                   std::cout << "... updating infection and detection status ..." << std::endl;
                   step_and_detect(agents);
                   std::cout << "... applying/updating social distancing measures for daytime ..." << std::endl;
                   update_Isolation_states(agents, SD_trigger, CI_intervention, HQ_intervention, SC_trigger); // this function assesses compliance and intervention conditions for each agent
                   set_Isolation_Factors_work(agents, FoI_SD_work, FoI_CI_work, FoI_HQ_work, FoI_SC_work);
                   std::cout << "... simulating agents during the day period ..." << std::endl;
                   simulateDaytime_SD(agents);

                   // simulate daytime tells us which ill agents have been detected, so it must run before updating HQ states

                   std::cout << "... updating infection status ..." << std::endl;
                   step_without_detection(agents);
                   std::cout << "... applying/updating social distancing measures for nighttime ..." << std::endl;
                   update_Isolation_states(agents, SD_trigger, CI_intervention, HQ_intervention, SC_trigger);
                   set_Isolation_Factors_home(agents, FoI_SD_home, FoI_CI_home, FoI_HQ_home, FoI_SC_home); // translates isolation states into the appropriate FoI factors.
                   set_Isolation_Factors_community(agents, FoI_SD_community, FoI_CI_community, FoI_HQ_community, FoI_SC_community);
                   std::cout << "... simulating agents during the night period ..." << std::endl;
                   simulateNighttime_SD(agents);

                   if (HQ_intervention)
                    {
                       std::cout << "updating home quarantine clocks" << std::endl;
                       update_HQ(agents);
                    }                                   // advances clock for quarantined agents. turns quarantine off if duration is exceeded
                }

               else
                {
                   std::cout << "... simulating agents during the day period ..." << std::endl;
                   simulateDaytime(agents);
                   // the "step" function is integrated into this version of simulate daytime/nighttime

                   /* Update probabilities and infect agents during the first part of the day (at home) */
                   std::cout << "... simulating agents during the night period ..." << std::endl;

                   simulateNighttime(agents);
                }
            }
           else
            {
               std::cout << "it is a weekend so we're essentially simulating nighttime twice" << std::endl;
               /* Update probabilities and infect agents during the first part of the day (at work/school) */

               if (SD_intervention || CI_intervention || School_closure)
                {
                   // CZ: if prevalence threshold was reached at the end of the previous day,
                   // trigger Social Distancing regimen
                   if (SD_timer == 1)
                    {
                       std::cout << "SD trigger reached." << std::endl;
                    }

                   if (SC_timer == 1)
                    {
                       std::cout << "SC trigger reached. Children, Teachers, and some parents will stay home." << std::endl;
                    }

                   if (SD_trigger)
                    {
                       SD_timer += 1;
                       std::cout << "SD applied, timer: " << SD_timer << std::endl;
                    }

                   if (SC_trigger)
                    {
                       assign_SC_compliance_parents(agents, p_parent_stays_home);
                       SC_timer += 1;
                       std::cout << "SC applied, timer: " << SC_timer << std::endl;
                    }


                   std::cout << "... updating infection and detection status ..." << std::endl;
                   step_and_detect(agents);                                                                   // it's still "daytime" with respect to detection, but "nighttime" with respect to contact patterns
                   std::cout << "... applying/updating social distancing measures for daytime (weekend) ..." << std::endl;
                   update_Isolation_states(agents, SD_trigger, CI_intervention, HQ_intervention, SC_trigger); // this function assesses compliance and intervention conditions for each agent
                   //set_Isolation_Factors_work(agents, FoI_SD_work, FoI_CI_work, FoI_HQ_work, FoI_SC_work);
                   set_Isolation_Factors_home(agents, FoI_SD_home, FoI_CI_home, FoI_HQ_home, FoI_SC_home);    // translates isolation states into the appropriate FoI factors.
                   set_Isolation_Factors_community(agents, FoI_SD_community, FoI_CI_community, FoI_HQ_community, FoI_SC_community);
                   std::cout << "... simulating agents during the daytime (weekend) period ..." << std::endl;
                   simulateDaytimeWeekend_SD(agents);

                   // simulate daytime tells us which ill agents have been detected, so it must run before updating HQ states

                   std::cout << "... updating infection status ..." << std::endl;
                   step_without_detection(agents);
                   std::cout << "... applying/updating social distancing measures for nighttime ..." << std::endl;
                   update_Isolation_states(agents, SD_trigger, CI_intervention, HQ_intervention, SC_trigger);
                   set_Isolation_Factors_home(agents, FoI_SD_home, FoI_CI_home, FoI_HQ_home, FoI_SC_home); // translates isolation states into the appropriate FoI factors.
                   set_Isolation_Factors_community(agents, FoI_SD_community, FoI_CI_community, FoI_HQ_community, FoI_SC_community);
                   std::cout << "... simulating agents during the night period ..." << std::endl;
                   simulateNighttime_SD(agents);

                   if (HQ_intervention)
                    {
                       std::cout << "updating home quarantine clocks" << std::endl;
                       update_HQ(agents);
                    }                          // advances clock for quarantined agents. turns quarantine off if duration is exceeded
                }

               else
                {
                   std::cout << "... simulating agents during the day (weekend) period ..." << std::endl;
                   simulateDaytimeWeekend(agents);

                   /* Update probabilities and infect agents during the first part of the day (at home) */
                   std::cout << "... simulating agents during the night period ..." << std::endl;

                   simulateNighttime(agents);
                }
            }



           g_detected_cumulative += g_detected_incidence;
           g_illness_cumulative  += g_illness_incidence;
           g_infected_cumulative += g_infected_incidence;

           // CZ COV19 SD 17032020 - SD trigger to 2000
           if ((g_detected_cumulative >= n_detected_SD_trigger) && !SD_trigger && SD_intervention && (SD_timer < duration_SD + 1))
            {
               SD_trigger = true;
               SD_timer   = 1;
            }

           // CZ COV19 SD 17032020 - SD trigger to 2000
           if ((g_detected_cumulative >= n_detected_SC_trigger) && !SC_trigger && School_closure && (SC_timer < duration_SC + 1))
            {
               SC_trigger = true;
               SC_timer   = 1;
            }

           if (SD_timer == duration_SD + 1)
            {
               SD_trigger = false;
               std::cout << "SD duration exceeded, stopping social distancing" << std::endl;
            }

           if (SC_timer == duration_SC + 1)
            {
               SC_trigger = false;
               std::cout << "SC duration exceeded, re-opening schools" << std::endl;
            }

           num_CD_incidence  = count_CD_incidence(census_districts);
           num_CD_prevalence = count_CD_prevalence(census_districts);

           std::vector<double> perc_infected_in_group;

           unsigned long       tot_in_groups = 0;

           //double              cumulative_ill_without_ap = double(g_cumulative_ill_check - n_airport_illnesses);
           double cumulative_infected_without_ap = double(g_infected_cumulative - n_airport_infections);
           // CZ: cumulative_ill_check is no longer useful since we're counting some asymptomatic agents.
           for (auto cg : g_cumulative_infected_in_group)
            {
               perc_infected_in_group.push_back(double(cg) / cumulative_infected_without_ap);
               tot_in_groups += cg;
            }

           std::cout << " computing child symptomatic rate " << std::endl;

           double child_symptomatic_fraction =
              double(g_ag_illness_prevalence_map.at(AgeGroups::Age0to4) +
                     g_ag_illness_prevalence_map.at(AgeGroups::Age5to18)) /
              double(g_ag_infection_prevalence_map.at(AgeGroups::Age0to4) +
                     g_ag_infection_prevalence_map.at(AgeGroups::Age5to18));

           std::cout << " computing adult symptomatic rate " << std::endl;

           double adult_symptomatic_fraction =
              double(g_ag_illness_prevalence_map.at(AgeGroups::Age19to29) +
                     g_ag_illness_prevalence_map.at(AgeGroups::Age30to64) +
                     g_ag_illness_prevalence_map.at(AgeGroups::Age65Plus)) /
              double(g_ag_infection_prevalence_map.at(AgeGroups::Age19to29) +
                     g_ag_infection_prevalence_map.at(AgeGroups::Age30to64) +
                     g_ag_infection_prevalence_map.at(AgeGroups::Age65Plus));

           std::cout << "... day analysis ..." << std::endl;
           std::cout << "\tDetection Incidence (Prop.): " << g_detected_incidence << " (" << double(g_detected_incidence) / double(agents.size()) << ")\n\t"
                     << "Infection Incidence (Prop.): " << g_infected_incidence << " (" << double(g_infected_incidence) / double(agents.size()) << ")\n\t"
                     << "Illness Incidence (Prop.): " << g_illness_incidence << " (" << double(g_illness_incidence) / double(agents.size()) << ")\n\t"
                     << "\n\t"
                     << "Detection Incidence - Vac. General: " << g_detected_incidence_VACge << "\n\t"
                     << "Detection Incidence - Vac. Priority: " << g_detected_incidence_VACpr << "\n\t"
                     << "Infection Incidence - Vac. General: " << g_infected_incidence_VACge << "\n\t"
                     << "Infection Incidence - Vac. Priority: " << g_infected_incidence_VACpr << "\n\t"
                     << "\n\t"
                     << "Detected Case Prevalence (Prop.): " << g_detected_prevalence << " (" << double( g_detected_prevalence ) / double( agents.size()) << ")\n\t"
                     << "Infection Prevalence (Prop.): " << g_infected_prevalence << " (" << double( g_infected_prevalence ) / double( agents.size()) << ")\n\t"
                     << "Illness Prevalence (Prop.): " << g_illness_prevalence << " (" << double(g_illness_prevalence) / double(agents.size()) << ")\n\t"
                     << "\n\t"
                     << "Cumulative Detected (Prop.): " << g_detected_cumulative << " (" << double( g_detected_cumulative ) / double( agents.size()) << ")\n\t"
                     << "Cumulative Infected (Prop.): " << g_infected_cumulative << " (" << double( g_infected_cumulative ) / double( agents.size()) << ")\n\t"
                     << "Cumulative Ill (Prop.): " << g_illness_cumulative << " (" << double( g_illness_cumulative ) / double( agents.size()) << ")\n\t"
                     << "Number removed (Prop.): " << g_number_removed << " (" << double( g_number_removed ) / double( agents.size()) << ")\n\t"
                     << "\n\t"
                     << "Symptomatic Fraction in Children: " << child_symptomatic_fraction << "\n\t"
                     << "Symptomatic Fraction in Adults: " << adult_symptomatic_fraction << "\n\t"
                     << "\n\t"
                     << "Per cent infected from group [HH, HC, CD, SLA, WG, CL, GR, SCH]: [";
           double perc_sum = 0.0;
           for (auto cp : perc_infected_in_group)
            {
               std::cout << cp * 100.0 << ", ";
               perc_sum += cp;
            }
           std::cout << "]" << std::endl;
           std::cout << "\tPercentage sum: " << perc_sum * 100.0 << std::endl;
           std::cout << "\tNumber of airport infections: " << n_airport_infections << std::endl;
           std::cout << "\tGroup sum diff: " << (long int)(tot_in_groups) - (long int)(g_infected_cumulative) + (long int)(n_airport_infections) << std::endl;

           /* Print incidence and prevalence for each SLA */
           outfile_stream << g_detected_incidence << "," << g_detected_prevalence << std::endl;

           outfile_stream_breakdown << g_infected_incidence
                                    << "," << g_infected_prevalence
                                    << "," << g_illness_incidence
                                    << "," << g_illness_prevalence
                                    << "," << g_detected_incidence
                                    << "," << g_detected_prevalence
                                    << "," << num_CD_incidence
                                    << "," << num_CD_prevalence
                                    << std::endl;

          outfile_stream_breakthrough << g_infected_incidence_VACge
                                      << "," << g_infected_incidence_VACpr
                                      << "," << g_detected_incidence_VACge
                                      << "," << g_detected_incidence_VACpr
                                      << std::endl;


          outfile_stream_P_vaccination << num_vac_PV
                                       << "," << num_vac_children_PV
                                       << "," << num_vac_18_to_64_PV
                                       << "," << num_vac_65_plus_PV
                                       << std::endl;


          outfile_stream_G_vaccination << num_vac_GV
                                       << "," << num_vac_children_GV
                                       << "," << num_vac_18_to_64_GV
                                       << "," << num_vac_65_plus_GV
                                       << std::endl;


           for (auto sla_entry : statistical_local_areas)
            {
               outfile_stream_incidence << double( g_sla_incidence_map.at(sla_entry.first)) / double( sla_entry.second->getAgentIdxs().size()) << " ";
               outfile_stream_prevalence << double( g_sla_prevalence_map.at(sla_entry.first)) / double( sla_entry.second->getAgentIdxs().size()) << " ";

               for (AgeGroup ag = AgeGroups::Age0to4; ag <= AgeGroups::Age65Plus; ag = AgeGroup(ag + 1))
                {
                   auto inc_outfile  = ag_outfile_stream_incidence.at(ag);
                   auto prev_outfile = ag_outfile_stream_prevalence.at(ag);

                   *inc_outfile << double( g_ag_sla_incidence_map.at(ag).at(sla_entry.first)) / double( sla_entry.second->getAgentIdxs().size()) << " ";
                   *prev_outfile << double( g_ag_sla_prevalence_map.at(ag).at(sla_entry.first)) / double( sla_entry.second->getAgentIdxs().size()) << " ";
                }
            }
           outfile_stream_incidence << std::endl;
           outfile_stream_prevalence << std::endl;

           for (AgeGroup ag = AgeGroups::Age0to4; ag <= AgeGroups::Age65Plus; ag = AgeGroup(ag + 1))
            {
               *ag_outfile_stream_incidence.at(ag) << std::endl;
               *ag_outfile_stream_prevalence.at(ag) << std::endl;

               outfile_stream_infection_I_ag << g_ag_infection_incidence_map.at(ag) << " ";
               outfile_stream_infection_P_ag << g_ag_infection_prevalence_map.at(ag) << " ";

               outfile_stream_illness_I_ag << g_ag_illness_incidence_map.at(ag) << " ";
               outfile_stream_illness_P_ag << g_ag_illness_prevalence_map.at(ag) << " ";

               outfile_stream_detection_I_ag << g_ag_detection_incidence_map.at(ag) << " ";
               outfile_stream_detection_P_ag << g_ag_detection_prevalence_map.at(ag) << " ";

               outfile_stream_detection_I_VACge_ag << g_ag_detection_incidence_VACge_map.at(ag) << " ";
               outfile_stream_detection_I_VACpr_ag << g_ag_detection_incidence_VACpr_map.at(ag) << " ";

               outfile_stream_infection_I_VACge_ag << g_ag_infection_incidence_VACge_map.at(ag) << " ";
               outfile_stream_infection_I_VACpr_ag << g_ag_infection_incidence_VACpr_map.at(ag) << " ";
            }

           outfile_stream_infection_I_ag << std::endl;
           outfile_stream_infection_P_ag << std::endl;
           outfile_stream_illness_I_ag << std::endl;
           outfile_stream_illness_P_ag << std::endl;
           outfile_stream_detection_I_ag << std::endl;
           outfile_stream_detection_P_ag << std::endl;
           outfile_stream_detection_I_VACge_ag << std::endl;
           outfile_stream_detection_I_VACpr_ag << std::endl;
           outfile_stream_infection_I_VACge_ag << std::endl;
           outfile_stream_infection_I_VACpr_ag << std::endl;

           /* beginning 's edit - intervention*/
           // Running the AV protocols
           if (pharm_intervention)
            {
               progressive_vaccination_priority(agents,
                                                unvaccinated_senior_IDs_s,
                                                unvaccinated_adult_IDs_s,
                                                unvaccinated_child_IDs_s);

               progressive_vaccination_general(agents,
                                               unvaccinated_senior_IDs_s,
                                               unvaccinated_adult_IDs_s,
                                               unvaccinated_child_IDs_s);

               if ((GTAP > 0) || (TAP > 0)) // This code is not designed to mix TAP and GTAP
                {
                   dynamic_intervention(agents);
                }

               //ount_vaccinated_agents(agents); // CZ: just a check.
            }
           /* end Kit's edit - intervention*/

           if (HQ_intervention || CI_intervention)
            {
               outfile_stream_CI_HQ << g_CI_agents << "\t" << g_HQ_agents << std::endl;
            }

           std::cout << "\tPercentage sum: " << perc_sum * 100.0 << "\n\t"
                     << "=== End of day " << day << " ===\n\t" << std::endl;
        }

       outfile_stream.close();
       outfile_stream_breakdown.close();
       outfile_stream_breakthrough.close();
       outfile_stream_CI_HQ.close();
       outfile_stream_infection_I_ag.close();
       outfile_stream_infection_P_ag.close();
       outfile_stream_illness_I_ag.close();
       outfile_stream_illness_P_ag.close();
       outfile_stream_detection_I_ag.close();
       outfile_stream_detection_P_ag.close();
       outfile_stream_P_vaccination.close();
       outfile_stream_G_vaccination.close();

       outfile_stream_detection_I_VACge_ag.close();
       outfile_stream_detection_I_VACpr_ag.close();
       outfile_stream_infection_I_VACge_ag.close();
       outfile_stream_infection_I_VACpr_ag.close();


       std::cout << "Time taken: %.2fs\n" << (double)(clock() - tStart) / CLOCKS_PER_SEC << std::endl;
    }
   outfile_stream_seed.close(); /* Kit's edit - random seed */



   return(0);
}


/* Parse argument inputs (code "borrowed" from https://www.gnu.org/software/libc/manual/html_node/Getopt-Long-Option-Example.html#Getopt-Long-Option-Example).
 *  Options are:
 *  - # runs
 *  - # days per run
 *  - # SLAs to infect with disease
 *  - # people in each SLA to infect
 *  - # Database containing census data/for results
 */
void parse_arguments(int argc, char *argv[],
   double& p_trans, double& trans_scaler,
   unsigned& n_runs, unsigned& n_days,
   unsigned& n_seeded_slas, double& p_seeded_agents, double& p_pop_seeded,
   std::string& database_output, std::string& database_input,
   bool& reuse_seed,
   std::string& SD_intervention_input_fname,
   std::string& Pharm_intervention_input_fname,
   std::string& Parameter_input_fname)                    /* Kit's edit - random seed - intervention; Cam's edit COV19 SD input file 17032020*/
{
   static struct option long_options[] =
   {
      /*
       * Following options set a flag.
       */
      //{"runs", no_argument, &verbose_flag, 1},
      //{"brief",   no_argument,       &verbose_flag, 0},

      /*
       * Following options don’t set a flag.
       * We distinguish them by their indices.
       */
      { "ptrans",                         no_argument, 0, 'p' },
      { "scaler",                         no_argument, 0, 't' },
      { "runs",                           no_argument, 0, 'r' },
      { "days",                           no_argument, 0, 'd' },
      { "slas",                           no_argument, 0, 's' },
      { "agents",                         no_argument, 0, 'a' },
      { "proportion",                     no_argument, 0, 'c' },
      { "database_output",                no_argument, 0, 'b' },
      { "database_input",                 no_argument, 0, 'f' },
      { "reuse_seed",                     no_argument, 0, 'u' }, /* Kit's edit - random seed */
      { "SD_intervention_input_fname",    no_argument, 0, 'q' },
      { "Pharm_intervention_input_fname", no_argument, 0, 'V' }, // CZ: added pharmaceutical intervention fname (vaccine and antiviral parameters, flag -V)
      { "Parameter_input_fname",          no_argument, 0, 'P' },
      {                                0,           0, 0,   0 }
   };

   for ( ; ;)
    {
       /* getopt_long stores the option index here. */
       int  option_index = 0;

       char c = getopt_long(argc, argv, "p:t:r:d:s:a:c:b:f:u:q:V:P:", long_options, &option_index); /* Kit's edit - random seed - intervention*/

       /* Options are all parsed */
       if (c == -1)
        {
           break;
        }

       switch (c)
        {
        case 'p':
           p_trans = std::stof(optarg);
           break;

        case 't':
           trans_scaler = std::stof(optarg);
           break;

        case 'r':
           n_runs = std::stoi(optarg);
           break;

        case 'd':
           n_days = std::stoi(optarg);
           break;

        case 's':
           n_seeded_slas = std::stoi(optarg);
           break;

        case 'a':
           p_seeded_agents = std::stof(optarg);
           break;

        case 'c':
           p_pop_seeded = std::stof(optarg);
           break;

        case 'b':
           database_output = std::string(optarg);
           break;

        case 'f':
           database_input = std::string(optarg);
           break;

        case 'u': //  Kit's edit - random seed
           reuse_seed = std::stoi(optarg);
           break;

        case 'q': //CZ: SD intervention inputs
           SD_intervention_input_fname = std::string(optarg);
           break;

        case 'V':    //CZ: SD intervention inputs
           Pharm_intervention_input_fname = std::string(optarg);
           break;

        case 'P':
           Parameter_input_fname = std::string(optarg);
           break;


        case '?':
           /* getopt_long already printed an error message. */
           break;

        default:
           abort();
        }
       std::cout << "option -" << c << " with value `" << optarg << "'" << std::endl;
    }

   /* Print any remaining command line arguments (not options). */
   if (optind < argc)
    {
       std::cout << "Non-option ARGV-elements: ";
       while (optind < argc)
        {
           std::cout << argv[optind++] << ", ";
        }
       std::cout << std::endl;
    }
}


int makeDir(const std::string& dir)
{
   int status = mkdir(dir.c_str(), S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);

   if (status != 0)
    {
       if (errno == EEXIST)
        {
           std::cout << "Folder " << dir << " already exists." << std::endl;
        }
       else
        {
           std::cerr << "Could not create folder " << dir << " for results (err:" << errno << ")." << std::endl;
        }
    }
   return(status);
}


/* For efficiency, these three should probably be combined or computed along the way instead */
unsigned long getNumAgentsRemoved(agent_map_type& agents)
{
   unsigned long num = 0;

   for (auto agent_entry : agents)
    {
       if (agent_entry.second->isRemoved())
        {
           num++;
        }
    }
   return(num);
}


unsigned long getNumAgentsRecovered(agent_map_type& agents)
{
   unsigned long num = 0;

   for (auto agent_entry : agents)
    {
       if (agent_entry.second->isInfectedPrior() && !agent_entry.second->isInfected())
        {
           num++;
        }
    }
   return(num);
}


unsigned long getNumAgentsIll(agent_map_type& agents)
{
   unsigned long num = 0;

   for (auto agent_entry : agents)
    {
       if (agent_entry.second->isIll())
        {
           num++;
        }
    }
   return(num);
}


unsigned long getNumAgentsInfected(agent_map_type& agents)
{
   unsigned long num = 0;

   for (auto agent_entry : agents)
    {
       if (agent_entry.second->isInfected())
        {
           num++;
        }
    }
   return(num);
}

// why doesn't this function need AVEd as an argument? must be declared in global scope...
void updateAgentState(std::shared_ptr<Agent> c_agent, double P, unsigned agroup)
{
   if (c_agent->isSusceptible())
    {
       // Probability of infection is moderated by a multiplier if the agent is either vaccinated or given AV
       double vac_multiplier  = 1.0;
       double AVEs_multiplier = 1.0;

       if ((c_agent->getAVState() == AVStates::AV_prophylaxis) || (c_agent->getAVState() == AVStates::AV_treatment)) // 28.02.19 // The AVP multiplier for susceptibility if the agent is recieving AVP
        {
           AVEs_multiplier = 1 - AVEs;
        }

       // CZ: use VacStates::vaccinated to apply VEs to susceptibility
       if (c_agent->getVacState() == VacStates::vaccinated_priority) // The vaccination multiplier for susceptibility if the agent has recieved a vaccination
        {
           vac_multiplier = 1 - VEs_1;
        }
        else if (c_agent->getVacState() == VacStates::vaccinated_general)
        {
          vac_multiplier = 1 - VEs_2;
        }

       P = P * vac_multiplier * AVEs_multiplier; // Probability of transmition with intervention multipliers

       if (rand() / (double)RAND_MAX < P)
        {
           c_agent->infect(AVEd, VEd_1, VEd_2);                  // AVEd and VEd are global variables...

           g_infected_incidence++;                      // counting the total infected incidence
           g_ag_infection_incidence_map.at(c_agent->getAgeGroup())++;

           if (c_agent->getVacState() == VacStates::vaccinated_general)
           {
             g_infected_incidence_VACge++;
             g_ag_infection_incidence_VACge_map.at(c_agent->getAgeGroup())++;
           }
           else if (c_agent->getVacState() == VacStates::vaccinated_priority)
           {
             g_infected_incidence_VACpr++;
             g_ag_infection_incidence_VACpr_map.at(c_agent->getAgeGroup())++;
           }

           g_cumulative_infected_in_group.at(agroup)++; // counting infection incidence in mixing groups
        }
    }
   else if (c_agent->isInfected())
    {
       if (c_agent->isNewIllness())
        {
           g_illness_incidence++;
           g_ag_illness_incidence_map.at(c_agent->getAgeGroup())++;
           c_agent->resetNewIllness();
        }

       if (c_agent->isNewlyDetected())  //includes all detected agents, regardless of symptom status
        {
           auto my_sla = c_agent->getSLA()->getID();
           g_detected_incidence++;
           g_ag_detection_incidence_map.at(c_agent->getAgeGroup())++;
           if (c_agent->getVacState() == VacStates::vaccinated_general)
           {
             g_detected_incidence_VACge++;
             g_ag_detection_incidence_VACge_map.at(c_agent->getAgeGroup())++;
           }
           else if (c_agent->getVacState() == VacStates::vaccinated_priority)
           {
             g_detected_incidence_VACpr++;
             g_ag_detection_incidence_VACpr_map.at(c_agent->getAgeGroup())++;
           }
           g_sla_incidence_map.at(my_sla)++;
           g_sla_attack_rate_map.at(my_sla)++;
           g_ag_sla_incidence_map.at(c_agent->getAgeGroup()).at(my_sla)++;
           c_agent->resetNewDetection();
        }
    }
}


void simulateDaytime(agent_map_type& agents)
{
   /* Update transmissibility at work */
   for (auto agent_entry : agents)
    {
       bool detected_agent = false;
       auto c_agent        = agent_entry.second;


       if (c_agent->isInfected())
        {
           c_agent->step();

           if ((c_agent->getState() == DiseaseStates::StateSymptomatic) && (c_agent->isIllnessDetected() == false))
           // If the agent is symptomatic and not detected check for detection
            {
               detected_agent = c_agent->Illness_detection(Detection, Response_delay, TAP, GTAP);
            }
           else
            {
               if ((c_agent->getState() == DiseaseStates::StateAsymptomatic) && (c_agent->isIllnessDetected() == false)) // if agent is asymptomatic
               // remember that the state 'asymptomatic' applies to both pre-symptomatic and asymptomatic individuals who are otherwise
               // differentiated by the sign of the incubating_period_ (this value is set to -1 for asymptomatic agents)
                {
                   detected_agent = c_agent->Illness_detection(Detection_asymp, Response_delay, TAP, GTAP);
                }
            }



           c_agent->addToDaytimeProbabilityOfInfection(VEi_1, VEi_2, AVEi);
        }

       // Does the agent require a new AV dose? happens once every 10 steps unless they are recovered
       int new_dose = 0;
       if ((c_agent->getAVState() == AVStates::ExposedAV) || (c_agent->getAVState() == AVStates::AV_prophylaxis) || (c_agent->getAVState() == AVStates::AV_treatment)) // 28.02.19
        {
           new_dose  = c_agent->stepAV(Response_delay);
           AV_doses += new_dose;
        }
    }

   // decide who is now infected
   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       if (c_agent->isStudent())
        {
           updateAgentState(c_agent, c_agent->getDaytimeProbabilityOfInfection(2), MixingGroups::GroupClass);
           updateAgentState(c_agent, c_agent->getDaytimeProbabilityOfInfection(1), MixingGroups::GroupGrade);
           updateAgentState(c_agent, c_agent->getDaytimeProbabilityOfInfection(0), MixingGroups::GroupSchool);
        }
       else
        {
           updateAgentState(c_agent, c_agent->getDaytimeProbabilityOfInfection(), MixingGroups::GroupWorkplace);
        }
    }
}


void simulateNighttime(agent_map_type& agents)
{
   /* Update distributions for home */

   //std::cout << "A1" << std::endl;

   int counter = 0;

   for (auto& agent_entry : agents)
    {
       auto c_agent = agent_entry.second;


       if (c_agent->isInfected())
        {
           c_agent->step();
           // note: no detection at night

           c_agent->addToNighttimeProbabilityOfInfection(VEi_1, VEi_2, AVEi);
           //std::cout << "done adding foi" << std::endl;
        }
    }

   //std::cout << "A2" << std::endl;
   /* Decide who is now infected */
   for (auto& agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       updateAgentState(c_agent, c_agent->getHouseholdProbabilityOfInfection(), MixingGroups::GroupHousehold);
       updateAgentState(c_agent, c_agent->getHouseholdClusterProbabilityOfInfection(), MixingGroups::GroupHouseholdCluster);
       updateAgentState(c_agent, c_agent->getNeighbourhoodProbabilityOfInfection(), MixingGroups::GroupNeighbourhood);
       updateAgentState(c_agent, c_agent->getCommunityProbabilityOfInfection(), MixingGroups::GroupCommunity);

       auto my_sla = c_agent->getSLA()->getID();

// bookkeeping
       if (!(c_agent->isSusceptible()))
        {
           if (c_agent->isRemoved())
            {
               g_number_removed++;
               g_sla_number_removed_map.at(my_sla)++;
               g_ag_sla_number_removed_map.at(c_agent->getAgeGroup()).at(my_sla)++;
            }
           else if (c_agent->isInfected())
            {
               g_infected_prevalence++;
               g_ag_infection_prevalence_map.at(c_agent->getAgeGroup())++;

               if (c_agent->isIll())
                {
                   g_illness_prevalence++;
                   g_ag_illness_prevalence_map.at(c_agent->getAgeGroup())++;
                }
               if (c_agent->isIllnessDetected())
                {
                   g_detected_prevalence++;
                   g_ag_detection_prevalence_map.at(c_agent->getAgeGroup())++;
                   g_sla_prevalence_map.at(my_sla)++;
                   g_ag_sla_prevalence_map.at(c_agent->getAgeGroup()).at(my_sla)++;
                }
            }
        }
    }
}


void simulateDaytimeWeekend(agent_map_type& agents)
{
   /* Update distributions for home */

   //std::cout << "A1" << std::endl;

   for (auto& agent_entry : agents)
    {
       bool detected_agent = false;
       auto c_agent        = agent_entry.second;


       if (c_agent->isInfected())
        {
           c_agent->step();

           // detection occurs in the daytime

           if ((c_agent->getState() == DiseaseStates::StateSymptomatic) && (c_agent->isIllnessDetected() == false)) // If the agent is symptomatic and not detected check for detection
            {
               detected_agent = c_agent->Illness_detection(Detection, Response_delay, TAP, GTAP);
            }
           else
            {
               if ((c_agent->getState() == DiseaseStates::StateAsymptomatic) && (c_agent->isIllnessDetected() == false)) // if agent is asymptomatic
               // remember that the state 'asymptomatic' applies to both pre-symptomatic and asymptomatic individuals who are otherwise
               // differentiated by the sign of the incubating_period_ (this value is set to -1 for asymptomatic agents)
                {
                   detected_agent = c_agent->Illness_detection(Detection_asymp, Response_delay, TAP, GTAP);
                }
            }

           // CZ: removing weekend 'daytime' infections as per Mikhail's suggestion
           //c_agent->addToNighttimeProbabilityOfInfection(VEi, AVEi, AVEd);
           //std::cout << "done adding foi" << std::endl;
        }
    }

   //std::cout << "A2" << std::endl;
   /* Decide who is now infected */
   // CZ: removing this loop to reduce the number of weekend infections
   // bookkeeping should be OK as long as updateAgentState is called during the nighttime cycle
   /*for (auto& agent_entry : agents)
    * {
    *  auto c_agent = agent_entry.second;
    *
    *  updateAgentState(c_agent, c_agent->getHouseholdProbabilityOfInfection(), MixingGroups::GroupHousehold);
    *  updateAgentState(c_agent, c_agent->getHouseholdClusterProbabilityOfInfection(), MixingGroups::GroupHouseholdCluster);
    *  updateAgentState(c_agent, c_agent->getNeighbourhoodProbabilityOfInfection(), MixingGroups::GroupNeighbourhood);
    *  updateAgentState(c_agent, c_agent->getCommunityProbabilityOfInfection(), MixingGroups::GroupCommunity);
    *
    * }*/
}





void step_and_detect(agent_map_type& agents)
{
   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       if (c_agent->isInfected())
        {
           c_agent->step(); // CZ: this will update HH infection status (i.e., for HQ) if agent has already been detected, Kit hardcoded this for his targeted quarantine implementation

           bool detected_agent = false;

           if ((c_agent->getState() == DiseaseStates::StateSymptomatic) && (c_agent->isIllnessDetected() == false)) // If the agent is symptomatic and not detected check for detection
            {
               detected_agent = c_agent->Illness_detection(Detection, Response_delay, TAP, GTAP);
            }
           else
            {
               if ((c_agent->getState() == DiseaseStates::StateAsymptomatic) && (c_agent->isIllnessDetected() == false))  // if agent is asymptomatic
               // remember that the state 'asymptomatic' applies to both pre-symptomatic and asymptomatic individuals who are otherwise
               // differentiated by the sign of the incubating_period_ (this value is set to -1 for asymptomatic agents)
                {
                   detected_agent = c_agent->Illness_detection(Detection_asymp, Response_delay, TAP, GTAP);
                }
            }
        }
    }
}


void step_without_detection(agent_map_type& agents)
{
   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       if (c_agent->isInfected())
        {
           c_agent->step();
        }
    }
}


//CZ COV19 17032020 - can be combined with vaccination and antivirals
void simulateDaytime_SD(agent_map_type& agents)
{
   /* Update transmissibility at work */
   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       if (c_agent->isInfected())
        {
           c_agent->addToDaytimeProbabilityOfInfection_SD(VEi_1, VEi_2, AVEi);
        }
    }

   /* Update susceptibility at work */
   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       if (c_agent->isStudent())
        {
           updateAgentState(c_agent, c_agent->getDaytimeProbabilityOfInfection(2) * c_agent->getIsolationFactor_work(), MixingGroups::GroupClass);
           updateAgentState(c_agent, c_agent->getDaytimeProbabilityOfInfection(1) * c_agent->getIsolationFactor_work(), MixingGroups::GroupGrade);
           updateAgentState(c_agent, c_agent->getDaytimeProbabilityOfInfection(0) * c_agent->getIsolationFactor_work(), MixingGroups::GroupSchool);
        }
       else
        {
           updateAgentState(c_agent, c_agent->getDaytimeProbabilityOfInfection() * c_agent->getIsolationFactor_work(), MixingGroups::GroupWorkplace);
        }
    }
}


//CZ COV19 17032020, updated 01042020
void simulateNighttime_SD(agent_map_type& agents)
{
   /* Update distributions for home */
   for (auto& agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       if (c_agent->isInfected())
        {
           c_agent->addToNighttimeProbabilityOfInfection_SD(VEi_1, VEi_2, AVEi); // CZ COV10 08042020 - removed agent->step() so isolation states can be updated before computing FoI
        }
    }
   /* Decide who is now infected */
   for (auto& agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       updateAgentState(c_agent, c_agent->getHouseholdProbabilityOfInfection() * c_agent->getIsolationFactor_home(), MixingGroups::GroupHousehold);
       updateAgentState(c_agent, c_agent->getHouseholdClusterProbabilityOfInfection() * c_agent->getIsolationFactor_community(), MixingGroups::GroupHouseholdCluster);
       updateAgentState(c_agent, c_agent->getNeighbourhoodProbabilityOfInfection() * c_agent->getIsolationFactor_community(), MixingGroups::GroupNeighbourhood);
       updateAgentState(c_agent, c_agent->getCommunityProbabilityOfInfection() * c_agent->getIsolationFactor_community(), MixingGroups::GroupCommunity);

       auto my_sla = c_agent->getSLA()->getID();


       // bookkeeping
       if (!(c_agent->isSusceptible()))
        {
           if (c_agent->isRemoved())
            {
               g_number_removed++;
               g_sla_number_removed_map.at(my_sla)++;
               g_ag_sla_number_removed_map.at(c_agent->getAgeGroup()).at(my_sla)++;
            }
           else if (c_agent->isInfected())
            {
               g_infected_prevalence++;
               g_ag_infection_prevalence_map.at(c_agent->getAgeGroup())++;

               if (c_agent->isIll())
                {
                   g_illness_prevalence++;
                   g_ag_illness_prevalence_map.at(c_agent->getAgeGroup())++;
                }
               if (c_agent->isIllnessDetected())
                {
                   g_detected_prevalence++;
                   g_ag_detection_prevalence_map.at(c_agent->getAgeGroup())++;
                   g_sla_prevalence_map.at(my_sla)++;
                   g_ag_sla_prevalence_map.at(c_agent->getAgeGroup()).at(my_sla)++;
                }
            }
        }
    }
}

void simulateDaytimeWeekend_SD(agent_map_type& agents)
{
   /* Update distributions for home */
   for (auto& agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       if (c_agent->isInfected())
        {
           c_agent->addToNighttimeProbabilityOfInfection_SD(VEi_1, VEi_2, AVEi); // CZ COV10 08042020 - removed agent->step() so isolation states can be updated before computing FoI
        }
    }
   //CZ: removing weekend daytime infections, as per Mikhail's suggestion...
   /* Decide who is now infected */
   /* for (auto& agent_entry : agents)
    * {
    *   auto c_agent = agent_entry.second;
    *
    *   updateAgentState(c_agent, c_agent->getHouseholdProbabilityOfInfection() * c_agent->getIsolationFactor_home(), MixingGroups::GroupHousehold);
    *   updateAgentState(c_agent, c_agent->getHouseholdClusterProbabilityOfInfection() * c_agent->getIsolationFactor_community(), MixingGroups::GroupHouseholdCluster);
    *   updateAgentState(c_agent, c_agent->getNeighbourhoodProbabilityOfInfection() * c_agent->getIsolationFactor_community(), MixingGroups::GroupNeighbourhood);
    *   updateAgentState(c_agent, c_agent->getCommunityProbabilityOfInfection() * c_agent->getIsolationFactor_community(), MixingGroups::GroupCommunity);
    *
    *
    * }*/
}


//CZ COV19 31032020
void update_HQ(agent_map_type& agents)
{
   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;
       c_agent->step_HQ(duration_HQ);
    }
}


// this function sets up the population for correct factor computation
// an agents isolation state is a function of the implemented policies and their compliance state.
void update_Isolation_states(agent_map_type& agents, bool SD_flag, bool CI_flag, bool HQ_flag, bool SC_flag)
{
   int  CI_counter = 0;
   int  SD_counter = 0;
   int  HQ_counter = 0;
   int  SC_counter = 0;
   bool assigned;

   // This script implicily assumes a ranking of isolation states
   // CI > HQ > SD > SC
   // This should be reflected in the FoI terms. TODO add some check that sends
   // a warning if this hierarchy is violated.


   // have to add a separate loop for CI because HQ is conditioned on it.
   // CI will apply to all compliant agents who are ill, it trumps all other measures they may comply with
   // This means there also needs to be a separate loop for clearing all residual states
   // may as well combine these.

   for (auto agent_entry : agents)
    {
       // first, set the isolation state to nil, to make sure only currently active measures are applied
       // the exception is for quarantined agents, who get reset in the function update_HQ (As this measure has a clock attached to it)
       auto c_agent = agent_entry.second;

       if (c_agent->getIsolationState() != Isolation_States::HQ)
        {
           c_agent->setIsolationState(Isolation_States::nil);
        }

       if (CI_flag)
        {
           // if ((c_agent->getComplianceState() > 1) &&                    // agent is CI compliant
           // (c_agent->getState() == DiseaseStates::StateSymptomatic)) // agent is symptomatic, should I change this to isIllnessDetected?
           if ((c_agent->getComplianceState() > 1) &&                                                      // agent is CI compliant
               ((c_agent->getState() == DiseaseStates::StateSymptomatic) || c_agent->isIllnessDetected())) // added case isolation for detected
           // asymptomatic agents as well.
            {
               c_agent->setIsolationState(Isolation_States::CI);
               c_agent->getHH()->setCIinHousehold(true);
               CI_counter += 1;
            }
        }
    }

   // now go through and apply all other possible interventions in order of 'strength' of the intervention

   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       if (c_agent->getIsolationState() != Isolation_States::CI)
        {
           // HQ will apply to all compliant agents who have a symptomatic (detected) illness in their home, unless they
           // are in case isolation themselves or are already quarantined.
           if (HQ_flag)
            {
               if (c_agent->getIsolationState() == Isolation_States::HQ)
                {
                   HQ_counter += 1;
                   continue;
                }
               else if ((c_agent->getHH()->getCIinHousehold()) &&
                        ((c_agent->getComplianceState() > 2) && (c_agent->getComplianceState() < 5)))
                {
                   c_agent->setIsolationState(Isolation_States::HQ);
                   HQ_counter += 1;
                   continue;
                }
            }

           // SD will apply to any compliant agents who are not actively in HQ or CI
           if (SD_flag)
            {
               if (((c_agent->getComplianceState() > 0) && (c_agent->getComplianceState() < 4)))
                {
                   c_agent->setIsolationState(Isolation_States::SD);
                   SD_counter += 1;
                   continue;
                }
            }

           // school closure only applies to teachers, students, and parents who are not
           // already isolated by one of the other measures.
           if (SC_flag)
            {
               if ((c_agent->getSCState() > 0))
                {
                   c_agent->setIsolationState(Isolation_States::SC);
                   SC_counter += 1;
                   continue;
                }
            }
        }
    }

   g_HQ_agents = HQ_counter;
   g_CI_agents = CI_counter;

   std::cout << "\t" << " CI count: " << CI_counter << ","
             << "\t" << " HQ count: " << HQ_counter << ","
             << "\t" << " SD count: " << SD_counter << ","
             << "\t" << " SC count: " << SC_counter << std::endl;
}


// CZ COV19 01042020 set isolation factors for FoI based on Agent's isolation states.

void set_Isolation_Factors_home(agent_map_type& agents, double FoI_SD, double FoI_CI, double FoI_HQ, double FoI_SC)
{
   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;
       c_agent->setIsolationFactor_home(FoI_SD, FoI_CI, FoI_HQ, FoI_SC);
    }
}


void set_Isolation_Factors_work(agent_map_type& agents, double FoI_SD, double FoI_CI, double FoI_HQ, double FoI_SC)
{
   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;
       c_agent->setIsolationFactor_work(FoI_SD, FoI_CI, FoI_HQ, FoI_SC);
    }
}


void set_Isolation_Factors_community(agent_map_type& agents, double FoI_SD, double FoI_CI, double FoI_HQ, double FoI_SC)
{
   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;
       c_agent->setIsolationFactor_community(FoI_SD, FoI_CI, FoI_HQ, FoI_SC);
    }
}


void assign_SC_compliance_parents(agent_map_type& agents, double p_parent_stays_home)
{
   int n_parents_SC = 0;

   // reset SC compliance for parents
   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       if (c_agent->isAdult() && !c_agent->isTeacher())
        {
           c_agent->setSCState(SC_States::SC_nil);
        }
    }

   // decide which parents are staying home

   int           parent_index;
   unsigned long parent_ID;

   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       if (c_agent->getSCState() == SC_States::SC_child)
        {
           parent_index = c_agent->findParent_ID(agents);

           if (parent_index < 0)
            {
               std::cout << "WARNING: Something went wrong in findParent_ID(), no parent found" << std::endl;
            }
           else
            {
               parent_ID = (unsigned long)parent_index;

               if (agents.at(parent_ID)->getSCState() == SC_States::SC_nil)
                {
                   if ((rand() / (double)RAND_MAX) < p_parent_stays_home)
                    {
                       agents.at(parent_ID)->setSCState(SC_States::SC_parent);
                       n_parents_SC += 1;
                    }
                }
            }
        }
    }

   std::cout << "***** Parents taking care of children at home: " << n_parents_SC << std::endl;
}


void assign_SC_compliance_students_teachers(agent_map_type& agents)
{
   int n_students = 0;
   int n_teachers = 0;

   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       if (c_agent->isStudent())
        {
           c_agent->setSCState(SC_States::SC_child);
           n_students += 1;
        }

       else if (c_agent->isTeacher())
        {
           c_agent->setSCState(SC_States::SC_teacher);
           n_teachers += 1;
        }
    }

   std::cout << "...SC applied to " << n_students << " students" << std::endl;
   std::cout << "...SC applied to " << n_teachers << " teachers" << std::endl;
}


//CZ COV19 17032020
void assign_SD_compliance(agent_map_type& agents)
{
   int n_SD_compliant = 0;

   for (auto agent_entry : agents)
    {
       auto   c_agent          = agent_entry.second;
       double agent_compliance = rand() / (double)RAND_MAX;
       if (agent_compliance < compliance_rate_SD)
        {
           c_agent->setSDState(SD_States::SD_compliant);
           n_SD_compliant += 1;
        }
    }

   std::cout << " ***** SD compliance assigned to " << n_SD_compliant << " agents by assign_SD_compliance()" << std::endl;
}


// CZ COV19 24032020
void assign_CI_compliance(agent_map_type& agents)
{
   int n_CI_compliant = 0;

   for (auto agent_entry : agents)
    {
       auto   c_agent          = agent_entry.second;
       double agent_compliance = rand() / (double)RAND_MAX;
       if (agent_compliance < compliance_rate_CI)
        {
           c_agent->setCIState(CI_States::CI_compliant);
           n_CI_compliant += 1;
        }
    }

   std::cout << " ***** CI compliance assigned to " << n_CI_compliant << " agents by assign_CI_compliance()" << std::endl;
}


// CZ COV19 24032020
void assign_HQ_compliance(agent_map_type& agents)
{
   int n_HQ_compliant = 0;

   for (auto agent_entry : agents)
    {
       auto   c_agent          = agent_entry.second;
       double agent_compliance = rand() / (double)RAND_MAX;
       if ((c_agent->getCIState() == CI_States::CI_compliant) && (agent_compliance < compliance_rate_HQ))
        {
           c_agent->setHQState(HQ_States::HQ_compliant);
           n_HQ_compliant += 1;
        }
    }

   std::cout << "***** HQ compliance assigned to " << n_HQ_compliant << " agents by assign_HQ_compliance()" << std::endl;
}


void set_Compliance_states_all(agent_map_type& agents)
{
   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;
       c_agent->eval_Compliance_state();
    }
}


void reZero(hh_map_type& households, hc_map_type& household_clusters, cd_map_type& census_districts, dz_map_type& destination_zones,
   wg_map_type& working_groups, school_map_type& schools, wg_map_type& grades, wg_map_type& classes, sla_map_type& statistical_local_areas)
{
   for (auto& chh : households)
    {
       chh.second->reZero();
    }

   for (auto& chc : household_clusters)
    {
       chc.second->reZero();
    }

   for (auto& ccd : census_districts)
    {
       ccd.second->reZero();
    }

   for (auto& cdz : destination_zones)
    {
       cdz.second->reZero();
    }

   for (auto& cwg : working_groups)
    {
       cwg.second->reZero();
    }

   for (auto& csch : schools)
    {
       csch.second->reZero();
    }

   for (auto& cgr : grades)
    {
       cgr.second->reZero();
    }

   for (auto& ccl : classes)
    {
       ccl.second->reZero();
    }

   for (auto& csla : statistical_local_areas)
    {
       csla.second->reZero();

       /* Probably should just have these in each statistical area */
       g_sla_prevalence_map.at(csla.first)     = 0;
       g_sla_incidence_map.at(csla.first)      = 0;
       g_sla_number_removed_map.at(csla.first) = 0;

       for (AgeGroup ag = AgeGroups::Age0to4; ag <= AgeGroups::Age65Plus; ag = AgeGroup(ag + 1))
        {
           g_ag_sla_prevalence_map.at(ag).at(csla.first)     = 0;
           g_ag_sla_incidence_map.at(ag).at(csla.first)      = 0;
           g_ag_sla_number_removed_map.at(ag).at(csla.first) = 0;
        }
    }
   /* Zero our global variables too */
   g_detected_incidence  = 0;
   g_detected_prevalence = 0;

   g_infected_incidence  = 0; // this will count infections
   g_infected_prevalence = 0; // Rezero the prevalence (asymptomatic and symptomatic)

   g_illness_incidence  = 0;
   g_illness_prevalence = 0;

   g_detected_incidence_VACge = 0;
   g_detected_incidence_VACpr = 0;
   g_infected_incidence_VACge = 0;
   g_infected_incidence_VACpr = 0;

   for (AgeGroup ag = AgeGroups::Age0to4; ag <= AgeGroups::Age65Plus; ag = AgeGroup(ag + 1))
    {
       g_ag_infection_incidence_map.at(ag)  = 0;
       g_ag_illness_incidence_map.at(ag)    = 0;
       g_ag_detection_incidence_map.at(ag)  = 0;
       g_ag_infection_prevalence_map.at(ag) = 0;
       g_ag_illness_prevalence_map.at(ag)   = 0;
       g_ag_detection_prevalence_map.at(ag) = 0;
       g_ag_detection_incidence_VACpr_map.at(ag)  = 0;
       g_ag_detection_incidence_VACge_map.at(ag)  = 0;
       g_ag_infection_incidence_VACpr_map.at(ag)  = 0;
       g_ag_infection_incidence_VACge_map.at(ag)  = 0;
    }

   g_number_removed = 0;

   num_CD_incidence  = 0;
   num_CD_prevalence = 0;         // Rezero the CD prevalence (asymptomatic and symptomatic)

   AV_doses = 0;
}


// parsing in the intervention inputs
void parse_pharm_intervention_params(const std::string& pharm_input_fname, const sla_map_type& sla_map)
{
   std::string   input_file(pharm_input_fname);
   std::ifstream input_stream(input_file);
   std::string   cline;
   unsigned long c_sla;

   while (std::getline(input_stream, cline))
    {
       std::stringstream i_ss;
       i_ss << cline;

       std::string first_element;
       i_ss >> first_element;

       // If the SLAs are designated for the intervention then they are put in list of SLA IDs, if not, all SLAs are added to the list i.e. the intervention is applied to all SLAs
       if (first_element == "vac_SLAs:")
        {
           vac_SLA_bool = 1;
           while (i_ss >> c_sla)
            {
               vac_SLAs.push_back(c_sla);
            }
        }

       if (first_element == "AV_SLAs:")
        {
           while (i_ss >> c_sla)
            {
               AV_SLAs.push_back(c_sla);
            }
        }


       // Collecting the intervention input parrameters from the intervention_input.dat file
       if (first_element == "Pharm_intervention:")
        {
           i_ss >> pharm_intervention;
        }

       // CZ: progressive, age-stratified vaccination
       if (first_element == "n_vac_per_week_1:")
        {
           i_ss >> n_vac_per_week_1;
        }

       if (first_element == "n_vac_pre_epidemic_1:")
        {
           i_ss >> n_vac_pre_epidemic_1;
        }

        if (first_element == "n_vac_per_week_2:")
         {
            i_ss >> n_vac_per_week_2;
         }

        if (first_element == "n_vac_pre_epidemic_2:")
         {
            i_ss >> n_vac_pre_epidemic_2;
         }

       if (first_element == "p_vac_65_and_over:")
        {
           i_ss >> p_vac_65_and_over;
        }

       if (first_element == "p_vac_18_to_64:")
        {
           i_ss >> p_vac_18_to_64;
        }

       if (first_element == "p_vac_children:")
        {
           i_ss >> p_vac_children;
        }

       if (first_element == "VEs_1:")
        {
           i_ss >> VEs_1;
        }
       if (first_element == "VEi_1:")
        {
           i_ss >> VEi_1;
        }
       // CZ: adding VEd
       if (first_element == "VEd_1:")
        {
           i_ss >> VEd_1;
        }


        if (first_element == "VEs_2:")
         {
            i_ss >> VEs_2;
         }
        if (first_element == "VEi_2:")
         {
            i_ss >> VEi_2;
         }
        // CZ: adding VEd
        if (first_element == "VEd_2:")
         {
            i_ss >> VEd_2;
         }


       if (first_element == "AVEs:")
        {
           i_ss >> AVEs;
        }
       if (first_element == "AVEi:")
        {
           i_ss >> AVEi;
        }
       if (first_element == "AVEd:")
        {
           i_ss >> AVEd;
        }
       if (first_element == "GTAP:")
        {
           i_ss >> GTAP;
        }
       if (first_element == "TAP:")
        {
           i_ss >> TAP;
        }
       if (first_element == "Intervention_delay:")
        {
           i_ss >> Response_delay;
        }


       // if (first_element == "Detection:")
       //  {
       //     i_ss >> Detection;
       //  }
       //
       // if (first_element == "Detection_asymp:")
       //  {
       //     i_ss >> Detection_asymp;
       //  }
    }


   // If the SLAs aren't specified then the list is populated with all of the SLAs
   if (vac_SLAs.empty())
    {
       for (auto& sla : sla_map)
        {
           c_sla = sla.first;
           vac_SLAs.push_back(c_sla);
        }
    }

   if (AV_SLAs.empty())
    {
       for (auto& sla : sla_map)
        {
           c_sla = sla.first;
           AV_SLAs.push_back(c_sla);
        }
    }


   input_stream.close();
}


void parse_SD_intervention_inputs(const std::string& SD_interv_fname)
{
   {
      std::string   param_file(SD_interv_fname);
      std::ifstream param_stream(param_file);
      std::string   cline;

      while (std::getline(param_stream, cline))
       {
          std::stringstream i_ss;
          i_ss << cline;

          std::string first_element;
          i_ss >> first_element;

          if (first_element == "SD_intervention:")
           {
              i_ss >> SD_intervention;
           }

          if (first_element == "n_detected_SD_trigger:")
           {
              i_ss >> n_detected_SD_trigger;
           }
          if (first_element == "compliance_rate_SD:")
           {
              i_ss >> compliance_rate_SD;
           }
          if (first_element == "duration_SD:")
           {
              i_ss >> duration_SD;
           }
          if (first_element == "FoI_SD_community:")
           {
              i_ss >> FoI_SD_community;
           }
          if (first_element == "FoI_SD_work:")
           {
              i_ss >> FoI_SD_work;
           }
          if (first_element == "FoI_SD_home:")
           {
              i_ss >> FoI_SD_home;
           }


          if (first_element == "School_closure:")
           {
              i_ss >> School_closure;
           }
          if (first_element == "n_detected_SC_trigger:")
           {
              i_ss >> n_detected_SC_trigger;
           }
          if (first_element == "duration_SC:")
           {
              i_ss >> duration_SC;
           }
          if (first_element == "FoI_SC_community:")
           {
              i_ss >> FoI_SC_community;
           }
          if (first_element == "FoI_SC_work:")
           {
              i_ss >> FoI_SC_work;
           }
          if (first_element == "FoI_SC_home:")
           {
              i_ss >> FoI_SC_home;
           }
          if (first_element == "p_parent_stays_home:")
           {
              i_ss >> p_parent_stays_home;
           }


          if (first_element == "CI_intervention:")
           {
              i_ss >> CI_intervention;
           }

          if (first_element == "compliance_rate_CI:")
           {
              i_ss >> compliance_rate_CI;
           }

          if (first_element == "FoI_CI_community:")
           {
              i_ss >> FoI_CI_community;
           }

          if (first_element == "FoI_CI_work:")
           {
              i_ss >> FoI_CI_work;
           }

          if (first_element == "FoI_CI_home:")
           {
              i_ss >> FoI_CI_home;
           }


          if (first_element == "HQ_intervention:")
           {
              i_ss >> HQ_intervention;
           }

          if (first_element == "compliance_rate_HQ:")
           {
              i_ss >> compliance_rate_HQ;
           }

          if (first_element == "duration_HQ:")
           {
              i_ss >> duration_HQ;
           }

          if (first_element == "FoI_HQ_community:")
           {
              i_ss >> FoI_HQ_community;
           }

          if (first_element == "FoI_HQ_work:")
           {
              i_ss >> FoI_HQ_work;
           }

          if (first_element == "FoI_HQ_home:")
           {
              i_ss >> FoI_HQ_home;
           }
       }

      param_stream.close();
   }
}


// CZ 10 06 2020 - read in the disease parameters in the Parameters_input.dat file

void parse_Parameters(const std::string& Parameter_input_fname)
{
   {
      std::string   param_file(Parameter_input_fname);
      std::ifstream param_stream(param_file);
      std::string   cline;

      while (std::getline(param_stream, cline))
       {
          std::stringstream i_ss;
          i_ss << cline;

          std::string first_element;
          i_ss >> first_element;

          if (first_element == "incubation_period_mu:")
           {
              i_ss >> incubation_period_mu;
           }
          if (first_element == "incubation_period_sig:")
           {
              i_ss >> incubation_period_sig;
           }
          if (first_element == "recovery_period_hi:")
           {
              i_ss >> recovery_period_hi;
           }
          if (first_element == "recovery_period_low:")
           {
              i_ss >> recovery_period_low;
           }
          if (first_element == "latent_period:")
           {
              i_ss >> latent_period;
           }
          if (first_element == "adult_symptomatic_fraction:")
           {
              i_ss >> adult_symptomatic_fraction;
           }
          if (first_element == "child_symptomatic_fraction:")
           {
              i_ss >> child_symptomatic_fraction;
           }
          if (first_element == "asymptomatic_infectivity:")
           {
              i_ss >> asymptomatic_infectivity;
           }
          if (first_element == "Detection:")
           {
              i_ss >> Detection;
           }

          if (first_element == "Detection_asymp:")
           {
              i_ss >> Detection_asymp;
           }
       }

      param_stream.close();
   }
}


// Checking if agent is exposed and, if so, starting the intervention delay for AV
void dynamic_intervention(agent_map_type& agents)
{
   std::cout << "... Administering anti-virals ..." << std::endl;

   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       if (c_agent->getAVState() == AVStates::Notexposed)
        {
           auto my_sla = c_agent->getSLA()->getID();
           if (std::find(AV_SLAs.begin(), AV_SLAs.end(), my_sla) != AV_SLAs.end()) // Only in speficed SLAs. Defult is all.
            {
               if (GTAP > 0)                                                       // if anyone is meant to have GTAP
                {
                   if (c_agent->CD_Sympotmatic_present())                          // Only in symptomatic infection is present in SA1
                    {
                       c_agent->GTAP_administer(GTAP);                             // relables agent as exposed in geographical area and starts count for antiviral
                    }
                }
               if (TAP > 0)
                {
                   if (c_agent->TAP_Sympotmatic_present()) // Only in symptomatic infection is present in HH or WG
                    {
                       c_agent->TAP_administer(TAP);       // relables WG and HH agent as exposed and starts count for antiviral
                    }
                }
            }
        }
    }
   std::cout << "\tNumber of AV doses used: " << AV_doses << std::endl;
}


void count_vaccinated_agents(agent_map_type& agents)
{
   unsigned long n_children = 0;
   unsigned long n_adults   = 0;
   unsigned long n_seniors  = 0;

   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       if (c_agent->getVacState() == VacStates::vaccinated_priority ||
           c_agent->getVacState() == VacStates::vaccinated_general )
        {
           if (c_agent->getAgeGroup() == AgeGroups::Age65Plus)
            {
               n_seniors++;
            }
           else if (c_agent->getAgeGroup() == AgeGroups::Age0to4 ||
                    c_agent->getAgeGroup() == AgeGroups::Age5to18)
            {
               n_children++;
            }
           else
            {
               n_adults++;
            }
        }
    }
   std::cout << "checking the vaccine count: " << std::endl;
   std::cout << " children vaccinated: " << n_children << std::endl;
   std::cout << " adults vaccinated: " << n_adults << std::endl;
   std::cout << " seniors vaccinated: " << n_seniors << std::endl;
}

void progressive_vaccination_priority(agent_map_type& agents,
   std::stack<unsigned long>& unvaccinated_senior_IDs_s,
   std::stack<unsigned long>&  unvaccinated_adult_IDs_s,
   std::stack<unsigned long>&  unvaccinated_child_IDs_s)
{
   // CZ: modified age-stratified progressive-vaccination

   int           num_vac_today          = 0;
   int           num_vac_65_plus_today  = 0;
   int           num_vac_18_to_64_today = 0;
   int           num_vac_children_today = 0;

   unsigned long n_vac_daily = (unsigned long)round(n_vac_per_week_1 / 7);

   unsigned long n_unvaccinated = unvaccinated_senior_IDs_s.size() +
                                  unvaccinated_adult_IDs_s.size();

   while (num_vac_today < n_vac_daily && n_unvaccinated > 0)
    {
       if (unvaccinated_senior_IDs_s.size() > 0)
        {
           for (int i = 1; i < p_vac_65_and_over + 1; i++)
            {
               if (unvaccinated_senior_IDs_s.size() > 0 && num_vac_today < n_vac_daily)
                {
                   auto c_agent = agents.at(unvaccinated_senior_IDs_s.top());

                   if (c_agent->wasIllnessDetected())
                    {
                       unvaccinated_senior_IDs_s.pop();
                    }
                   else
                    {
                       c_agent->setVacState(VacState(VacStates::vaccinated_priority));
                       unvaccinated_senior_IDs_s.pop();
                       num_vac++;
                       num_vac_PV++;
                       num_vac_today++;
                       num_vac_65_plus++;
                       num_vac_65_plus_PV++;
                       num_vac_65_plus_today++;
                    }
                }
            }
        }

       if (unvaccinated_adult_IDs_s.size() > 0)
        {
           for (int i = 1; i < p_vac_18_to_64 + 1; i++)
            {
               if (unvaccinated_adult_IDs_s.size() > 0 && num_vac_today < n_vac_daily)
                {
                   auto c_agent = agents.at(unvaccinated_adult_IDs_s.top());

                   if (c_agent->wasIllnessDetected())
                    {
                       unvaccinated_adult_IDs_s.pop();
                    }
                   else
                    {
                       c_agent->setVacState(VacState(VacStates::vaccinated_priority));
                       unvaccinated_adult_IDs_s.pop();
                       num_vac++;
                       num_vac_PV++;
                       num_vac_today++;
                       num_vac_18_to_64++;
                       num_vac_18_to_64_PV++;
                       num_vac_18_to_64_today++;
                    }
                }
            }
        }


       if (unvaccinated_child_IDs_s.size() > 0)
        {
           for (int i = 0; i < p_vac_children; i++)
            {
               if (unvaccinated_child_IDs_s.size() > 0 && num_vac_today < n_vac_daily)
                {
                   auto c_agent = agents.at(unvaccinated_child_IDs_s.top());

                   if (c_agent->wasIllnessDetected())
                    {
                       unvaccinated_child_IDs_s.pop();
                    }
                   else
                    {
                       c_agent->setVacState(VacState(VacStates::vaccinated_priority));
                       unvaccinated_child_IDs_s.pop();

                       num_vac++;
                       num_vac_PV++;
                       num_vac_today++;
                       num_vac_children++;
                       num_vac_children_PV++;
                       num_vac_children_today++;
                    }
                }
            }
        }

       n_unvaccinated = unvaccinated_senior_IDs_s.size() +
                        unvaccinated_adult_IDs_s.size();
    }


   std::cout << "\tNumber of agents vaccinated today, priority (total): "
             << num_vac_today << " (" << num_vac_PV << ")" << std::endl;
   std::cout << "\tNumber of children vaccinated today, priority (total): "
             << num_vac_children_today << " (" << num_vac_children_PV << ")" << std::endl;
   std::cout << "\tNumber of agents aged 18-64 vaccinated today, priority (total): "
             << num_vac_18_to_64_today << " (" << num_vac_18_to_64_PV << ")" << std::endl;
   std::cout << "\tNumber of agents aged 65+ vaccinated today, priority (total): "
             << num_vac_65_plus_today << " (" << num_vac_65_plus_PV << ")" << std::endl;
}


void progressive_vaccination_general(agent_map_type& agents,
   std::stack<unsigned long>& unvaccinated_senior_IDs_s,
   std::stack<unsigned long>&  unvaccinated_adult_IDs_s,
   std::stack<unsigned long>&  unvaccinated_child_IDs_s)
{
   // CZ: modified age-stratified progressive-vaccination

   int           num_vac_today          = 0;
   int           num_vac_65_plus_today  = 0;
   int           num_vac_18_to_64_today = 0;
   int           num_vac_children_today = 0;

   unsigned long n_vac_daily = (unsigned long)round(n_vac_per_week_2 / 7);

   unsigned long n_unvaccinated = unvaccinated_senior_IDs_s.size() +
                                  unvaccinated_adult_IDs_s.size();

   while (num_vac_today < n_vac_daily && n_unvaccinated > 0)
    {
       if (unvaccinated_senior_IDs_s.size() > 0)
        {
           for (int i = 1; i < p_vac_65_and_over + 1; i++)
            {
               if (unvaccinated_senior_IDs_s.size() > 0 && num_vac_today < n_vac_daily)
                {
                   auto c_agent = agents.at(unvaccinated_senior_IDs_s.top());

                   if (c_agent->wasIllnessDetected())
                    {
                       unvaccinated_senior_IDs_s.pop();
                    }
                   else
                    {
                       c_agent->setVacState(VacState(VacStates::vaccinated_general));
                       unvaccinated_senior_IDs_s.pop();
                       num_vac++;
                       num_vac_GV++;
                       num_vac_today++;
                       num_vac_65_plus++;
                       num_vac_65_plus_GV++;
                       num_vac_65_plus_today++;
                    }
                }
            }
        }

       if (unvaccinated_adult_IDs_s.size() > 0)
        {
           for (int i = 1; i < p_vac_18_to_64 + 1; i++)
            {
               if (unvaccinated_adult_IDs_s.size() > 0 && num_vac_today < n_vac_daily)
                {
                   auto c_agent = agents.at(unvaccinated_adult_IDs_s.top());

                   if (c_agent->wasIllnessDetected())
                    {
                       unvaccinated_adult_IDs_s.pop();
                    }
                   else
                    {
                       c_agent->setVacState(VacState(VacStates::vaccinated_general));
                       unvaccinated_adult_IDs_s.pop();
                       num_vac++;
                       num_vac_GV++;
                       num_vac_today++;
                       num_vac_18_to_64++;
                       num_vac_18_to_64_GV++;
                       num_vac_18_to_64_today++;
                    }
                }
            }
        }


       if (unvaccinated_child_IDs_s.size() > 0)
        {
           for (int i = 0; i < p_vac_children; i++)
            {
               if (unvaccinated_child_IDs_s.size() > 0 && num_vac_today < n_vac_daily)
                {
                   auto c_agent = agents.at(unvaccinated_child_IDs_s.top());

                   if (c_agent->wasIllnessDetected())
                    {
                       unvaccinated_child_IDs_s.pop();
                    }
                   else
                    {
                       c_agent->setVacState(VacState(VacStates::vaccinated_general));
                       unvaccinated_child_IDs_s.pop();

                       num_vac++;
                       num_vac_GV++;
                       num_vac_today++;
                       num_vac_children++;
                       num_vac_children_GV++;
                       num_vac_children_today++;
                    }
                }
            }
        }

       n_unvaccinated = unvaccinated_senior_IDs_s.size() +
                        unvaccinated_adult_IDs_s.size();
    }


   std::cout << "\tNumber of agents vaccinated today, general (total): "
             << num_vac_today << " (" << num_vac_GV << ")" << std::endl;
   std::cout << "\tNumber of children vaccinated today, general (total): "
             << num_vac_children_today << " (" << num_vac_children_GV << ")" << std::endl;
   std::cout << "\tNumber of agents aged 18-64 vaccinated today, general (total): "
             << num_vac_18_to_64_today << " (" << num_vac_18_to_64_GV << ")" << std::endl;
   std::cout << "\tNumber of agents aged 65+ vaccinated today, general (total): "
             << num_vac_65_plus_today << " (" << num_vac_65_plus_GV << ")" << std::endl;
}




void pre_epidemic_vaccination(agent_map_type& agents,
   std::stack<unsigned long>&  unvaccinated_senior_IDs_s,
   std::stack<unsigned long>&  unvaccinated_adult_IDs_s,
   std::stack<unsigned long>&  unvaccinated_child_IDs_s)
{
   unsigned long n_unvaccinated = unvaccinated_senior_IDs_s.size() +
                                  unvaccinated_adult_IDs_s.size();

   while ((num_vac < n_vac_pre_epidemic_1 + n_vac_pre_epidemic_2) && (n_unvaccinated > 0))
    {
       //unsigned long rand_index;
       //unsigned long n_agents;

       if (unvaccinated_senior_IDs_s.size() > 0)
        {
           for (int i = 1; i < p_vac_65_and_over + 1; i++)
            {
               if ((unvaccinated_senior_IDs_s.size() > 0) && (num_vac < n_vac_pre_epidemic_1 + n_vac_pre_epidemic_2))
                {
                   //n_agents = unvaccinated_senior_IDs.size();

                   //rand_index = floor((rand() / (double)RAND_MAX) * n_agents);

                   //auto c_agent = agents.at(unvaccinated_senior_IDs.at(rand_index));

                   auto c_agent = agents.at(unvaccinated_senior_IDs_s.top());

                   if (c_agent->wasIllnessDetected())
                    {
                       //unvaccinated_senior_IDs.erase(unvaccinated_senior_IDs.begin() + rand_index);
                       unvaccinated_senior_IDs_s.pop();
                    }
                   else
                    {
                       if (num_vac < n_vac_pre_epidemic_1)
                        {
                           c_agent->setVacState(VacState(VacStates::vaccinated_priority));
                           num_vac_PV++;
                           num_vac_65_plus_PV++;
                        }
                       else
                        {
                           c_agent->setVacState(VacState(VacStates::vaccinated_general));
                           num_vac_GV++;
                           num_vac_65_plus_GV++;
                        }
                       //unvaccinated_senior_IDs.erase(unvaccinated_senior_IDs.begin() + rand_index);
                       unvaccinated_senior_IDs_s.pop();
                       num_vac++;
                       num_vac_65_plus++;
                    }
                }
            }
        }

       if (unvaccinated_adult_IDs_s.size() > 0)
        {
           for (int i = 1; i < p_vac_18_to_64 + 1; i++)
            {
               if ((unvaccinated_adult_IDs_s.size() > 0) && (num_vac < n_vac_pre_epidemic_1 + n_vac_pre_epidemic_2))
                {
                   //n_agents = unvaccinated_adult_IDs.size();

                   //rand_index = floor((rand() / (double)RAND_MAX) * n_agents);

                   // auto c_agent = agents.at(unvaccinated_adult_IDs.at(rand_index));

                   auto c_agent = agents.at(unvaccinated_adult_IDs_s.top());

                   if (c_agent->wasIllnessDetected())
                    {
                       //unvaccinated_adult_IDs.erase(unvaccinated_adult_IDs.begin() + rand_index);
                       unvaccinated_adult_IDs_s.pop();
                    }
                   else
                    {
                       if (num_vac < n_vac_pre_epidemic_1)
                        {
                           c_agent->setVacState(VacState(VacStates::vaccinated_priority));
                           num_vac_PV++;
                           num_vac_18_to_64_PV++;
                        }
                       else
                        {
                           c_agent->setVacState(VacState(VacStates::vaccinated_general));
                           num_vac_GV++;
                           num_vac_18_to_64_GV++;
                        }
                       //unvaccinated_adult_IDs.erase(unvaccinated_adult_IDs.begin() + rand_index);
                       unvaccinated_adult_IDs_s.pop();
                       num_vac++;
                       num_vac_18_to_64++;
                    }
                }
            }
        }


       if (unvaccinated_child_IDs_s.size() > 0)
        {
           for (int i = 0; i < p_vac_children; i++)
            {
               if (unvaccinated_child_IDs_s.size() > 0 && (num_vac < n_vac_pre_epidemic_1 + n_vac_pre_epidemic_2))
                {
                   //n_agents = unvaccinated_child_IDs.size();

                   //rand_index = floor((rand() / (double)RAND_MAX) * n_agents);

                   //auto c_agent = agents.at(unvaccinated_child_IDs.at(rand_index));

                   auto c_agent = agents.at(unvaccinated_child_IDs_s.top());

                   if (c_agent->wasIllnessDetected())
                    {
                       //unvaccinated_child_IDs.erase(unvaccinated_child_IDs.begin() + rand_index);
                       unvaccinated_child_IDs_s.pop();
                    }
                   else
                    {
                       if (num_vac < n_vac_pre_epidemic_1)
                        {
                           c_agent->setVacState(VacState(VacStates::vaccinated_priority));
                           num_vac_PV++;
                           num_vac_children_PV++;
                        }
                       else
                        {
                           c_agent->setVacState(VacState(VacStates::vaccinated_general));
                           num_vac_GV++;
                           num_vac_children_GV++;
                        }
                       //unvaccinated_child_IDs.erase(unvaccinated_child_IDs.begin() + rand_index);
                       unvaccinated_child_IDs_s.pop();

                       num_vac++;
                       num_vac_children++;
                    }
                }
            }
        }

       n_unvaccinated = unvaccinated_senior_IDs_s.size() +
                        unvaccinated_adult_IDs_s.size() ;
    }

   std::cout << "\tNumber of pre-vaccinated agents: " << num_vac << std::endl;
   std::cout << "\tNumber of pre-vaccinated children: " << num_vac_children << std::endl;
   std::cout << "\tNumber of pre-vaccinated aged 18-64: " << num_vac_18_to_64 << std::endl;
   std::cout << "\tNumber of pre-vaccinated aged 65+: " << num_vac_65_plus << std::endl;

   std::cout << "\tNumber of pre-vaccinated agents (Priority): " << num_vac_PV << std::endl;
   std::cout << "\tNumber of pre-vaccinated children (Priority): " << num_vac_children_PV << std::endl;
   std::cout << "\tNumber of pre-vaccinated aged 18-64 (Priority): " << num_vac_18_to_64_PV << std::endl;
   std::cout << "\tNumber of pre-vaccinated aged 65+ (Priority): " << num_vac_65_plus_PV << std::endl;

   std::cout << "\tNumber of pre-vaccinated agents (General): " << num_vac_GV << std::endl;
   std::cout << "\tNumber of pre-vaccinated children (General): " << num_vac_children_GV << std::endl;
   std::cout << "\tNumber of pre-vaccinated aged 18-64 (General): " << num_vac_18_to_64_GV << std::endl;
   std::cout << "\tNumber of pre-vaccinated aged 65+ (General): " << num_vac_65_plus_GV << std::endl;
}

int count_CD_incidence(cd_map_type& census_districts)
{
   int num_CD_incidence = 0;  //counts number of CDs with newly infected agents

   for (auto& ccd : census_districts)
    {
       if (ccd.second->is_CD_newly_infected())
        {
           num_CD_incidence++;
           ccd.second->is_not_newly_infected();
        }
    }
   return(num_CD_incidence);
}


int count_CD_prevalence(cd_map_type& census_districts)
{
   for (auto& ccd : census_districts)
    {
       if (ccd.second->get_n_infecteds_tot() > 0)
        {
           num_CD_prevalence++;
        }
    }
   return(num_CD_prevalence);
}


//end Kit's edit -AVE
