/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright   Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-201), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright   Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   This code is released only for internal review by Nature Communications, manuscript: NCOMMS-21-29568.

   *EndCopyright:
*/

#ifdef _MSC_VER
#define _CRT_SECURE_NO_WARNINGS
#endif

#include "agent.h"
#include "create_population.h"
#include "epidemic.h"
#include "debug.h"


#include <cstddef>
#include <fstream>
#include <iostream>
#include <map>
#include <string>
#include <vector>
#include <memory>
#include <time.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <getopt.h>
#include <random>

//#define INFECT_ILL_ONLY
//#define COUNT_ILL_ONLY
//#define COUNT_SYDNEY_ONLY

unsigned long              g_cumulative_ill   = 0;
unsigned long              g_presymptomatic   = 0;
unsigned long              g_asymptomatic     = 0;
unsigned long              g_generation_steps = 0; // for generation time
unsigned long              g_serial_steps     = 0; // for serial interval
std::vector<unsigned long> g_cumulative_ill_in_group {
   0, 0, 0, 0, 0, 0, 0, 0
};
// update
void init(hh_map_type&, hc_map_type&, cd_map_type&, dz_map_type&, wg_map_type&, school_map_type&, wg_map_type&, wg_map_type&, sla_map_type&);
void updateAgentState(std::shared_ptr<Agent>, double, unsigned long, unsigned);
void simulateDaytime(agent_map_type&, std::shared_ptr<Agent>&, unsigned long);
void simulateNighttime(agent_map_type&, std::shared_ptr<Agent>&, unsigned long);
// for R0 code, this is identical to simulateNighttime(), there are some differences in the main.cpp function
// due to test detection and bookkeeping
void simulateDaytimeWeekend(agent_map_type&, std::shared_ptr<Agent>&, unsigned long); //CZ COV19 21112020
void reZero(std::shared_ptr<Agent>&);

void parse_arguments(int argc, char *argv[], double&, double&, unsigned&, std::string &, std::string &, std::string &);
void parse_Parameters(const std::string& input_fname);

//CZ: COV19 - disease parameters for sensitivity analysis etc. 10 06 2020
// set default disease parameters - these can be altered by Parameters_input.dat
// read by the parse_Parameters() function.
double latent_period         = 0;
double incubation_period_mu  = 1.644;  //from Lauer et al.
double incubation_period_sig = 0.363;  // from Lauer et al.
double recovery_period_hi    = 10;     //time between end of exponential period an recovery - upper bound - CDC replication competent virus reports
double recovery_period_low   = 5;
//double infectious_period          = 15; // this will be the recovery period plus the incubation period
double adult_symptomatic_fraction = 0.67;
double child_symptomatic_fraction = 0.134;
double asymptomatic_infectivity   = 0.3;

//CZ 11 06 2020 incubation period will be set from a lognormal distribution
// CZ 11 06 2020 symptomatic period (time from onset to recovery) will be set as a uniform
// distribution (Qifang Bi et al., Figure S5 ) from 14 - 28 days
std::default_random_engine generator;


double dum = 0.0;


int main(int argc, char *argv[])
{
   double       p_trans      = 0.08;
   double       trans_scaler = 2.5;//3.33;
   unsigned    n_runs       = 100;
   std::string current_dir  = (char *)get_current_dir_name();

   std::string database_output = current_dir + "/AmTraC_outputs/2016_population_1/";
   std::string database_input  = current_dir + "/AmTraC_inputs/2016_population_1/";

   std::string Parameter_input_fname = "Parameter_input.dat";

   parse_arguments(argc, argv, p_trans, trans_scaler, n_runs, database_output, database_input, Parameter_input_fname);

   std::string pop_fname = database_input + "population.dat";

   std::cout << "Starting simulations with following settings:"
             << "\n\tInput Database: " << database_input
             << "\n\tOutput Database: " << database_output
             << "\n\tPopulation datafile: " << pop_fname
             << "\n\tParameter file: " << Parameter_input_fname
             << "\n\tRuns: " << n_runs
             << "\n\tP(trans): " << p_trans
             << "\n\tTransmission Scaler: " << trans_scaler
             << std::endl;

   // CZ 10 06 2020 parse input parameter file
   std::string   Parameter_input_file = database_input + "Parameter_input/" + Parameter_input_fname;

   std::ifstream Parameter_file(Parameter_input_file);

   if (!Parameter_file.good())
    {
       std::cout << "Parameter input file named: " << Parameter_input_file << " does not exist, using default parameters" << std::endl;
    }
   else
    {
       parse_Parameters(Parameter_input_file);
    }

   std::cout << "\n"
             << "\n" << "-----disease parameters-----"
             << "\n" << "incubation period mu: " << incubation_period_mu
             << "\n" << "incubation period sigma: " << incubation_period_sig
             << "\n" << "days from end of incubation to recovery (upper bound): " << recovery_period_hi
             << "\n" << "days from end of incubation to recovery (lower bound): " << recovery_period_low
             << "\n" << "adult symptomatic fraction: " << adult_symptomatic_fraction
             << "\n" << "child symptomatic fraction: " << child_symptomatic_fraction
             << "\n" << "asymptomatic infectivity factor: " << asymptomatic_infectivity << "\n" << std::endl;


   srand(time(NULL));

   /* Set up data structures */
   agent_map_type agents;
   hh_map_type    households;
   hc_map_type    household_clusters;
   cd_map_type    census_districts;
   wg_map_type    working_groups;
   wg_map_type    schools, grades, classes;
   dz_map_type    destination_zones;
   sla_map_type   statistical_local_areas;

   /* Scale contact rates by `p_trans' for inputting to agents */
   Epidemic::init_transmission_rates(p_trans, trans_scaler);

   /* Make a folder for todays results */
   struct timeval tv;
   gettimeofday(&tv, NULL);
   struct tm      *tm_info = localtime(&tv.tv_sec);
   char           buffer[26];
   strftime(buffer, 26, "%Y-%m-%d_%H-%M-%S", tm_info);

   std::size_t       last_fs  = pop_fname.find_last_of('L');
   std::size_t       last_dot = pop_fname.find_last_of('.');

   std::stringstream fname_ss;
   fname_ss << database_output << "results/R0_"
            << "P" << p_trans
            << "_T" << trans_scaler
            << "_D" << buffer;

   if (last_fs != -1)
    {
       std::string pl_str = pop_fname.substr(last_fs + 1, last_dot - last_fs - 1);
       fname_ss << "_L" << pl_str;
    }

   std::string fname(fname_ss.str());

   std::size_t dot_loc;
   while ((dot_loc = fname.find(".")) != std::string::npos)
    {
       fname.replace(dot_loc, 1, "-");
    }

   std::ofstream outfile_stream(fname + ".dat");
   std::cout << "Created file " << fname << ".dat for results." << std::endl;

   /* Create the population */
   std::cout << "Reading Population..." << std::endl;

   CreatePopulation::readPopulation(agents, households, household_clusters, census_districts, destination_zones, statistical_local_areas,
                                    working_groups, schools, grades, classes, database_input, pop_fname);

   /* Read population sets the id of the parameters in input */
   std::cout << "Done" << std::endl;

   std::vector<unsigned long> valid_agent_ids;
#ifdef COUNT_SYDNEY_ONLY
   /* Only include agents from sydney */
   for (auto c_agent : agents)
    {
       auto csla = c_agent.second->getSLA()->getID();
       if (floor(csla / 1000000) == 105)
        {
           valid_agent_ids.push_back(c_agent.second->getID());
        }
    }
   std::cout << "Number of agents from sydney: " << valid_agent_ids.size() << std::endl;
#endif


// seed generator for distributed parameters
//generator.seed(seed + 1); // shouldn't need this for r0 script, but if we want to do
// replicate runs we can define seed and uncomment.

// use mu and sigma parameters to make lognormal distribution for incubation period
   double                              incubation_period;
   double                              recovery_period;

   std::lognormal_distribution<double> incubation_period_dist(incubation_period_mu, incubation_period_sig);
   std::uniform_int_distribution<int>  recovery_period_dist(recovery_period_low, recovery_period_hi);

// set disease parameters for each agent
   double incubation_period_avg = 0;
   double recovery_period_avg   = 0;
   int    n_agents              = 0;

   // hack in case isolation for a test
   //double p_CI = 0.7;

   std::cout << "setting disease parameters for agents..." << std::endl;
   for (auto agent_entry : agents)
    {
       n_agents              += 1;
       incubation_period      = incubation_period_dist(generator);
       incubation_period_avg += incubation_period;

       recovery_period = recovery_period_dist(generator);

       //if ((rand() / RAND_MAX) < p_CI){
       //recovery_period = 1;
       //}

       recovery_period_avg += recovery_period;



       // hack in some case isolation

       agent_entry.second->setLATENT_PERIOD_((unsigned)latent_period * 2);

       agent_entry.second->setPEAK_VIRAL_TITER_((unsigned)(incubation_period * 2 + latent_period * 2));

       //agent_entry.second->setINFECTIOUS_PERIOD_((unsigned)(incubation_period * 2 + recovery_period * 2 + latent_period * 2));

       //agent_entry.second->setRECOVERY_TIME_((unsigned)(incubation_period * 2 + recovery_period * 2 + latent_period * 2));

       agent_entry.second->setINFECTION_PERIOD_((unsigned)(incubation_period * 2 + recovery_period * 2 + latent_period * 2));

       agent_entry.second->setPERCENT_SYMPTOMATIC(adult_symptomatic_fraction);

       agent_entry.second->setCHILD_PERCENT_SYMPTOMATIC(child_symptomatic_fraction);

       agent_entry.second->setASYMPTOMATIC_INFECTIVITY_(asymptomatic_infectivity);

       if (agent_entry.first == 0)
        {
           agent_entry.second->printTiterVsTime();
        }
       //CZ COV19 25032020 - added a SD and CI state reset to init() in agent.cpp
    }

   incubation_period_avg /= n_agents;
   recovery_period_avg   /= n_agents;

   std::cout << "average incubation period: " << incubation_period_avg << std::endl;
   std::cout << "average recovery period for: " << recovery_period_avg << std::endl;

   //Agent::computeViralLoad(); // CZ - no longer static.

   unsigned long cumulative_ill_running_total = 0;

   //CZ: counting pre-symptomatic transmissions
   unsigned long presymptomatic_transmissions_running_total = 0;

   //CZ: counting asymptomatic transmissions
   unsigned long asymptomatic_transmissions_running_total = 0;

   unsigned long chosen_airport = 0;
   for (unsigned run = 0; run < n_runs; ++run)
    {
       std::cout << "Initiating run " << run << "." << std::endl;

       /* Infect random agent at the start of each run */
#ifdef COUNT_SYDNEY_ONLY
       unsigned long valid_infected_agent_idx = rand() % valid_agent_ids.size();
       unsigned long infected_agent_idx       = valid_agent_ids.at(valid_infected_agent_idx);
#else
       unsigned long infected_agent_idx = rand() % agents.size();
#endif

       auto& agent_of_interest = agents.at(infected_agent_idx);

       /* Seeded SLAs are those with international airports */
       //Epidemic::init_seeded_slas( database, statistical_local_areas );

       std::cout << "Chosen to infect agent " << agent_of_interest->getID()
                 << " (in SLA:" << agent_of_interest->getSLA()->getID()
                 << ") for run " << run << "." << std::endl;

       auto             my_sla_id = agent_of_interest->getSLA()->getID();
       signed long long my_wg_id  = -1;
       if (!agent_of_interest->getWGs().empty())
        {
           my_wg_id = agent_of_interest->getWGs().front()->getID();
        }

       agent_map_type possible_infected_agents;
       for (auto& agent_entry : agents)
        {
           auto& cagent = agent_entry.second;
           if (cagent->getSLA()->getID() == my_sla_id)
            {
               possible_infected_agents.emplace(agent_entry);
               agent_entry.second->init();
               continue;
            }

           if ((my_wg_id > 0) && !cagent->getWGs().empty())
            {
               if (cagent->getWGs().front()->getID() == my_wg_id)
                {
                   possible_infected_agents.emplace(agent_entry);
                   agent_entry.second->init();
                }
            }
        }

       std::cout << "Number of agents in simulation: " << possible_infected_agents.size() << std::endl;

       /* Zero the buckets */
       init(households, household_clusters, census_districts, destination_zones,
            working_groups, schools, grades, classes, statistical_local_areas);

       g_cumulative_ill = 0;

       g_presymptomatic = 0;

       g_asymptomatic = 0;

       g_generation_steps = 0;

       g_serial_steps = 0;

       /* Infect our agent and simulate until recovered */
#ifdef INFECT_ILL_ONLY
       agent_of_interest->infectWithSymptoms();
#else
       agent_of_interest->infect(dum, dum, dum);
#endif

       // incorporating weekends - select day of week to start on
       int           day_of_week = rand() % 7;
       bool          weekend     = false; // initialise weekend to false, this is updated in the day loop

       unsigned long c_step            = 0;
       double        c_gen_time        = 0.0;
       double        c_serial_interval = 0.0;

       for (unsigned int day = 0; agent_of_interest->isInfected(); ++day)
        {
           // compute time between symptom onset of index case and symptom onset of infectees


           if (day_of_week == 5 || day_of_week == 6)
            {
               weekend = true;
            }
           else
            {
               weekend = false;
            }


           /* Zero the buckets the agent of interest interacts with */
           reZero(agent_of_interest);

           if (!weekend)
            {
               /* Update probabilities and infect agents during the first part of the day (at work/school) */
               c_step++;
               simulateDaytime(possible_infected_agents, agent_of_interest, c_step);

               /* Update probabilities and infect agents during the first part of the day (at home) */
               c_step++;
               simulateNighttime(possible_infected_agents, agent_of_interest, c_step);
            }
           else
            {
               /* Update probabilities and infect agents during the first part of the day (at work/school) */
               c_step++;
               simulateDaytimeWeekend(possible_infected_agents, agent_of_interest, c_step);

               /* Update probabilities and infect agents during the first part of the day (at home) */
               c_step++;
               simulateNighttime(possible_infected_agents, agent_of_interest, c_step);
            }

           if (agent_of_interest->willBeSymptomatic() &&
               agent_of_interest->getState() == DiseaseStates::StateAsymptomatic)
            {
               g_presymptomatic = g_cumulative_ill;
            }

           if (!agent_of_interest->willBeSymptomatic() &&
               agent_of_interest->getState() == DiseaseStates::StateAsymptomatic)
            {
               g_asymptomatic = g_cumulative_ill;
            }

           double incubation_steps_i = agent_of_interest->getPEAK_VIRAL_TITER_();

           c_gen_time = double( g_generation_steps ) / double( g_cumulative_ill ) / 2.0;

           c_serial_interval = double( g_serial_steps - incubation_steps_i * g_cumulative_ill ) / double( g_cumulative_ill ) / 2.0;

           unsigned long       c_ill_running_total = cumulative_ill_running_total + g_cumulative_ill;

           std::vector<double> perc_ill_in_group;
           for (auto cg : g_cumulative_ill_in_group)
            {
               perc_ill_in_group.push_back(double(cg) / c_ill_running_total);
            }

           std::cout << "Total number ill after work on day " << day
                     << ":" << g_cumulative_ill
                     << "\n" << "in mixing group percentage [HH, HC, CD, SLA, WG, CL, GR, SCH]: [";
           double perc_sum = 0.0;
           for (auto cp : perc_ill_in_group)
            {
               std::cout << cp * 100.0 << ", ";
            }
           std::cout << "];" << "\n" << "AOI state: " << agent_of_interest->getState()
                     << ";" << "\n" << "Current generation time: " << c_gen_time << std::endl;

           std::cout << "total transmissions: " << g_cumulative_ill << std::endl;
           std::cout << "pre-symptomatic transmissions: " << g_presymptomatic << std::endl;
           std::cout << "asymptomatic transmissions: " << g_asymptomatic << std::endl;
           std::cout <<"****"<<std::endl;


           day_of_week += 1;
           if (day_of_week >= 7)
            {
               day_of_week = 0;
            }
        }
       cumulative_ill_running_total += g_cumulative_ill;

       /*outfile_stream << g_cumulative_ill << "," << c_gen_time <<  std::endl; Unweighted method */
       outfile_stream << g_cumulative_ill << ","
                      << g_presymptomatic << ","
                      << g_asymptomatic << ","
                      << c_gen_time << ","
                      << c_serial_interval << ","
                      << agent_of_interest->getAgeGroup() << std::endl;
    }
   outfile_stream.close();

   return(0);
}


void updateAgentState(std::shared_ptr<Agent> c_agent, double P, unsigned long step, unsigned agroup)
{
   if (c_agent->isSusceptible())
    {
       if (rand() / (double)RAND_MAX < P)
        {
           c_agent->infect(dum, dum, dum);

           std::cout << "Agent " << c_agent->getID() << " infected at step " << step << " (P=" << P << ")" << std::endl;

#ifdef COUNT_ILL_ONLY
           /* Add to ill counter only if the agent will become symptomatic */
           if (c_agent->willBeSymptomatic())
            {
               g_cumulative_ill++;
               g_generation_steps += step;
               g_cumulative_ill_in_group.at(agroup)++;
            }
#else
           /* Add to counter */
           g_cumulative_ill++;
           g_generation_steps += step;
           g_serial_steps     += step + c_agent->getPEAK_VIRAL_TITER_();
           g_cumulative_ill_in_group.at(agroup)++;
#endif
        }
    }
}


void simulateDaytime(agent_map_type& agents, std::shared_ptr<Agent>& agent_of_interest, unsigned long step)
{
   agent_of_interest->step();

   /* Update transmissibility at work */
   agent_of_interest->addToDaytimeProbabilityOfInfection(dum, dum, dum);

   //std::cout << "Agent factor at step " << step << ": " << agent_of_interest->computeFactor() << std::endl;

   /* Update transmissibility at work */
   for (auto agent_entry : agents)
    {
       auto c_agent = agent_entry.second;
       if (c_agent->isStudent())
        {
           updateAgentState(c_agent, c_agent->getDaytimeProbabilityOfInfection(2), step, MixingGroups::GroupClass);
           updateAgentState(c_agent, c_agent->getDaytimeProbabilityOfInfection(1), step, MixingGroups::GroupGrade);
           updateAgentState(c_agent, c_agent->getDaytimeProbabilityOfInfection(0), step, MixingGroups::GroupSchool);
        }
       else
        {
           updateAgentState(c_agent, c_agent->getDaytimeProbabilityOfInfection(), step, MixingGroups::GroupWorkplace);
        }
    }
}


void simulateNighttime(agent_map_type& agents, std::shared_ptr<Agent>& agent_of_interest, unsigned long step)
{
   agent_of_interest->step();

   /* Update distributions for nighttime */
   agent_of_interest->addToNighttimeProbabilityOfInfection(dum, dum, dum);

   /* Decide who is now infected in the household */
   for (auto& agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       updateAgentState(c_agent, c_agent->getHouseholdProbabilityOfInfection(), step, MixingGroups::GroupHousehold);
       updateAgentState(c_agent, c_agent->getHouseholdClusterProbabilityOfInfection(), step, MixingGroups::GroupHouseholdCluster);
       updateAgentState(c_agent, c_agent->getNeighbourhoodProbabilityOfInfection(), step, MixingGroups::GroupNeighbourhood);
       updateAgentState(c_agent, c_agent->getCommunityProbabilityOfInfection(), step, MixingGroups::GroupCommunity);
    }
}

// for the R0 code, this is identical to SimulateNighttime (for main.cpp it's not exactly the same - bookkeeping is different)
void simulateDaytimeWeekend(agent_map_type& agents, std::shared_ptr<Agent>& agent_of_interest, unsigned long step)
{
   agent_of_interest->step();

   /* Update distributions for nighttime */
   agent_of_interest->addToNighttimeProbabilityOfInfection(dum, dum, dum);

   /* Decide who is now infected in the household */
   for (auto& agent_entry : agents)
    {
       auto c_agent = agent_entry.second;

       updateAgentState(c_agent, 0.0, step, MixingGroups::GroupHousehold); // dummy for bookkeeping
       //updateAgentState(c_agent, c_agent->getHouseholdProbabilityOfInfection(), step, MixingGroups::GroupHousehold);
       //updateAgentState(c_agent, c_agent->getHouseholdClusterProbabilityOfInfection(), step, MixingGroups::GroupHouseholdCluster);
       //updateAgentState(c_agent, c_agent->getNeighbourhoodProbabilityOfInfection(), step, MixingGroups::GroupNeighbourhood);
       //updateAgentState(c_agent, c_agent->getCommunityProbabilityOfInfection(), step, MixingGroups::GroupCommunity);
    }
}


void reZero(std::shared_ptr<Agent>& agent_of_interest)
{
   agent_of_interest->getHH()->reZero();
   agent_of_interest->getHC()->reZero();
   agent_of_interest->getCD()->reZero();
   agent_of_interest->getSLA()->reZero();

   if (!agent_of_interest->getWGs().empty())
    {
       for (auto cwg : agent_of_interest->getWGs())
        {
           cwg->reZero();
        }
       agent_of_interest->getDZ()->reZero();
    }
}


void init(hh_map_type& households, hc_map_type& household_clusters, cd_map_type& census_districts, dz_map_type& destination_zones,
   wg_map_type& working_groups, school_map_type& schools, wg_map_type& grades, wg_map_type& classes, sla_map_type& statistical_local_areas)
{
   for (auto& chh : households)
    {
       chh.second->reZero();
    }

   for (auto& chc : household_clusters)
    {
       chc.second->reZero();
    }

   for (auto& ccd : census_districts)
    {
       ccd.second->reZero();
    }

   for (auto& cdz : destination_zones)
    {
       cdz.second->reZero();
    }

   for (auto& cwg : working_groups)
    {
       cwg.second->reZero();
    }

   for (auto& csla : statistical_local_areas)
    {
       csla.second->reZero();
    }

   for (auto& csch : schools)
    {
       csch.second->reZero();
    }

   for (auto& cgr : grades)
    {
       cgr.second->reZero();
    }

   for (auto& ccl : classes)
    {
       ccl.second->reZero();
    }
}


/* Parse argument inputs (code "borrowed" from https://www.gnu.org/software/libc/manual/html_node/Getopt-Long-Option-Example.html#Getopt-Long-Option-Example).
 *  Options are:
 *  - # runs
 *  - # Database containing census data/for results
 */
void parse_arguments(int argc, char *argv[],
   double& p_trans, double& trans_scaler, unsigned& n_runs,
   std::string& database_output, std::string& database_input, std::string& Parameter_input_fname)
{
   static struct option long_options[] =
   {
      /*
       * Following options set a flag.
       */
      //{"runs", no_argument, &verbose_flag, 1},
      //{"brief",   no_argument,       &verbose_flag, 0},

      /*
       * Following options don’t set a flag.
       * We distinguish them by their indices.
       */
      { "ptrans",                no_argument, 0, 'p' },
      { "scaler",                no_argument, 0, 't' },
      { "runs",                  no_argument, 0, 'r' },
      { "database_output",       no_argument, 0, 'b' },
      { "database_input",        no_argument, 0, 'f' },
      { "Parameter_input_fname", no_argument, 0, 'P' },
      {                       0,           0, 0,   0 }
   };

   for ( ; ;)
    {
       /* getopt_long stores the option index here. */
       int  option_index = 0;

       char c = getopt_long(argc, argv, "p:t:r:d:s:a:b:f:P:", long_options, &option_index);

       /* Options are all parsed */
       if (c == -1)
        {
           break;
        }

       switch (c)
        {
        case 'p':
           p_trans = std::stof(optarg);
           break;

        case 't':
           trans_scaler = std::stof(optarg);
           break;

        case 'r':
           n_runs = std::stoi(optarg);
           break;

        case 'b':
           database_output = std::string(optarg);
           break;

        case 'f':
           database_input = std::string(optarg);
           break;

        case 'P':
           Parameter_input_fname = std::string(optarg);
           break;

        case '?':
           /* getopt_long already printed an error message. */
           break;

        default:
           abort();
        }
       std::cout << "option -" << c << " with value `" << optarg << "'" << std::endl;
    }

   /* Print any remaining command line arguments (not options). */
   if (optind < argc)
    {
       printf("Non-option ARGV-elements: ");
       while (optind < argc)
        {
           std::cout << argv[optind++] << ", ";
        }
       std::cout << std::endl;
    }
}


// CZ 10 06 2020 - read in the disease parameters in the Parameters_input.dat file

void parse_Parameters(const std::string& Parameter_input_fname)
{
   {
      std::string   param_file(Parameter_input_fname);
      std::ifstream param_stream(param_file);
      std::string   cline;

      while (std::getline(param_stream, cline))
       {
          std::stringstream i_ss;
          i_ss << cline;

          std::string first_element;
          i_ss >> first_element;

          if (first_element == "incubation_period_mu:")
           {
              i_ss >> incubation_period_mu;
           }
          if (first_element == "incubation_period_sig:")
           {
              i_ss >> incubation_period_sig;
           }
          if (first_element == "recovery_period_hi:")
           {
              i_ss >> recovery_period_hi;
           }
          if (first_element == "recovery_period_low:")
           {
              i_ss >> recovery_period_low;
           }
          if (first_element == "latent_period:")
           {
              i_ss >> latent_period;
           }
          if (first_element == "adult_symptomatic_fraction:")
           {
              i_ss >> adult_symptomatic_fraction;
           }
          if (first_element == "child_symptomatic_fraction:")
           {
              i_ss >> child_symptomatic_fraction;
           }
          if (first_element == "asymptomatic_infectivity:")
           {
              i_ss >> asymptomatic_infectivity;
           }
       }

      param_stream.close();
   }
}
