/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright   Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-201), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright   Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   This code is released only for internal review by Nature Communications, manuscript: NCOMMS-21-29568.

   *EndCopyright:
*/

#ifndef __STATISTICAL_AREAS_H__
#define __STATISTICAL_AREAS_H__

#include "enums.h"
#include "debug.h"

#include <iostream>
#include <vector>
#include <map>
#include <memory>

/*
 * Base class - all statistical areas (SAs) have a number of agents, an ID (of type T), and a transmission probability
 */
template <class T> class StatisticalAreaBase
{
protected:
   const double P_ZERO = 1.0;

   /* ID associated with this SA */
   const T ID_;

   /* Agents in this SA */
   std::vector <unsigned long> agent_idxs_;

   /* Joint probability of transmission in this SA (dynamic) */
   std::vector <double> joint_transmission_probability_;

   bool Symptomatic_infection_;   // An indicator variable for detection in the area

   unsigned long n_infecteds_tot; // A counter for the total infected prevalence (symptomatic and asymptomatic)

   unsigned long n_infecteds_;

public:

   /* Construct/default destructor */
   StatisticalAreaBase(T ID) : ID_(ID), n_infecteds_(0)
   {
   }

   virtual ~StatisticalAreaBase()
   {
   };

   /* Get ID of this SA */
   inline T getID(void)
   {
      return(ID_);
   }

   /* Get vector of agents' (IDs) in this SA */
   inline std::vector <unsigned long>& getAgentIdxs(void)
   {
      return(agent_idxs_);
   }

   /* Get probability of infection in this SA */
   inline double getProbabilityOfTransmission(unsigned short idx)
   {
      return(joint_transmission_probability_.at(idx));
   }

   /* Add an agent (ID) to this SA */
   inline void addAgentIdx(unsigned long idx)
   {
      return(agent_idxs_.push_back(idx));
   }

   /* Re-zero the transmission probabilities */
   //inline virtual void reZero(void)

   inline void reZero(void)
   {
      for (auto& prob : joint_transmission_probability_)
       {
          prob = P_ZERO;
       }

      Symptomatic_infection_ = false;

      n_infecteds_ = 0;
   }

   inline unsigned long getNumberOfInfects(void)
   {
      return(n_infecteds_);
   }
};

/*
 * Statistical local area data structure
 */
class StatisticalLocalArea : public StatisticalAreaBase <unsigned long>
{
   /* Independent probability of transmission in this SLA (constant) */
   static std::vector <double> independent_transmission_probability_;

   static constexpr double ORIG_CR_SIZE_ = 4300.0;
   static constexpr double ORIG_CR_AREA_ = 150518872.0;
   static constexpr double OUR_CR_AREA_  = 5409282700.0;
   double density_scaling_;

public:

   ~StatisticalLocalArea()
   {
   }

   StatisticalLocalArea(unsigned long ID) : StatisticalAreaBase <unsigned long>(ID)
   {
      joint_transmission_probability_ = std::vector <double>(1);
   }

   inline static void addTransmissionRate(double rate)
   {
      independent_transmission_probability_.push_back(rate);
   }

   inline void computeDensity(void)
   {
      //density_scaling_ = ( ORIG_CR_AREA_ / ORIG_CR_SIZE_ ) / ( OUR_CR_AREA_ / double(agent_idxs_.size()) );
      density_scaling_ = ORIG_CR_SIZE_ / double(agent_idxs_.size());

#ifdef DEBUG_MODE_
      std::cout << "density scaling SLA:" << this->getID() << " (" << agent_idxs_.size() << " agents): " << density_scaling_ << std::endl;
#endif
   }

   //inline virtual double addInfectedAgent(AgeGroup ag, double factor)
   inline void addInfectedAgent(AgeGroup ag, double factor)
   {
      double P_inf = independent_transmission_probability_.at(ag) * factor * density_scaling_;

#ifdef DEBUG_MODE_
      std::cout << "P_inf for SLA:" << this->getID()
                << "\n\tOrig P_inf (ag:" << ag << "): " << independent_transmission_probability_.at(ag)
                << "\n\tFactor: " << factor
                << "\n\tDensity scaling: " << density_scaling_
                << "\n\tFinal P_inf: " << P_inf << std::endl;
#endif

      joint_transmission_probability_.at(0) *= 1.0 - P_inf;
      n_infecteds_++;
   }

   inline double getProbabilityOfTransmission(void)
   {
      return(joint_transmission_probability_.at(0));
   }
};

/*
 * Census district data structure
 */
class CensusDistrict : public StatisticalAreaBase <unsigned long>
{
private:
   const std::shared_ptr <StatisticalLocalArea> sla_;
   const float latitude_;
   const float longitude_;

   static constexpr double ORIG_CR_SIZE_ = 2000.0;      // people
   static constexpr double ORIG_CR_AREA_ = 69992883.0;  // m^2
   static constexpr double OUR_CR_AREA_  = 198816201.0; // m^2
   double density_scaling_;
   bool infected_CD       = false;
   bool newly_infected_CD = false;

   static std::vector <double> independent_transmission_probability_;

public:

   ~CensusDistrict()
   {
   }

   CensusDistrict(const unsigned long ID,
      const std::shared_ptr <StatisticalLocalArea>& sla,
      const double latitude,
      const double longitude)
      : StatisticalAreaBase <unsigned long>(ID),
      sla_(sla), latitude_(latitude), longitude_(longitude)
   {
      joint_transmission_probability_ = std::vector <double>(1);
      Symptomatic_infection_          = false;
      n_infecteds_tot                 = 0;
   }

   inline const std::shared_ptr <StatisticalLocalArea>& getSLA(void)
   {
      return(sla_);
   }

   inline static void addTransmissionRate(double rate)
   {
      independent_transmission_probability_.push_back(rate);
   }

   inline void computeDensity(void)
   {
      //density_scaling_ = ( ORIG_CR_AREA_ / ORIG_CR_SIZE_ ) / ( OUR_CR_AREA_ / double(agent_idxs_.size()) );
      density_scaling_ = ORIG_CR_SIZE_ / double(agent_idxs_.size());
      //std::cout << "density scaling CD:" << this->getID() << " (" << agent_idxs_.size() << " agents): " << density_scaling_ << std::endl;
   }

   /* overloaded method */
   inline void addAgentIdx(unsigned long idx)
   {
      sla_->addAgentIdx(idx);
      agent_idxs_.push_back(idx);
   }


   // CZ: something weird here - I think this is supposed to be inline void, since this is the derived class
   //inline virtual double addInfectedAgent(AgeGroup ag, double factor)
   inline void addInfectedAgent(AgeGroup ag, double factor)
   {
      double P_inf = independent_transmission_probability_.at(ag) * factor * density_scaling_;

#ifdef DEBUG_MODE_
      std::cout << "P_inf for CD:" << this->getID()
                << "\n\tOrig P_inf (ag:" << ag << "): " << independent_transmission_probability_.at(ag)
                << "\n\tFactor: " << factor
                << "\n\tDensity scaling: " << density_scaling_
                << "\n\tFinal P_inf: " << P_inf << std::endl;
#endif

      //std::cout << "adding to joint trans prob" << std::endl;

      joint_transmission_probability_.at(0) *= 1.0 - P_inf;
      n_infecteds_++;

      //std::cout << "added to joint trans prob" << std::endl;
   }



   inline double getProbabilityOfTransmission(void)
   {
      return(joint_transmission_probability_.at(0));
   }

   // This is for the intervention. It adds 1 to the number of sypmtomatic infected for a given location
   inline void addSymptomaticAgent(void)
   {
      Symptomatic_infection_ = true;
   }

   inline bool getSymptomaticInfection(void)
   {
      return(Symptomatic_infection_);
   }

   inline void add_n_infecteds_tot(void)
   {
      n_infecteds_tot += 1.0;
   }

   inline void remove_n_infecteds_tot(void)
   {
      n_infecteds_tot -= 1.0;
   }

   inline void rezero_n_infecteds_tot(void)
   {
      n_infecteds_tot = 0;
   }

   inline unsigned long get_n_infecteds_tot(void)
   {
      return(n_infecteds_tot);
   }

   inline void infect_CD(void)
   {
      newly_infected_CD = true;
      infected_CD       = true;
   }

   inline void is_not_newly_infected(void)
   {
      newly_infected_CD = false;
   }

   inline void uninfect_CD(void)
   {
      infected_CD = false;
   }

   inline bool is_CD_newly_infected(void)
   {
      return(newly_infected_CD);
   }
};

/*
 * Destination zone data structure
 */
class DestinationZone : public StatisticalAreaBase <unsigned long long>
{
private:

   const std::shared_ptr <StatisticalLocalArea> sla_;

public:

   ~DestinationZone()
   {
   }

   DestinationZone(const unsigned long long ID,
      const std::shared_ptr <StatisticalLocalArea>& sla)
      : StatisticalAreaBase <unsigned long long>(ID), sla_(sla)
   {
   }

   inline const std::shared_ptr <StatisticalLocalArea>& getSLA(void)
   {
      return(sla_);
   }
};

/*
 * Working group data structure
 */
class WorkingGroup : public StatisticalAreaBase <unsigned long>
{
private:

   const std::shared_ptr <DestinationZone> dz_;

   double density_scaling_;
   const int type_;

   static const std::vector <double> ORIG_TR_SIZE_; // people

   static std::vector <double> independent_transmission_probability_;

public:

   ~WorkingGroup()
   {
   }

   WorkingGroup(const unsigned long ID,
      const int type,
      const std::shared_ptr <DestinationZone>& dz)
      : StatisticalAreaBase <unsigned long>(ID), dz_(dz), type_(type)
   {
      joint_transmission_probability_ = std::vector <double>(1);
      Symptomatic_infection_          = false;
   }

   inline void computeDensity(void)
   {
      density_scaling_ = ORIG_TR_SIZE_.at(type_) / double( agent_idxs_.size());
   }

   /* Getter methods */
   inline const std::shared_ptr <DestinationZone>& getDZ(void)
   {
      return(dz_);
   }

   inline int getType(void)
   {
      return(type_);
   }

   inline double getProbabilityOfTransmission(void)
   {
      /* Only one prob of transmission in a WG */
      return(joint_transmission_probability_.at(0));
   }

   /* Overloaded methods */
   inline void addAgentIdx(unsigned long idx)
   {
      dz_->addAgentIdx(idx);
      agent_idxs_.push_back(idx);
   }

   inline static void addTransmissionRate(double rate)
   {
      independent_transmission_probability_.push_back(rate);
   }

   inline void addInfectedAgent(double factor)
   {
      double P_inf = independent_transmission_probability_.at(type_);

      //std::cout << "P_inf for wg: " << P_inf << std::endl;
      //std::cout << "Factor: " << factor << std::endl;

      /* only one prob/rate for WGs */
      joint_transmission_probability_.at(0) *= 1.0 - independent_transmission_probability_.at(type_) * factor * density_scaling_;
      n_infecteds_++;
   }

   inline void addSymptomaticAgent(void)
   {
      Symptomatic_infection_ = true;
   }

   inline bool getSymptomaticInfection(void)
   {
      return(Symptomatic_infection_);
   }
};

/*
 * Household cluster data structure
 */
class HouseholdCluster : public StatisticalAreaBase <unsigned long>
{
private:

   const static unsigned short CHILD_IDX = 0, ADULT_IDX = 1;

   const std::shared_ptr <CensusDistrict> cd_;

   /* Special cases for household where we have different transmission rates/probs depending on who it's coming from */
   static std::vector <std::vector <double> > independent_transmission_probability_matrix_;
   std::vector <std::vector <double> > joint_transmission_probability_matrix_;

public:

   ~HouseholdCluster()
   {
   }

   HouseholdCluster(const unsigned long ID,
      const std::shared_ptr <CensusDistrict>& cd)
      : StatisticalAreaBase <unsigned long>(ID), cd_(cd)
   {
      joint_transmission_probability_ = std::vector <double>(5);
      for (unsigned i = 0; i < 5; ++i)
       {
          joint_transmission_probability_matrix_.push_back(std::vector <double>(5));
       }
   }

   inline const std::shared_ptr <CensusDistrict>& getCD(void)
   {
      return(cd_);
   };

   /* Overloaded method */
   inline void addInfectedAgent(bool is_adult, double factor)
   {
      /* Update by individual elements */
      double P_adult_to_child = independent_transmission_probability_matrix_.at(is_adult).at(CHILD_IDX);
      double P_adult_to_adult = independent_transmission_probability_matrix_.at(is_adult).at(ADULT_IDX);

      //std::cout << "P_inf for cluster (adult->child): " << P_adult_to_child << std::endl;
      //std::cout << "P_inf for cluster (adult->adult): " << P_adult_to_adult << std::endl;
      //std::cout << "Factor: " << factor << std::endl;

      joint_transmission_probability_matrix_.at(is_adult).at(CHILD_IDX) *= 1.0 - P_adult_to_child * factor;
      joint_transmission_probability_matrix_.at(is_adult).at(ADULT_IDX) *= 1.0 - P_adult_to_adult * factor;

      n_infecteds_++;
   }

   inline double getProbabilityOfTransmission(bool is_adult)
   {
      return(joint_transmission_probability_matrix_.at(CHILD_IDX).at(is_adult)
             * joint_transmission_probability_matrix_.at(ADULT_IDX).at(is_adult));
   }

   inline static void addTransmissionRate(bool is_adult, double rate)
   {
      if (independent_transmission_probability_matrix_.size() <= is_adult)
       {
          independent_transmission_probability_matrix_.push_back(std::vector <double>());
       }
      independent_transmission_probability_matrix_.at(is_adult).push_back(rate);
   }

   inline void addAgentIdx(unsigned long idx)
   {
      cd_->addAgentIdx(idx);
      agent_idxs_.push_back(idx);
   }

   inline void reZero(void)
   {
      for (auto& tp : joint_transmission_probability_)
       {
          tp = P_ZERO;
       }


      for (auto& tp_row : joint_transmission_probability_matrix_)
       {
          for (auto& tp : tp_row)
           {
              tp = P_ZERO;
           }
       }
      n_infecteds_ = 0;
   }
};

/*
 * Household data structure
 */
class Household : public StatisticalAreaBase <unsigned long>
{
private:
   const std::shared_ptr <HouseholdCluster> hc_;

   static std::vector <std::vector <double> > independent_transmission_probability_matrix_;

   // CZ COV19 08042020 - this is a flag indicating that a household has a case-isolated illness
   // this will be used as a condition for home quarantine.

   bool CI_in_household = false;


public:

   ~Household()
   {
   }

   Household(const unsigned long ID,
      const std::shared_ptr <HouseholdCluster>& hc)
      : StatisticalAreaBase <unsigned long>(ID), hc_(hc)
   {
      joint_transmission_probability_ = std::vector <double>(1),
      Symptomatic_infection_          = false;
   }

   inline const std::shared_ptr <HouseholdCluster>& getHC(void)
   {
      return(hc_);
   }

   inline unsigned getHouseholdSize(void)
   {
      return(this->agent_idxs_.size());
   }

   inline HouseholdSizeGroup getHouseholdSizeGroup(void)
   {
      HouseholdSizeGroup my_size_group;

      switch (this->getHouseholdSize())
       {
       case 1:
       case 2:
          my_size_group = HouseholdSizeGroups::TwoOrLessPeople;
          break;

       case 3:
          my_size_group = HouseholdSizeGroups::ThreePeople;
          break;

       case 4:
          my_size_group = HouseholdSizeGroups::FourPeople;
          break;

       case 5:
          my_size_group = HouseholdSizeGroups::FivePeople;
          break;

       default:
          my_size_group = HouseholdSizeGroups::SixOrMorePeople;
       }
      return(my_size_group);
   }

   /* Get probability of infection in this household */
   inline double getProbabilityOfTransmission(void)
   {
      return(joint_transmission_probability_.at(0));
   }

   inline static void addTransmissionRate(unsigned short hh_size, double rate)
   {
      if (independent_transmission_probability_matrix_.size() <= hh_size)
       {
          independent_transmission_probability_matrix_.push_back(std::vector <double>());
       }
      independent_transmission_probability_matrix_.at(hh_size).push_back(rate);
   }

   inline void addInfectedAgent(AgeGroup ag, double factor)
   {
      double P_inf = independent_transmission_probability_matrix_.at(this->getHouseholdSizeGroup()).at(ag);

      //std::cout << "P_inf for household (size " << this->getHouseholdSizeGroup() << ", ag: " << ag << "): " << P_inf << std::endl;
      //std::cout << "Factor: " << factor << std::endl;
      joint_transmission_probability_.at(0) *= 1.0 - P_inf * factor;

      n_infecteds_++;
   }

   inline void addAgentIdx(unsigned long idx)
   {
      hc_->addAgentIdx(idx);
      agent_idxs_.push_back(idx);
   }

   inline void reZero(void)
   {
      for (auto& tp : joint_transmission_probability_)
       {
          tp = P_ZERO;
       }

      Symptomatic_infection_ = false;

      n_infecteds_ = 0;

      CI_in_household = false;
   }

   inline void addSymptomaticAgent(void)
   {
      Symptomatic_infection_ = true;
   }

   inline bool getSymptomaticInfection(void)
   {
      return(Symptomatic_infection_);
   }

   inline bool getCIinHousehold(void)
   {
      return CI_in_household;
   }

   inline void setCIinHousehold(bool CI_flag)
   {
      CI_in_household = CI_flag;
   }
};


typedef std::map <unsigned long, std::shared_ptr <StatisticalLocalArea> >   sla_map_type;
typedef std::map <unsigned long, std::shared_ptr <CensusDistrict> >         cd_map_type;
typedef std::map <unsigned long long, std::shared_ptr <DestinationZone> >   dz_map_type;
typedef std::map <signed long, std::shared_ptr <WorkingGroup> >             wg_map_type;
typedef std::map <signed long, std::shared_ptr <WorkingGroup> >             school_map_type;
typedef std::map <unsigned long, std::shared_ptr <HouseholdCluster> >       hc_map_type;
typedef std::map <unsigned long, std::shared_ptr <Household> >              hh_map_type;

#endif // __STATISTICAL_AREAS_H__
